/*
 * C
 *
 * Copyright 2014-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

/* Includes --------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>

#include "sni.h"
#include "LLNET_CONSTANTS.h"
#include "LLNET_ERRORS.h"
#include "NET_delegate_task.h"
#include "LLECOM_WIFI_impl.h"


//#define LLECOM_WIFI_ENABLE_TRACE

#ifdef LLECOM_WIFI_ENABLE_TRACE
#define ASSERT_LLECOM_WIFI(x) do{if(!(x)){printf("assert %s:%d\n", __FILE__, __LINE__);while(1);}}while(0)
#else
#define ASSERT_LLECOM_WIFI(x) ((void) 0)
#endif


int32_t LLECOM_WIFI_IMPL_disableSoftAP(int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_DISABLE_SOFT_AP_EVENT;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_disable_soft_ap_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_enableSoftAP(int8_t* ssid, int32_t ssidOffset, int32_t ssidLength, int8_t* passphrase, int32_t passphraseOffset, int32_t passphraseLength, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	ASSERT_LLECOM_WIFI(ssid != NULL);
	ASSERT_LLECOM_WIFI(ssidOffset < ssidLength);
	ASSERT_LLECOM_WIFI(passphrase != NULL);
	ASSERT_LLECOM_WIFI(passphraseOffset < passphraseLength);

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_ENABLE_SOFT_AP_EVENT;
			pst_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.pc_ssid_in = (char*) (ssid + ssidOffset);
			pst_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.i_ssid_length_in = ssidLength - ssidOffset;
			pst_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.pc_passphrase_in = (char*) (passphrase + passphraseOffset);
			pst_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.i_passphrase_length_in = passphraseLength - passphraseOffset;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_enable_soft_ap_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_setNameSoftAP(int8_t* name, int32_t nameOffset, int32_t nameLength, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	ASSERT_LLECOM_WIFI(name != NULL);
	ASSERT_LLECOM_WIFI(nameOffset < nameLength);

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_SET_NAME_SOFT_AP_EVENT;
			pst_event->u_ctx_event.st_wifi_set_name_soft_ap_ctx_t.pc_name_in = (char*) (name + nameOffset);
			pst_event->u_ctx_event.st_wifi_set_name_soft_ap_ctx_t.i_name_length_in = nameLength - nameOffset;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_set_name_soft_ap_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_getBSSID(int8_t* bssid, int32_t bssidOffset, int32_t bssidLength, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	ASSERT_LLECOM_WIFI(bssid != NULL);
	ASSERT_LLECOM_WIFI(bssidOffset < bssidLength);

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("(%s) bssid=%p, bssidOffset=%d, bssidLength=%d, getResult=%hhd\n", __func__, bssid, bssidOffset, bssidLength, getResult);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_BSSID_EVENT;
			pst_event->u_ctx_event.st_wifi_get_bssid_ctx_t.i_bssid_length_in = bssidLength - bssidOffset;
			pst_event->u_ctx_event.st_wifi_get_bssid_ctx_t.pc_bssid_in = (char*) (bssid + bssidOffset);

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_get_bssid_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}


int32_t LLECOM_WIFI_IMPL_getChannel(int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_CHANNEL_EVENT;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_getClientState(int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s: %d\n", __func__, getResult);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_CLIENT_STATE;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_get_client_state_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}
	return i_return;
}

int32_t LLECOM_WIFI_IMPL_getRSSI(float* rssi, int32_t rssiOffset, int32_t rssiLength, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	ASSERT_LLECOM_WIFI(rssi != NULL);
	ASSERT_LLECOM_WIFI(rssiOffset < rssiLength);

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_RSSI_EVENT;
			pst_event->u_ctx_event.st_wifi_get_rssi_ctx_t.pf_rssi_in = rssi + rssiOffset;
			pst_event->u_ctx_event.st_wifi_get_rssi_ctx_t.i_rssi_length_in = rssiLength - rssiOffset;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_get_rssi_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_getSecurityMode(int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_SECURITY_MODE_EVENT;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_getSSID(int8_t* ssid, int32_t ssidOffset, int32_t ssidLength, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	ASSERT_LLECOM_WIFI(ssid!= NULL);
	ASSERT_LLECOM_WIFI(ssidOffset < ssidLength);

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_SSID_EVENT;
			pst_event->u_ctx_event.st_wifi_get_ssid_ctx_t.i_ssid_length_in = ssidLength - ssidOffset;
			pst_event->u_ctx_event.st_wifi_get_ssid_ctx_t.pc_ssid_in = (char*) (ssid + ssidOffset);

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_get_ssid_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_getWPSModes(int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_WPS_MODES_EVENT;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_join(int8_t* ssid, int32_t ssidOffset, int32_t ssidLength, int8_t* passphrase, int32_t passphraseOffset, int32_t passphraseLength, int8_t getResult)
{

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	return LLECOM_WIFI_IMPL_joinWithSecurityMode(ssid, ssidOffset, ssidLength, passphrase, passphraseOffset, passphraseLength, SECURITY_MODE_WPA2, getResult);
}

int32_t LLECOM_WIFI_IMPL_joinWithSecurityMode(int8_t* ssid, int32_t ssidOffset, int32_t ssidLength, int8_t* passphrase, int32_t passphraseOffset, int32_t passphraseLength, int32_t securityModeValue, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	ASSERT_LLECOM_WIFI(ssid != NULL);
	ASSERT_LLECOM_WIFI(ssidOffset < ssidLength);
	ASSERT_LLECOM_WIFI(passphrase != NULL);
	ASSERT_LLECOM_WIFI(passphraseOffset < passphraseLength);
	ASSERT_LLECOM_WIFI(securityModeValue >= 0);


	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s ssid %p ssidOffset last byte 0x%02X %d ssidLength %d passphrase %p passphraseOffset %d passphraseLength %d\n", __func__, ssid, ssid[ssidLength - 1], ssidOffset, ssidLength, passphrase,passphraseOffset, passphraseLength);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_JOIN_EVENT;
			pst_event->u_ctx_event.st_wifi_join_ctx_t.pc_ssid_in = (char*) (ssid + ssidOffset);
			pst_event->u_ctx_event.st_wifi_join_ctx_t.i_ssid_length_in = ssidLength - ssidOffset;
			pst_event->u_ctx_event.st_wifi_join_ctx_t.pc_passphrase_in = (char*) (passphrase + passphraseOffset);
			pst_event->u_ctx_event.st_wifi_join_ctx_t.i_passphrase_length_in = passphraseLength - passphraseOffset;
			pst_event->u_ctx_event.st_wifi_join_ctx_t.i_security_mode_in = securityModeValue;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_join_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s ssid %p ssidOffset last byte 0x%02X %d ssidLength %d passphrase %p passphraseOffset %d passphraseLength %d\n", __func__, ssid, ssid[ssidLength - 1], ssidOffset, ssidLength, passphrase,passphraseOffset, passphraseLength);
	#endif

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_leave(int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return = 0;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_LEAVE_EVENT;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_scanAPCount(int8_t active, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	i_return = 0;

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_AP_COUNT_EVENT;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_generic_only_return_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}


int32_t LLECOM_WIFI_IMPL_scanAPssidAt(int8_t* ssid, int32_t ssidOffset,	int32_t ssidLength, int32_t index, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	ASSERT_LLECOM_WIFI(ssid != NULL);
	ASSERT_LLECOM_WIFI(ssidOffset < ssidLength);

	i_return = 0;

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_AP_SSID_EVENT;
			pst_event->u_ctx_event.st_wifi_get_ap_ssid_ctx_t.puc_ssid_in = (unsigned char *) (ssid + ssidOffset);
			pst_event->u_ctx_event.st_wifi_get_ap_ssid_ctx_t.i_ssid_length_in = ssidLength - ssidOffset;
			pst_event->u_ctx_event.st_wifi_get_ap_ssid_ctx_t.i_index_in = index;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_get_ap_ssid_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}


int32_t LLECOM_WIFI_IMPL_scanAPbssidAt(int8_t* bssid, int32_t bssidOffset, int32_t bssidLength, int32_t index, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	ASSERT_LLECOM_WIFI(bssid != NULL);
	ASSERT_LLECOM_WIFI(bssidOffset < bssidLength);

	i_return = 0;

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_AP_BSSID_EVENT;
			pst_event->u_ctx_event.st_wifi_get_ap_bssid_ctx_t.puc_bssid_in = (unsigned char *) bssid + bssidOffset;
			pst_event->u_ctx_event.st_wifi_get_ap_bssid_ctx_t.i_bssid_length_in = bssidLength - bssidOffset;
			pst_event->u_ctx_event.st_wifi_get_ap_bssid_ctx_t.i_index_in = index;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_get_ap_bssid_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_scanAPchannelAt(int32_t index, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	i_return = 0;

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_AP_CHANNEL_EVENT;
			pst_event->u_ctx_event.st_wifi_get_ap_channel_ctx_t.i_index_in = index;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_get_ap_channel_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_scanAPrssiAt(float* rssi, int32_t rssiOffset, int32_t rssiLength, int32_t index, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	ASSERT_LLECOM_WIFI(rssi != NULL);
	ASSERT_LLECOM_WIFI(rssiOffset < rssiLength);

	i_return = 0;

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_AP_RSSI_EVENT;
			pst_event->u_ctx_event.st_wifi_get_ap_rssi_ctx_t.pf_rssi_in = rssi + rssiOffset;
			pst_event->u_ctx_event.st_wifi_get_ap_rssi_ctx_t.i_rssi_length_in = rssiLength - rssiOffset;
			pst_event->u_ctx_event.st_wifi_get_ap_rssi_ctx_t.i_index_in = index;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_get_ap_rssi_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_scanAPsecurityModeAt(int32_t index, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	i_return = 0;

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_AP_SECURITY_MODE_EVENT;
			pst_event->u_ctx_event.st_wifi_get_ap_security_mode_ctx_t.i_index_in = index;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_get_ap_security_mode_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_scanAPwpsModeAt(int32_t index, int8_t getResult)
{
	NET_event_t * pst_event;
	int32_t i_return;

	#ifdef LLECOM_WIFI_ENABLE_TRACE
	printf("%s\n", __func__);
	#endif

	i_return = 0;

	// If not retry, this is the first execution so post event
	if (getResult != JTRUE)
	{
		if (NET_get_event(&pst_event) != NET_NO_ERROR)
		{
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			pst_event->e_id_event = NET_WIFI_GET_AP_WPS_MODES_EVENT;
			pst_event->u_ctx_event.st_wifi_get_ap_wps_modes_ctx_t.i_index_in = index;

			if (NET_post_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
			else
			{
				// Must return this code in case of on error to be recall with retry to true
				i_return = (int32_t) J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT;
			}
		}
	}
	else // Else it's second execution the thread has just been resumed by end of delegate process
	{
		// Retrieve event in pool according to java thread id
		if (NET_retrieve_event(&pst_event, SNI_getCurrentJavaThreadID()) != NET_NO_ERROR)
		{
			// An error occurred
			i_return = (int32_t) J_ENOMEM;
		}
		else
		{
			// Return value from context
			i_return = (int32_t) pst_event->u_ctx_event.st_wifi_get_ap_wps_modes_ctx_t.s_error_out;

			// Release pool event
			if (NET_release_event(pst_event) != NET_NO_ERROR)
			{
				// An error occurred
				i_return = (int32_t) J_ENOMEM;
			}
		}
	}

	return i_return;
}

int32_t LLECOM_WIFI_IMPL_getCapability(int8_t getResult){
	return WIFI_CAPABILITY_BOTH_EXCLUSIVE; // FIXME link to the hardware part
}
