/*
 * C
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

// BSP
#include "fsl_debug_console.h"
#include "fsl_dma.h"
#include "fsl_fbdev.h"

// RTOS
#include "FreeRTOS.h"
#include "semphr.h"

// Project
#include "display_support.h"

#include "board.h"
#include "pin_mux.h"

#include "LLUI_DISPLAY_impl.h"
#include "display_task.h"
#include "display_dma.h"
#include "display_utils.h"
#include "display_vglite.h"
#include "interrupts.h"
#include "LLVECTOR_FONT_impl.h"

#include "mej_debug.h"
#include "mej_log.h"

// -----------------------------------------------------------------------------
// Defines
// -----------------------------------------------------------------------------

#if (CUSTOM_VGLITE_MEMORY_CONFIG != 1)
#error "Application must be compiled with CUSTOM_VGLITE_MEMORY_CONFIG=1"
#else
#define VGLITE_COMMAND_BUFFER_SZ (128 << 10)
#define VGLITE_HEAP_SZ 0x100000
#endif

// -----------------------------------------------------------------------------
// Global Variables
// -----------------------------------------------------------------------------

/* Allocate the heap and set the command buffer(s) size */
AT_NONCACHEABLE_SECTION_ALIGN(uint8_t vglite_heap[VGLITE_HEAP_SZ], 64);

void *vglite_heap_base        = &vglite_heap;
uint32_t vglite_heap_size     = VGLITE_HEAP_SZ;
uint32_t vglite_cmd_buff_size = VGLITE_COMMAND_BUFFER_SZ;

/*
 * @brief Framebuffer context
 */
fbdev_t g_fbdev;

/*
 * @brief Framebuffer information
 */
fbdev_fb_info_t g_fbInfo;

// -----------------------------------------------------------------------------
// Public functions
// -----------------------------------------------------------------------------

// See the header file for the function documentation
void LLUI_DISPLAY_IMPL_initialize(LLUI_DISPLAY_SInitData* init_data)
{
	status_t status;
	vg_lite_error_t ret;
	vg_lite_window_t* window;

	//MEJ_DEBUG_enable(MEJ_DEBUG_DISPLAY_IMPL_LVL);
	MEJ_DEBUG_enable(MEJ_DEBUG_MICROUI_DRAWINGS_LVL);
	MEJ_DEBUG_enable(MEJ_DEBUG_VGLITE_LVL);

	MEJ_DEBUG(MEJ_DEBUG_DISPLAY_IMPL_LVL, "enter\n");

	/***************
	 * Init VGLite *
	 ***************/

	status = BOARD_PrepareVGLiteController();
	if (kStatus_Success != status) {
		MEJ_LOG_CRITICAL("Prepare VGlite contolor error");
	}

	ret = display_vglite_init();
	if (VG_LITE_SUCCESS != ret) {
		MEJ_LOG_CRITICAL("init_vg_lite failed: init_vg_lite() returned error %d", ret);
	}

	/************
	 * Init DMA *
	 ************/

#if APP_BUFFER_COUNT > 1
	static void ** tmp;

	tmp = VGLITE_GetBuffers();

	for (int i = 0; i < APP_BUFFER_COUNT; i++) {
		s_frameBufferAddress[i] = tmp[i];
	}
#endif

#if DISPLAY_DMA_ENABLED != 0
	display_dma_initialize(s_frameBufferAddress);
#endif // defined DISPLAY_DMA_ENABLED

	window = display_vglite_get_window();
	display_task_initialize(window);

	vg_lite_buffer_t *buffer = VGLITE_GetRenderTarget(window);
	init_data->binary_semaphore_0 = (void*)xSemaphoreCreateBinary();
	init_data->binary_semaphore_1 = (void*)xSemaphoreCreateBinary();
	init_data->lcd_width = window->width;
	init_data->lcd_height = window->height;
	init_data->memory_width = FRAME_BUFFER_STRIDE_PIXELS;
	init_data->back_buffer_address = (uint8_t*)buffer->memory;

	LLVECTOR_FONT_init();

	MEJ_DEBUG(MEJ_DEBUG_DISPLAY_IMPL_LVL, "leave\n");
}

// See the header file for the function documentation
uint8_t* LLUI_DISPLAY_IMPL_flush(MICROUI_GraphicsContext* gc, uint8_t* addr, uint32_t xmin, uint32_t ymin, uint32_t xmax, uint32_t ymax)
{
	vg_lite_buffer_t *next_buffer;

	MEJ_DEBUG(MEJ_DEBUG_DISPLAY_IMPL_LVL, "enter\n");

	MEJ_DEBUG(MEJ_DEBUG_DISPLAY_IMPL_LVL, "params: (addr: %p, xmin: %d, ymin: %d, xmax: %d, ymax: %d\n", addr, xmin, ymin, xmax, ymax);
	MEJ_DEBUG(MEJ_DEBUG_DISPLAY_IMPL_LVL, "synchronize display task\n");

	next_buffer = display_vglite_get_next_graphics_buffer();
	MEJ_DEBUG(MEJ_DEBUG_DISPLAY_IMPL_LVL, "next buffer: 0x%p\n", next_buffer->memory);

	display_task_flush(addr, xmin, ymin, xmax, ymax);

	MEJ_DEBUG(MEJ_DEBUG_DISPLAY_IMPL_LVL, "leave\n");

	return (uint8_t*) next_buffer->memory;
}

// See the header file for the function documentation
void LLUI_DISPLAY_IMPL_binarySemaphoreTake(void* sem)
{
	xSemaphoreTake((SemaphoreHandle_t)sem, portMAX_DELAY);
}

// See the header file for the function documentation
void LLUI_DISPLAY_IMPL_binarySemaphoreGive(void* sem, bool under_isr)
{
	if (under_isr)
	{
		portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
		xSemaphoreGiveFromISR((SemaphoreHandle_t)sem, &xHigherPriorityTaskWoken);
		if(xHigherPriorityTaskWoken != pdFALSE )
		{
			// Force a context switch here.
			portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
		}
	}
	else
	{
		xSemaphoreGive((SemaphoreHandle_t)sem);
	}
}

// See the header file for the function documentation
uint32_t LLUI_DISPLAY_IMPL_getNewImageStrideInBytes(MICROUI_ImageFormat image_format, uint32_t image_width, uint32_t image_height, uint32_t default_stride)
{
	/*
	 * VGLite library requires a stride on 16 pixels: 64 bytes for 32bpp, 32 bytes
	 * for 16bpp and 16 bytes for 8bpp. Exception 8 bytes for <= 4bpp.
	 */
	int bpp = display_utils_get_bpp(image_format);
	return bpp >= 8 ? ALIGN(image_width, 16) * (bpp / 8) : ALIGN(image_width, 8);
}

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------
