/*
 * C
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------
#include "LLVECTOR_FONT_impl.h"
#include "LLVGLITE_impl.h"

#include "math.h"
#include <stdbool.h>

#include "sni.h"

#include "ft2build.h"
#include FT_FREETYPE_H
#include FT_INTERNAL_OBJECTS_H
#include "ftvglite.h"

#include "display_vglite.h"
#include "display_task.h"

#include "trace_platform.h"

#include "LLUI_DISPLAY.h"

#include "fsl_debug_console.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

/*
 * Trace start macros
 *
 * param[in] event: LLVECTOR_FONT Event to trace
 */
#define __TRACE_START(event) \
		TRACE_PLATFORM_START_VOID(VECTOR_FONT, LLVECTOR_FONT_TRACE_ ## event)

/*
 * Trace start macros
 *
 * param[in] event: LLVECTOR_FONT Event to trace
 * param[in] v: LLVECTOR_FONT value to trace
 */
#define __TRACE_START_U32(event, v) \
		TRACE_PLATFORM_START_U32(VECTOR_FONT, LLVECTOR_FONT_TRACE_ ## event, v)

/*
 * Trace end macros
 *
 * param[in] event: LLVECTOR_FONT Event to trace
 */
#define __TRACE_END(event) \
		TRACE_PLATFORM_END_VOID(VECTOR_FONT, LLVECTOR_FONT_TRACE_ ## event);

/*
 * @brief Low Level VECTOR_FONT identifier to trace initialization
 */
#define LLVECTOR_FONT_TRACE_INIT               11

/*
 * @brief Low Level VECTOR_FONT identifier to trace font loading
 */
#define LLVECTOR_FONT_TRACE_LOAD_FONT          12

/*
 * @brief Low Level VECTOR_FONT identifier to trace string width computation
 */
#define LLVECTOR_FONT_TRACE_STRING_WIDTH       21

/*
 * @brief Low Level VECTOR_FONT identifier to trace gradient initializaiton
 */
#define LLVECTOR_FONT_TRACE_INIT_GRADIENT      31

/*
 * @brief Low Level VECTOR_FONT identifier to trace gradient closure
 */
#define LLVECTOR_FONT_TRACE_CLOSE_GRADIENT     32

/*
 * @brief Low Level VECTOR_FONT identifier to trace gradient translation
 */
#define LLVECTOR_FONT_TRACE_TRANSLATE_GRADIENT 33

/*
 * @brief Low Level VECTOR_FONT identifier to trace string drawing
 */
#define LLVECTOR_FONT_TRACE_DRAW_STRING        41

/*
 * @brief Low Level VECTOR_FONT identifier to trace string drawing on circle arc
 */
#define LLVECTOR_FONT_TRACE_DRAW_STRING_ARC    42

#define MEJ_PRINT(fmt, ...) PRINTF("%s: " fmt, __func__, ##__VA_ARGS__)
#define MEJ_PRINT_ON_ERROR(err, fmt, ...) if (0 != err) PRINTF("%s: " fmt, __func__, ##__VA_ARGS__)

#ifndef M_PI
#define M_PI 3.1415926f
#endif

// -----------------------------------------------------------------------------
// Types
// -----------------------------------------------------------------------------

typedef struct {
	void* data;
	uint32_t size;
} SNIX_resource;


// -----------------------------------------------------------------------------
// Extern functions
// -----------------------------------------------------------------------------

extern int32_t SNIX_get_resource(char* path, SNIX_resource* resource);
extern jint LLVGLITE_GRADIENT_OBJECT(create, );
extern jint LLVGLITE_GRADIENT_OBJECT(destroy, );

// -----------------------------------------------------------------------------
// Global Variables
// -----------------------------------------------------------------------------

static FT_Library library;

static FT_Renderer vglite_renderer;

// -----------------------------------------------------------------------------
// Internal function definitions
// -----------------------------------------------------------------------------

/*
 * @brief Computes the scale to apply to the font.
 *
 * @param[in] size: the font size
 * @param[in] face: the face of the font
 */
static inline float __get_scale(jfloat size, FT_Face face);

/*
 * @brief Sets renderer parameters.
 *
 * @param[in] gc: the graphic context
 * @param[in] matrix: the transformation matrix
 * @param[in] alpha: the opacity of the string (from 0 to 255)
 * @param[in] gradient: the gradient to apply
 */
static void __set_renderer(MICROUI_GraphicsContext* gc, vg_lite_matrix_t *matrix, jint alpha, vg_lite_linear_gradient_t* gradient);

/*
 * @brief Updates the angle to use for the next glyph when drawn on an arc.
 * When drawing on an arc, the glyph position is defined by its angle. We thus
 * convert the advance (distance to the next glyph) to an angle.
 *
 * @param[in] advance: the advance that should be used if the text was drawn on
 * a straight line
 * @param[in] radius: the radius of the circle
 * @param[in] glyph_height: the height of the glyph
 * @param[in] scale: the scale to render glyphs
 * @param[in] out: tells whether the circle outline should be used as baseline.
 * @param[in,out] angle: the position of the glyph to update
 */
static void __update_angle(float advance, int radius, short glyph_height, float scale, bool out, float *angle);

/**
 * @brief Initializes a transformation matrix with the given data.
 *
 * @param[in] data: the values to set to the matrix.
 * @param[in,out] matrix: the transformation matrix.
 */
static void __init_data(vg_lite_float_t* data, vg_lite_matrix_t * matrix);

/*
 * @brief Initializes a transformation matrix setting (x, y) as origin.
 * It is equivalent to a translation of the identity matrix.
 *
 * @param[in] x: the x coordinate of the origin.
 * @param[in] y: the y coordinate of the origin.
 * @param[in,out] matrix: the transformation matrix.
 */
static void __init_origin(vg_lite_float_t x, vg_lite_float_t y, vg_lite_matrix_t * matrix);

/*
 * @brief Translates the coordinate system.
 *
 * @param[in] x: the move along x axis to apply.
 * @param[in] y: the move along y axis to apply.
 * @param[in,out] matrix: the transformation matrix.
 */
static void __translate(vg_lite_float_t x, vg_lite_float_t y, vg_lite_matrix_t * matrix);

/*
 * @brief Scales the coordinate system.
 *
 * @param[in] scale_x: the ratio to apply to x axis.
 * @param[in] scale_y: the ratio to apply to y axis.
 * @param[in,out] matrix: the transformation matrix.
 */
static void __scale(vg_lite_float_t scale_x, vg_lite_float_t scale_y, vg_lite_matrix_t * matrix);

/*
 * @brief Rotates the coordinate system.
 *
 * @param[in] angle_radians: the number of radians to rotate around.
 * @param[in,out] matrix: the transformation matrix.
 */
static void __rotate(vg_lite_float_t angle_radians, vg_lite_matrix_t * matrix);

// -----------------------------------------------------------------------------
// FreeType API functions
// -----------------------------------------------------------------------------

void LLVECTOR_FONT_init(){
	FT_Error error;

	__TRACE_START(INIT);

	error = FT_Init_FreeType(&library);
	MEJ_PRINT_ON_ERROR(error, "error occurred during library initialization: %d\n", error);

	vglite_renderer = FT_RENDERER(FT_Get_Module(library, "vglite_renderer"));
	error = !vglite_renderer;
	MEJ_PRINT_ON_ERROR(error, "error occurred during font loading: %d\n", error);

	__TRACE_END(INIT);
}

jint LLVECTOR_FONT_load_font(char* font_name){
	int32_t error;
	void* font_data;
	int font_data_length;
	FT_Face face;
	SNIX_resource font_resource;

	__TRACE_START(LOAD_FONT);

	face = NULL;

	error = SNIX_get_resource(font_name, &font_resource);
	if (error != 0) {
		PRINTF("An error occurred while loading resource: %d\n", error);
		return 0;
	}

	font_data = font_resource.data;
	font_data_length = font_resource.size;

	error = FT_New_Memory_Face(library, font_data, font_data_length,  0, &face);
	if (error != 0) {
		PRINTF("An error occurred during font loading: %d\n", error);
		return 0;
	}

	// Use Unicode
	FT_Select_Charmap(face , ft_encoding_unicode);

	__TRACE_END(LOAD_FONT);

	return (jint) face;
}

jint LLVECTOR_FONT_init_gradient(jint start_x, jint start_y, jint end_x, jint end_y, jint* fractions, jint* colors){
	float length_x;
	float length_y;
	float scale;
	float angle;
	vg_lite_linear_gradient_t* gradient;

	__TRACE_START(INIT_GRADIENT);

	gradient = (vg_lite_linear_gradient_t* )LLVGLITE_GRADIENT_OBJECT(create);
	if (NULL == gradient)
	{
		return (jint)0;
	}

	int gradient_length = SNI_getArrayLength(colors);

	vg_lite_init_grad(gradient);

	// GPU displays ABGR instead of ARGB and vice versa
	gradient->image.format = VG_LITE_RGBA8888;

	for(int i = 0; i < gradient_length; i++){
		colors[i] = display_vglite_porter_duff_workaround_ARGB8888(colors[i]);
	}

	vg_lite_matrix_t* matrix = vg_lite_get_grad_matrix(gradient);
	length_x = end_x - start_x;
	length_y = end_y - start_y;
	scale = sqrtf(length_x * length_x + length_y * length_y) / 255.0f;
	angle = atanf((length_y) / (length_x));

	__init_origin(start_x, start_y, matrix);
	__scale(scale, 1.0f, matrix);
	__rotate(angle, matrix);

	// TODO Handle error
	vg_lite_set_grad(gradient, gradient_length, (uint32_t*) colors, (uint32_t*) fractions);
	vg_lite_update_grad(gradient);

	__TRACE_END(INIT_GRADIENT);

	return (jint) gradient;
}

void LLVECTOR_FONT_gradient_translate(vg_lite_linear_gradient_t* gradient, jint x, jint y){
	// The matrix does not need to be synchronized. At the end of a draw method,
	// the matrix data is loaded in he GPU registers.

	if (NULL == gradient)
	{
		return;
	}

	__TRACE_START(TRANSLATE_GRADIENT);

	vg_lite_matrix_t* matrix = vg_lite_get_grad_matrix(gradient);
	__translate(x, y, matrix);

	__TRACE_END(TRANSLATE_GRADIENT);
}

void LLVECTOR_FONT_close_gradient(vg_lite_linear_gradient_t* gradient){  

	if (NULL == gradient)
	{
		return;
	}

	__TRACE_START(CLOSE_GRADIENT);

	LLVGLITE_GRADIENT_OBJECT(destroy, (jint)gradient);

	__TRACE_END(CLOSE_GRADIENT);
}

jint LLVECTOR_FONT_char_width(FT_Face face, jfloat size, jchar character){
	float scale = __get_scale(size, face);

	FT_UInt glyph_index = FT_Get_Char_Index(face, character);

	// load glyph image into the slot (erase previous one)
	FT_Error error = FT_Load_Glyph(face, glyph_index, FT_LOAD_NO_SCALE); // Workaround for error 0x24: "invalid size handle"
	MEJ_PRINT_ON_ERROR(error, "error while loading glyph: %d\n", error);
	if(error){
		return 0; // ignore errors
	}

	return (jint)(scale * face->glyph->advance.x);
}

jint LLVECTOR_FONT_string_width(FT_Face face, jfloat size, jchar* text){
	float scale;
	int text_length;
	long unscaled_width;
	FT_UInt previous; // previous glyph index for kerning
	FT_Bool use_kerning;

	__TRACE_START(STRING_WIDTH);

	scale = __get_scale(size, face);

	text_length = SNI_getArrayLength(text);
	unscaled_width = 0;

	use_kerning = FT_HAS_KERNING(face);

	for(int text_index = 0;  text_index < text_length; text_index++){
		FT_UInt glyph_index = FT_Get_Char_Index(face, text[text_index]);

		if (use_kerning && previous && glyph_index){
			FT_Vector  delta;
			FT_Get_Kerning(face, previous, glyph_index, FT_KERNING_DEFAULT, &delta);

			unscaled_width += delta.x;
		}

		// load glyph image into the slot (erase previous one)
		FT_Error error = FT_Load_Glyph(face, glyph_index, FT_LOAD_NO_SCALE); // Workaround for error 0x24: "invalid size handle"
		MEJ_PRINT_ON_ERROR(error, "error while loading glyph: %d\n", error);
		if(error){
			continue; // ignore errors
		}

		unscaled_width += face->glyph->advance.x;
		previous = glyph_index;
	}

	__TRACE_END(STRING_WIDTH);
   
	// Workaround add an extra pixel to avoid cropping of last charactere in the string 
	return (jint)(scale * unscaled_width + 1);
}

jint LLVECTOR_FONT_get_baseline_position(FT_Face face, jfloat size){

	float scale;
	short glyph_height;
	short advance_y;

	scale = __get_scale(size, face);

	glyph_height = face->ascender - face->descender;
	advance_y = (int) (face->ascender * face->units_per_EM / glyph_height * scale);

	return(advance_y);

}

void LLVECTOR_FONT_draw_char(MICROUI_GraphicsContext* gc, FT_Face face, jfloat size, jchar character, jfloat* matrix_data, int alpha, vg_lite_linear_gradient_t* gradient){
	vg_lite_matrix_t matrix;
	float scale;

	if(!LLUI_DISPLAY_requestDrawing(gc, (SNI_callback)&(LLVECTOR_FONT_draw_char))){
		return ;
	}

	scale = __get_scale(size, face);

	__init_data(matrix_data, &matrix);
	__translate(0, LLVECTOR_FONT_get_baseline_position(face,size), &matrix);
	__scale(scale, scale, &matrix);
	__set_renderer(gc, &matrix, alpha, gradient);

	// Manage clipping with scissor
	vg_lite_enable_scissor();
	vg_lite_set_scissor(gc->clip_x1, gc->clip_y1, gc->clip_x2, gc->clip_y2);

	FT_UInt glyph_index = FT_Get_Char_Index(face, character);

	// load glyph image into the slot (erase previous one)
	FT_Error error = FT_Load_Glyph(face, glyph_index, FT_LOAD_NO_SCALE); // Workaround for error 0x24: "invalid size handle"
	MEJ_PRINT_ON_ERROR(error, "error while loading glyph: %d\n", error);
	if(error){
		return; // ignore errors
	}

	// convert to an anti-aliased bitmap
	error = FT_Render_Glyph(face->glyph, FT_RENDER_MODE_NORMAL);
	MEJ_PRINT_ON_ERROR(error, "error while rendering glyph: %d\n", error);
	if(error){
		return;
	}

	// Disable scissors
	vg_lite_disable_scissor();

	display_vglite_start_operation();
	vg_lite_flush();

	// wakeup display task to force to wait the end of VG operation.
	// graphical engine will wait the end of VG operation until next call to LLUI_DISPLAY_notifyAsynchronousDrawingEnd();
	display_task_wakeup(false);

	LLUI_DISPLAY_setDrawingStatus(DRAWING_RUNNING);
}

void LLVECTOR_FONT_draw_string(MICROUI_GraphicsContext* gc, FT_Face face, jfloat size, jchar* text, jint offset, jint length, jint x, jint y, jint alpha, vg_lite_linear_gradient_t* gradient){
	float scale;
	short glyph_height;
	short advance_y;
	long advance_x;
	vg_lite_matrix_t matrix;
	FT_UInt previous; // previous glyph index for kerning
	FT_Bool use_kerning;

	if (!LLUI_DISPLAY_requestDrawing(gc, (SNI_callback)&(LLVECTOR_FONT_draw_string))){
		return ;
	}

	__TRACE_START_U32(DRAW_STRING, length);

	scale = __get_scale(size, face);
	__init_origin(x, y, &matrix);
	__scale(scale, scale, &matrix);

	glyph_height = face->ascender - face->descender;
	advance_y = face->ascender * face->units_per_EM / glyph_height ;

	__translate(0, advance_y, &matrix);

	__set_renderer(gc, &matrix, alpha, gradient);

	// Manage clipping with scissor
	vg_lite_enable_scissor();
	vg_lite_set_scissor(gc->clip_x1, gc->clip_y1, gc->clip_x2, gc->clip_y2);

	advance_x = 0;

	use_kerning = FT_HAS_KERNING(face);

	for(int text_index = offset;  text_index <  offset + length; text_index++){
		FT_UInt glyph_index = FT_Get_Char_Index(face, text[text_index]);

		if (use_kerning && previous && glyph_index){
			FT_Vector  delta;
			FT_Get_Kerning(face, previous, glyph_index, FT_KERNING_DEFAULT, &delta);

			advance_x += delta.x;
		}

		// load glyph image into the slot (erase previous one)
		FT_Error error = FT_Load_Glyph(face, glyph_index, FT_LOAD_NO_SCALE); // Workaround for error 0x24: "invalid size handle"
		MEJ_PRINT_ON_ERROR(error, "error while loading glyph: %d\n", error);
		if(error){
			continue; // ignore errors
		}

		__translate(advance_x, 0, &matrix);

		// convert to an anti-aliased bitmap
		error = FT_Render_Glyph(face->glyph, FT_RENDER_MODE_NORMAL);
		MEJ_PRINT_ON_ERROR(error, "error while rendering glyph: %d\n", error);
		if(error){
			continue;
		}

		advance_x = face->glyph->advance.x;

		previous = glyph_index;
	}

	// Disable scissors
	vg_lite_disable_scissor();

	display_vglite_start_operation();
	vg_lite_flush();

	// wakeup display task to force to wait the end of VG operation.
	// graphical engine will wait the end of VG operation until next call to LLUI_DISPLAY_notifyAsynchronousDrawingEnd();
	display_task_wakeup(false);

	LLUI_DISPLAY_setDrawingStatus(DRAWING_RUNNING);

	__TRACE_END(DRAW_STRING);
}

void LLVECTOR_FONT_draw_string_arc(MICROUI_GraphicsContext* gc, FT_Face face, jfloat size, jchar* text, jint offset, jint length, jint x, jint y, jint diameter, jfloat angle_degrees, jint alpha, vg_lite_linear_gradient_t* gradient, bool out){
	float scale;
	int radius;
	int center_x;
	int center_y;
	short glyph_height;
	short advance_y;
	long advance_x;

	vg_lite_float_t angle_radians;
	vg_lite_matrix_t matrix;
	FT_UInt previous; // previous glyph index for kerning
	FT_Bool use_kerning;


	if (!LLUI_DISPLAY_requestDrawing(gc, (SNI_callback)&(LLVECTOR_FONT_draw_string_arc))){
		return ;
	}

	__TRACE_START_U32(DRAW_STRING_ARC, length);

	scale = __get_scale(size, face);

	radius = diameter / 2;
	center_x = x + radius;
	center_y = y + radius;

	// We add the angle because in transformations we put the first character
	// at 90°
	angle_radians = ((out ? -90 : 90) - angle_degrees) / 180.0f * M_PI;

	__set_renderer(gc, &matrix, alpha, gradient);

	// Manage clipping with scissor
	vg_lite_enable_scissor();
	vg_lite_set_scissor(gc->clip_x1, gc->clip_y1, gc->clip_x2, gc->clip_y2);

	use_kerning = FT_HAS_KERNING(face);

	glyph_height = face->ascender - face->descender;
	advance_y = out ? face->descender * face->units_per_EM / glyph_height
			        : face->ascender * face->units_per_EM / glyph_height;

	advance_x = 0;

	for(int text_index = offset;  text_index <  offset + length; text_index++){
		FT_UInt glyph_index = FT_Get_Char_Index(face, text[text_index]);

		if (use_kerning && previous && glyph_index){
			FT_Vector  delta;
			FT_Get_Kerning(face, previous, glyph_index, FT_KERNING_DEFAULT, &delta);

			advance_x += delta.x;
		}

		__update_angle(advance_x, radius, glyph_height, scale, out, &angle_radians);

		// load glyph image into the slot (erase previous one)
		FT_Error error = FT_Load_Glyph(face, glyph_index, FT_LOAD_NO_SCALE); // Workaround for error 0x24: "invalid size handle"
		MEJ_PRINT_ON_ERROR(error, "error while loading glyph: %d\n", error);
		if(error){
			continue; // ignore errors
		}

		// set the origin to the center of the circle
		__init_origin(center_x, center_y, &matrix);

		// set the angle of the glyph
		__rotate(angle_radians, &matrix);

		// move the glyph to the arc
		__translate(0, out ? radius : -radius, &matrix);

		// a scale is applied to use font metrics after this line
		__scale(scale, scale, &matrix);

		// move the baseline to the top of the glyph
		__translate(0, advance_y, &matrix);

		// adjust baseline angle to the tangent
		float adjust_angle = atanf(scale * face->glyph->advance.x / diameter);
		__rotate(out ? -adjust_angle : adjust_angle, &matrix);

		// convert to an anti-aliased bitmap
		error = FT_Render_Glyph(face->glyph, FT_RENDER_MODE_NORMAL);
		MEJ_PRINT_ON_ERROR(error, "error while rendering glyph: %d\n", error);
		if(error){
			continue;
		}

		advance_x = face->glyph->advance.x;

		previous = glyph_index;
	}

	// Disable scissors
	vg_lite_disable_scissor();

	display_vglite_start_operation();

	// ask to start VG operation
	vg_lite_flush();

	// wakeup display task to force to wait the end of VG operation.
	// graphical engine will wait the end of VG operation until next call to LLUI_DISPLAY_notifyAsynchronousDrawingEnd();
	display_task_wakeup(false);

	LLUI_DISPLAY_setDrawingStatus(DRAWING_RUNNING);

	__TRACE_END(DRAW_STRING_ARC);
}

// -----------------------------------------------------------------------------
// Internal functions
// -----------------------------------------------------------------------------

static inline float __get_scale(jfloat size, FT_Face face){
	return size / face->units_per_EM;
}

static void __set_renderer(MICROUI_GraphicsContext* gc, vg_lite_matrix_t* matrix, jint alpha, vg_lite_linear_gradient_t* gradient){
	int color;
	vg_lite_buffer_t* buffer;
	FT_Parameter vglite_params[8];

	buffer = display_vglite_configure_destination(gc);
	color = (gc->foreground_color & 0x00FFFFFF) + (alpha << 24);

	vglite_params[0].tag = FT_PARAM_TAG_VGLITE_BUFFER_ADDR;
	vglite_params[0].data = buffer;

	vglite_params[1].tag = FT_PARAM_TAG_VGLITE_MATRIX;
	vglite_params[1].data = matrix;

	vglite_params[2].tag = FT_PARAM_TAG_VGLITE_QUALITY;
	vglite_params[2].data = (void *) VG_LITE_HIGH;

	vglite_params[3].tag = FT_PARAM_TAG_VGLITE_FORMAT;
	vglite_params[3].data = (void *) VG_LITE_S16;

	vglite_params[4].tag = FT_PARAM_TAG_VGLITE_FILL_RULE;
	vglite_params[4].data = (void *) VG_LITE_FILL_EVEN_ODD;

	vglite_params[5].tag = FT_PARAM_TAG_VGLITE_BLEND;
	vglite_params[5].data = (void *) VG_LITE_BLEND_SRC_OVER;

	vglite_params[6].tag = FT_PARAM_TAG_VGLITE_COLOR;
	vglite_params[6].data = (void *) color;

	vglite_params[7].tag = FT_PARAM_TAG_VGLITE_GRADIENT;
	vglite_params[7].data = (void *) gradient;

	FT_Set_Renderer(library, vglite_renderer, 8, &vglite_params[0]);
}

static void __update_angle(float advance, int radius, short glyph_height, float scale, bool out, float* angle_radians){
	float advance_scaled;

	advance_scaled = advance * scale;

	if(out){
		*angle_radians -= (advance_scaled / (radius - (glyph_height * scale)));
	}
	else {
		*angle_radians += (advance_scaled / (radius - (glyph_height * scale)));
	}
}

static void __init_data(vg_lite_float_t* data, vg_lite_matrix_t * matrix){
	matrix->m[0][0] = data[0];
	matrix->m[0][1] = data[1];
	matrix->m[0][2] = data[2];
	matrix->m[1][0] = data[3];
	matrix->m[1][1] = data[4];
	matrix->m[1][2] = data[5];
	matrix->m[2][0] = data[6];
	matrix->m[2][1] = data[7];
	matrix->m[2][2] = data[8];
	//	memcpy(data, matrix->m, 9);
}

static void __init_origin(vg_lite_float_t x, vg_lite_float_t y, vg_lite_matrix_t * matrix){
	matrix->m[0][0] = 1.0f;
	matrix->m[0][1] = 0.0f;
	matrix->m[0][2] = x;
	matrix->m[1][0] = 0.0f;
	matrix->m[1][1] = 1.0f;
	matrix->m[1][2] = y;
	matrix->m[2][0] = 0.0f;
	matrix->m[2][1] = 0.0f;
	matrix->m[2][2] = 1.0f;
}

static void __translate(vg_lite_float_t x, vg_lite_float_t y, vg_lite_matrix_t * matrix){
	matrix->m[0][2] = matrix->m[0][0] * x + matrix->m[0][1] * y + matrix->m[0][2];
	matrix->m[1][2] = matrix->m[1][0] * x + matrix->m[1][1] * y + matrix->m[1][2];
	matrix->m[2][2] = matrix->m[2][0] * x + matrix->m[2][1] * y + matrix->m[2][2];
}

static void __scale(vg_lite_float_t x, vg_lite_float_t y, vg_lite_matrix_t * matrix){
	matrix->m[0][0] *= x;
	matrix->m[0][1] *= y;
	matrix->m[1][0] *= x;
	matrix->m[1][1] *= y;
	matrix->m[2][0] *= x;
	matrix->m[2][1] *= y;
}

static void __rotate(vg_lite_float_t angle_radians, vg_lite_matrix_t * matrix){
	vg_lite_float_t cos_angle = cosf(angle_radians);
	vg_lite_float_t sin_angle = sinf(angle_radians);

	vg_lite_float_t tmp;

	tmp = cos_angle * matrix->m[0][0] + sin_angle * matrix->m[0][1];
	matrix->m[0][1] = cos_angle * matrix->m[0][1] - sin_angle * matrix->m[0][0];
	matrix->m[0][0] = tmp;

	tmp = cos_angle * matrix->m[1][0] + sin_angle * matrix->m[1][1];
	matrix->m[1][1] = cos_angle * matrix->m[1][1] - sin_angle * matrix->m[1][0];
	matrix->m[1][0] = tmp;

	tmp = cos_angle * matrix->m[2][0] + sin_angle * matrix->m[2][1];
	matrix->m[2][1] = cos_angle * matrix->m[2][1] - sin_angle * matrix->m[2][0];
	matrix->m[2][0] = tmp;
}

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------
