/*
 * C
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

#include <math.h>

// RTOS
#include "FreeRTOS.h"
#include "semphr.h"

#include "event_generator.h"

#include "vg_lite.h"

#include "sni.h"

#include "LLUI_DISPLAY.h"

#include "display_task.h"
#include "display_framebuffer.h"
#include "display_vglite.h"

#include "mej_log.h"

#include "LLVGLITE_impl.h"
#include "LLMJVM.h"

#include "trace_vglite.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

/*
 * Trace start macros
 *
 * param[in] event: LLVGLITE Event to trace
 */
#define __TRACE_START(event) \
		TRACE_PLATFORM_START_VOID(LLVGLITE, LLVGLITE_TRACE_ ## event)

/*
 * Trace end macros
 *
 * param[in] event: LLVGLITE Event to trace
 */
#define __TRACE_END(event) \
		TRACE_PLATFORM_END_VOID(LLVGLITE, LLVGLITE_TRACE_ ## event);

/*
 * @brief Low Level VGLite API DRAW (draw path) identifier
 * This event is expected to be used to trace a call to the java API
 * vg_lite_draw
 */
#define LLVGLITE_TRACE_DRAW (10 + TRACE_VGLITE_OP_DRAW)

/*
 * @brief VLow Level VGLite API DRAW_GRAD (draw path with gradient) identifier
 * This event is expected to be used to trace a call to the java API
 * vg_lite_draw_grad
 */
#define LLVGLITE_TRACE_DRAW_GRAD (10 + TRACE_VGLITE_OP_DRAW_GRAD)

/*
 * @brief VLow Level VGLite API MATRIX_IDENTITY (set matrix to identity)
 * identifier
 * This event is expected to be used to trace a call to the java API
 * vg_lite_matrix_identity
 */
#define LLVGLITE_TRACE_MATRIX_IDENTITY 21

/*
 * @brief VLow Level VGLite API MATRIX_TRANSLATE (update matrix for translation)
 * identifier
 * This event is expected to be used to trace a call to the java API
 * vg_lite_matrix_translate
 */
#define LLVGLITE_TRACE_MATRIX_TRANSLATE 22

/*
 * @brief VLow Level VGLite API MATRIX_TRANSLATE (update matrix for rotation)
 * identifier
 * This event is expected to be used to trace a call to the java API
 * vg_lite_matrix_rotate
 */
#define LLVGLITE_TRACE_MATRIX_ROTATE 23

/*
 * @brief VLow Level VGLite API MATRIX_TRANSLATE (update matrix for scale)
 * identifier
 * This event is expected to be used to trace a call to the java API
 * vg_lite_matrix_scale
 */
#define LLVGLITE_TRACE_MATRIX_SCALE 24

/*
 * @brief VLow Level VGLite API MATRIX_COPY (copy matrix parameters from an other matrix)
 * identifier
 * This event is expected to be used to trace a call to the java API
 * vg_lite_matrix_copy
 */
#define LLVGLITE_TRACE_MATRIX_COPY 25


// -----------------------------------------------------------------------------
// Internal function definitions
// -----------------------------------------------------------------------------

/*
 * @brief Close matrix
 */
static void _llvglite_close_matrix(void* resource);

/*
 * @brief Close gradient
 */
static void _llvglite_close_gradient(void* resource);

// -----------------------------------------------------------------------------
// Java APIs
// -----------------------------------------------------------------------------

/*
 * @brief See `vg_lite_identity` in `vg_lite.h`
 */
jint LLVGLITE_MATRIX_NATIVE(identity, jint matrix_handle) {
	vg_lite_matrix_t *matrix = (vg_lite_matrix_t *) matrix_handle;

	__TRACE_START(MATRIX_IDENTITY);

	vg_lite_identity(matrix);

	__TRACE_END(MATRIX_IDENTITY);

	return 0;
}

/*
 * @brief See `vg_lite_translate` in `vg_lite.h`
 */
jint LLVGLITE_MATRIX_NATIVE(translate, jint matrix_handle, jfloat x, jfloat y) {
	vg_lite_matrix_t *matrix = (vg_lite_matrix_t *) matrix_handle;

	__TRACE_START(MATRIX_TRANSLATE);

	vg_lite_translate(x, y, matrix);

	__TRACE_END(MATRIX_TRANSLATE);

	return 0;
}

/*
 * @brief See `vg_lite_rotate` in `vg_lite.h`
 */
jint LLVGLITE_MATRIX_NATIVE(rotate, jint matrix_handle, jfloat angle) {
	vg_lite_matrix_t *matrix = (vg_lite_matrix_t *) matrix_handle;

	__TRACE_START(MATRIX_ROTATE);

	vg_lite_rotate(angle, matrix);

	__TRACE_END(MATRIX_ROTATE);

	return 0;
}

/*
 * @brief See `vg_lite_scale` in `vg_lite.h`
 */
jint LLVGLITE_MATRIX_NATIVE(scale, jint matrix_handle, jfloat x, jfloat y) {
	vg_lite_matrix_t *matrix = (vg_lite_matrix_t *) matrix_handle;

	__TRACE_START(MATRIX_SCALE);

	vg_lite_scale(x, y, matrix);

	__TRACE_END(MATRIX_SCALE);

	return 0;
}

/*!
 @abstract Copy a matrix

 @discussion
 Copy matrix parameters from a matrix to an other matrix.

 @param src_matrix_handle
 Pointer to a <code>vg_lite_matrix_t</code> structure used as source matrix.

 @param dest_matrix_handle
 Pointer to a <code>vg_lite_matrix_t</code> structure used as destination matrix.
 */
jint LLVGLITE_MATRIX_NATIVE(copy, jint src_matrix_handle, jint dest_matrix_handle) {
	vg_lite_matrix_t *src_matrix = (vg_lite_matrix_t *) src_matrix_handle;
	vg_lite_matrix_t *dest_matrix = (vg_lite_matrix_t *) dest_matrix_handle;

	__TRACE_START(MATRIX_COPY);
	memcpy(dest_matrix, src_matrix, sizeof(vg_lite_matrix_t));

	__TRACE_END(MATRIX_COPY);

	return 0;
}

/*
 * @brief Destroys a matrix object.
 *
 * @param[in] matrix_handle: Handle on the matrix to destroy.
 */
jint LLVGLITE_MATRIX_OBJECT(destroy, jint matrix_handle) {

	int32_t res = SNI_unregisterResource((void*)matrix_handle, &_llvglite_close_matrix);
	assert(res == SNI_OK);
	_llvglite_close_matrix((void*)matrix_handle);
	return 0;
}

/*
 * @brief Creates a matrix object.
 *
 * @return: Handle on the created matrix.
 */
jint LLVGLITE_MATRIX_OBJECT(create, ) {

	vg_lite_matrix_t *matrix = malloc(sizeof(vg_lite_matrix_t));
	if (NULL != matrix)
	{
		vg_lite_identity(matrix);
		int32_t res = SNI_registerResource((void*)matrix, &_llvglite_close_matrix, NULL);
		assert(res == SNI_OK);
	}
	return (jint) matrix;
}

/*
 * @brief Draws a VGLite path.
 *
 * @param[in] gc: MicroUI graphics context.
 * @param[in] path_left: Horizontal coordinate of the left side of the bouding box.
 * @param[in] path_top: Vertical coordinate of top side the bouding box.
 * @param[in] path_right: Horizontal coordinate of the right side bouding box.
 * @param[in] path_bottom: Vertical coordinate of bottom side the bouding box.
 * @param[in] path_data: Buffer containing the path data (VGLite instructions).
 * @param[in] path_length: Number of bytes in the path data buffer.
 * @param[in] path_quality: Quality hint for the path (see: `vg_lite.h`).
 * @param[in] path_format: Coordinate format (see: `vg_lite.h`).
 * @param[in] fill_rule: Specified fill rule for the path (see: `vg_lite.h`).
 * @param[in] matrix_handle: Handle over a matrix object.
 * @param[in] blend: The blending mode to be applied to each drawn pixel (see: `vg_lite.h`).
 * @param[in] color: The color applied to each pixel drawn by the path (see: `vg_lite.h`).
 *
 * See `vg_lite_draw` in `vg_lite.h`
 */
jint LLVGLITE_NATIVE(draw_1path,
		MICROUI_GraphicsContext* gc,
		jfloat path_left, jfloat path_top, jfloat path_right, jfloat path_bottom,
		jbyte path_data[], jint path_offset, jint path_length,
		jint path_quality, jint path_format, jint fill_rule,
		jint matrix_handle, jint blend, jint color) {

	if (!LLUI_DISPLAY_requestDrawing(gc, (SNI_callback)&(LLVGLITE_NATIVE_NAME(draw_1path))))
	{
		return 0;
	}

	__TRACE_START(DRAW);

	vg_lite_matrix_t *matrix = (vg_lite_matrix_t *) matrix_handle;

	vg_lite_path_t path = {
			{path_left, path_top, // left,top
					path_right, path_bottom}, // right,bottom
					(vg_lite_quality_t) path_quality, // quality
					(vg_lite_format_t) path_format, // -128 to 127 coordinate range
					{0}, // uploaded
					path_length, // path length
					path_data + path_offset, // path data
					1
	};

	vg_lite_buffer_t* blit_context = display_vglite_configure_destination(gc);

	// Manage clipping with scissor
	vg_lite_enable_scissor();
	vg_lite_set_scissor(gc->clip_x1, gc->clip_y1, gc->clip_x2, gc->clip_y2);

	TRACE_VGLITE_INIT_GPU_START(DRAW);

	color = display_vglite_porter_duff_workaround_ARGB8888(color);

	vg_lite_error_t ret = vg_lite_draw(blit_context, &path, (vg_lite_fill_t) fill_rule, matrix, (vg_lite_blend_t) blend, color);

	TRACE_VGLITE_INIT_GPU_END();

	// Disable scissors
	vg_lite_disable_scissor();

	if (ret == VG_LITE_SUCCESS) {
		display_vglite_start_operation();
		vg_lite_flush();
		display_task_wakeup(false);
		LLUI_DISPLAY_setDrawingStatus(DRAWING_RUNNING);
	} else {
		MEJ_LOG_ERROR("vg_lite_draw failed with error: %d\n", ret);
		LLUI_DISPLAY_setDrawingStatus(DRAWING_DONE);
		display_vglite_cancel_operation();
	}

	__TRACE_END(DRAW);

	return (jint) ret;
}

/*
 * @brief Creates a gradient object.
 *
 * @return: Handle on the created gradient.
 */
jint LLVGLITE_GRADIENT_OBJECT(create, ) {
	vg_lite_linear_gradient_t *gradient = malloc(sizeof(vg_lite_linear_gradient_t));

	if (NULL != gradient)
	{
		memset(gradient, 0, sizeof(*gradient));
		int32_t res = SNI_registerResource((void*)gradient, &_llvglite_close_gradient, NULL);
		assert(res == SNI_OK);

		// initialize gradient's matrix
		vg_lite_matrix_t *  matrix = vg_lite_get_grad_matrix(gradient);
		vg_lite_identity(matrix);
	}

	return (jint) gradient;
}

/*
 * @brief Destroys a gradient object.
 *
 * @param[in] gradient_handle: Handle on the gradient.
 */
jint LLVGLITE_GRADIENT_OBJECT(destroy, jint gradient_handle) {

	int32_t res = SNI_unregisterResource((void*)gradient_handle, &_llvglite_close_gradient);
	assert(res == SNI_OK);
	_llvglite_close_gradient((void*)gradient_handle);
	return 0;
}

/*
 * @brief Returns the matrix handle of a gradient
 *
 * @param[in] gradient_handle: Handle on the gradient.
 */
jint LLVGLITE_GRADIENT_OBJECT(get_1matrix_1handle, jint gradient_handle) {
	vg_lite_linear_gradient_t *gradient = (vg_lite_linear_gradient_t *) gradient_handle;

	return (jint) vg_lite_get_grad_matrix(gradient);
}

jint LLVGLITE_NATIVE(init_1grad, jint gradient_handle) {
	vg_lite_linear_gradient_t * gradient = (vg_lite_linear_gradient_t *) gradient_handle;
	vg_lite_error_t ret;

	// clear previous conf (vg_lite_init_grad allocates a buffer in VGLite
	// buffer, we must free it before allocating another buffer)
	// also see _llvglite_close_gradient
	vg_lite_clear_grad(gradient);

	// this call allocates in VGLite buffer
	ret = vg_lite_init_grad(gradient);

	// GPU displays ABGR instead of ARGB and vice versa
	gradient->image.format = VG_LITE_RGBA8888;

	return (jint) ret;
}

jint LLVGLITE_NATIVE(set_1grad,
		jint gradient_handle,
		jint count,
		jint colors[],
		jint stops[]) {
	uint32_t colors_temp[VLC_MAX_GRAD];
	vg_lite_linear_gradient_t *gradient = (vg_lite_linear_gradient_t *) gradient_handle;

	if (count > VLC_MAX_GRAD) {
		return VG_LITE_INVALID_ARGUMENT;
	}

	for (int i = 0; i < count; i++) {
		colors_temp[i] = display_vglite_porter_duff_workaround_ARGB8888(colors[i]);
	}

	return (jint) vg_lite_set_grad(gradient,
			count,
			(uint32_t *) colors_temp,
			(uint32_t *) stops);
}

jint LLVGLITE_NATIVE(update_1grad, jint gradient_handle) {
	vg_lite_linear_gradient_t *gradient = (vg_lite_linear_gradient_t *) gradient_handle;

	return (jint) vg_lite_update_grad(gradient);
}

jint LLVGLITE_NATIVE(draw_1gradient,
		MICROUI_GraphicsContext* gc,
		jfloat path_left, jfloat path_top, jfloat path_right, jfloat path_bottom,
		jbyte path_data[], jint path_offset, jint path_length,
		jint path_quality, jint path_format, jint fill_rule,
		jint matrix_handle, jint gradient_handle, jint blend) {

	if (!LLUI_DISPLAY_requestDrawing(gc, (SNI_callback)&(LLVGLITE_NATIVE_NAME(draw_1gradient))))
	{
		return 0;
	}

	__TRACE_START(DRAW_GRAD);

	vg_lite_matrix_t *matrix = (vg_lite_matrix_t *) matrix_handle;
	vg_lite_linear_gradient_t *gradient = (vg_lite_linear_gradient_t *) gradient_handle;

	vg_lite_path_t path = {
			{path_left, path_top, // left,top
					path_right, path_bottom}, // right,bottom
					(vg_lite_quality_t) path_quality, // quality
					(vg_lite_format_t) path_format, // -128 to 127 coordinate range
					{0}, // uploaded
					path_length, // path length
					path_data + path_offset, // path data
					1
	};

	vg_lite_buffer_t* blit_context = display_vglite_configure_destination(gc);

	// Manage clipping with scissor
	vg_lite_enable_scissor();
	vg_lite_set_scissor(gc->clip_x1, gc->clip_y1, gc->clip_x2, gc->clip_y2);

	TRACE_VGLITE_INIT_GPU_START(DRAW_GRAD);

	vg_lite_error_t ret = vg_lite_draw_gradient(
			blit_context,
			&path,
			(vg_lite_fill_t) fill_rule,
			matrix, gradient,
			(vg_lite_blend_t) blend);

	TRACE_VGLITE_INIT_GPU_END();

	// Disable scissors
	vg_lite_disable_scissor();

	if (ret == VG_LITE_SUCCESS) {
		display_vglite_start_operation();
		vg_lite_flush();
		display_task_wakeup(false);
		LLUI_DISPLAY_setDrawingStatus(DRAWING_RUNNING);
	} else {
		MEJ_LOG_ERROR("vg_lite_draw failed with error: %d\n", ret);
		LLUI_DISPLAY_setDrawingStatus(DRAWING_DONE);
		display_vglite_cancel_operation();
	}

	__TRACE_END(DRAW_GRAD);

	return (jint) ret;
}

// -----------------------------------------------------------------------------
// Internal functions
// -----------------------------------------------------------------------------

static void _llvglite_close_matrix(void* resource)
{
	vg_lite_matrix_t *matrix = (vg_lite_matrix_t *) resource;
	free(matrix);
}

static void _llvglite_close_gradient(void* resource)
{
	vg_lite_linear_gradient_t *gradient = (vg_lite_linear_gradient_t *) resource;

	// vg_lite_init_grad allocates in VGLite buffer. No error even if init_1grad
	// is never called because vg_lite_clear_grad checks the allocation
	vg_lite_clear_grad(gradient);

	// free handle
	free(gradient);
}


// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------

