/*
 * C
 *
 * Copyright 2019-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

// -----------------------------------------------------------------------------
// Includes
// -----------------------------------------------------------------------------

// BSP
#include "fsl_debug_console.h"
#include "fsl_dma.h"
#include "fsl_fbdev.h"

// RTOS
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"

// Project
#include "mej_debug.h"

#include "display_support.h"

#include "vglite_support.h"
#include "vglite_window.h"
#include "vg_lite.h"

#include "board.h"
#include "pin_mux.h"

#include "LLUI_DISPLAY.h"

#include "display_task.h"
#include "display_dma.h"
#include "display_vglite.h"

#include "display_overlay.h"

#include "event_generator.h"

#include "monitor.h"

#include "ui_drawing_soft.h"
#include "framerate.h"

#include "mej_log.h"

#include "fsl_dc_fb_dsi_cmd.h"

// -----------------------------------------------------------------------------
// Macros and Defines
// -----------------------------------------------------------------------------

/*
 * @brief task specific constants
 */
#define OVERLAY_STACK_SIZE			( 4*1024 )
#define OVERLAY_TASK_PRIORITY		( 11 )
#define OVERLAY_TASK_STACK_SIZE		OVERLAY_STACK_SIZE/4
#define OVERLAY_TASK_SLEEP_TICKS	( MONITOR_SCHEDULE_TIME / portTICK_PERIOD_MS )

/*
 * @brief Horizontal coordinate of the "CPU" field
 */
#define CPU_X (150)

/*
 * @brief Vertical coordinate of the "CPU" field
 */
#define CPU_Y (5)

/*
 * @brief Horizontal coordinate of the "FPS" field
 */
#define FPS_X (150)
/*
 * @brief Vertical coordinate of the "FPS" field
 */
#define FPS_Y (25)

/*
 * @brief Indicates if the module is initialized
 */
#define IS_READY() ((__first_digit_x) > 0)

// -----------------------------------------------------------------------------
// Types
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// Global Variables
// -----------------------------------------------------------------------------

/*
 * @brief Graphics Context of the "CPU" MicroUi buffer
 */
static MICROUI_GraphicsContext* __gc_cpu;

/*
 * @brief Graphics Context of the "FPS" MicroUi buffer
 */
static MICROUI_GraphicsContext* __gc_fps;

/*
 * @brief MicroUI image containing the digits vertically (from 0 to 9)
 * The font used for generating the image
 * should be monospaced for a better rendering
 */
static MICROUI_Image* __digits;

/*
 * @brief Horizontal position of the first digit
 */
static jint __first_digit_x = 0; // 0 means "not initialized"

/*
 * @brief Width of the digits
 */
static jint __digit_width;

/*
 * @brief Height of the digits
 */
static jint __digit_height;

/*
 * @brief Back color to use
 */
static jint __back_color;

/*
 * @brief Flag to display the overlay
 */
static bool __display_overlay = false;

// -----------------------------------------------------------------------------
// Internal functions definition
// -----------------------------------------------------------------------------

/*
 * @brief Flushes an image to the frameboffer
 *
 * @param[in] buffer: pointer to the framebuffer
 * @param[in] gc: pointer to the graphics context of the image to flush
 * @param[in] x: horizontal coordinate where to flush the image
 * @param[in] y: vertical coordinate where to flush the image
 * @return 0 on success, -1 otherwise
 */
static int _flush_image(vg_lite_buffer_t *buffer, MICROUI_GraphicsContext* gc, uint32_t x, uint32_t y);

/*
 * @brief Clear an area of the overlay
 *
 * @param[in] dst: pointer to the VGLite buffer containing the area to clear
 * @param[in] x: coordinate of the left bound of the area to clear
 * @param[in] y: coordinate of the top bound of the area to clear
 * @param[in] w: width of the area to clear
 * @param[in] h: height of the area to clear
 * @param[in] color: color to fill the area with
 */
static void _clear(vg_lite_buffer_t *dst, int x, int y, int w, int h, vg_lite_color_t color);

/*
 * @brief Draws a number in an image
 *
 * @param[in] gc: pointer to the grapgics context of the image where to draw the number
 * @param[in] number: the number to draw
 */
static void _draw_number(MICROUI_GraphicsContext* gc, uint32_t number);

/*
 * @brief Draws a digit in an image
 *
 * @param[in] gc: pointer to the graphics context of the image where to draw the number
 * @param[in] number: the digit to draw
 * @param[in] x: horizontal coordinate of the digit to draw
 * @param[in] draw_zero: true to draw the zero digit, false to skip the zero digit
 * @return horizontal coordinate of the next digit
 */
static uint32_t _draw_digit(MICROUI_GraphicsContext* gc, uint32_t digit, uint32_t x, bool draw_zero);

/*
 * @brief FreeRTOS task that regularly update the overlay
 */
static void _overlay_task(void *params);

/*
 * @brief Callback that toggle the overlay
 */
static void __display_overlay_toggle_callback(void *params);


// -----------------------------------------------------------------------------
// Java functions
// -----------------------------------------------------------------------------

/*
 * @brief Java interface:
 * package: com.micrej.display.utils
 * class: Overlay
 * method: prepareOverlay
 *
 * @param[in] cpu: GraphicsContext of the "CPU" image.
 * @param[in] fps: GraphicsContext of the "FPS" image.
 * @param[in] digits: MicroUI image containing the digits.
 * @param[in] backColor: Color to use for the background.
 * @param[in] firstDigitX: Horizontal position of the first digit.
 * @param[in] nWidth: Width of the digits.
 * @param[in] nHeight: Height of the digits.
 */
void Java_com_microej_display_utils_NOverlay_prepareOverlay(MICROUI_GraphicsContext* cpu, MICROUI_GraphicsContext* fps, MICROUI_Image* digits, jint backColor, jint firstDigitX, jint nWidth, jint nHeight)
{
	__gc_cpu = cpu;
	__gc_fps = fps;
	__digits = digits;
	__digit_width = nWidth;
	__digit_height = nHeight;
	__back_color = backColor;
	__first_digit_x = firstDigitX;

	EVENT_GENERATOR_register_special_cb(__display_overlay_toggle_callback, 1);
	display_task_register_flush_callback(display_overlay_flush);

	BaseType_t xReturn = xTaskCreate( _overlay_task, "OVERLAY", OVERLAY_TASK_STACK_SIZE, NULL, OVERLAY_TASK_PRIORITY, NULL );
}

// -----------------------------------------------------------------------------
// Public functions
// -----------------------------------------------------------------------------

// See the header file for the function documentation
void display_overlay_update_cpu(uint32_t percent)
{
	if (__display_overlay) {
		_draw_number(__gc_cpu, percent);
	}
}

// See the header file for the function documentation
void display_overlay_update_fps(uint32_t fps)
{
	if (__display_overlay) {
		_draw_number(__gc_fps, fps);
	}
}

// See the header file for the function documentation
void display_overlay_flush(void *params)
{
	vg_lite_buffer_t *buffer = params;

	if (__display_overlay) {
		_clear(buffer, 0, 0, buffer->width, MAX(CPU_Y, FPS_Y) + __digit_height + 2, __back_color);
		_flush_image(buffer, __gc_cpu, CPU_X, CPU_Y);
		_flush_image(buffer, __gc_fps, FPS_X, FPS_Y);
	}
}

// See the header file for the function documentation
void display_overlay_toggle(void)
{
	__display_overlay = !__display_overlay;
}

// -----------------------------------------------------------------------------
// Internal functions
// -----------------------------------------------------------------------------

void _overlay_task(void *params) {
	while(1)
	{
		vTaskDelay(OVERLAY_TASK_SLEEP_TICKS);
		display_overlay_update_fps(framerate_get());
		display_overlay_update_cpu(cpuload_get());
	}
}

// See the section 'Internal function definitions' for the function documentation
static void _clear(vg_lite_buffer_t *dst, int x, int y, int w, int h, vg_lite_color_t color) {
	vg_lite_rectangle_t     rectangle;

	rectangle.x = x;
	rectangle.y = y;
	rectangle.width = w;
	rectangle.height = h;

	vg_lite_clear(dst, &rectangle, color);
}

// See the section 'Internal function definitions' for the function documentation
static int _flush_image(vg_lite_buffer_t *dst, MICROUI_GraphicsContext* gc, uint32_t x, uint32_t y) {
	vg_lite_buffer_t src;
	vg_lite_matrix_t matrix;

	if (IS_READY())
	{
		MICROUI_Image* image = &gc->image;

		if (!display_vglite_configure_source(&src, image)) {
			MEJ_LOG_ERROR("failed: image format not supported");
			return -1;
		}

		vg_lite_identity(&matrix);
		vg_lite_translate(x, y, &matrix);

		vg_lite_blit(dst, &src, &matrix, VG_LITE_BLEND_SRC_OVER, 0xFFFFFFFF, VG_LITE_FILTER_POINT);
	}

	return 0;
}

// See the section 'Internal function definitions' for the function documentation
static void _draw_number(MICROUI_GraphicsContext* gc, uint32_t number) {
	if (!IS_READY())
	{
		return;
	}

	uint32_t hundred = number / 100;
	if (hundred > 9)
	{
		// invalid number : draw nothing
		return;
	}

	// erase previous number
	jint gcColor = gc->foreground_color;
	gc->foreground_color = __back_color;
	UI_DRAWING_SOFT_fillRectangle(gc, __first_digit_x, 0,
			__first_digit_x + 3 * __digit_width - 1, __digit_height - 1);
	gc->foreground_color = gcColor;

	uint32_t x = _draw_digit(gc, hundred, __first_digit_x, false);
	number -= (hundred * 100);

	uint32_t ten = number / 10;
	x = _draw_digit(gc, ten, x, hundred > 0);
	number -= (ten * 10);

	x = _draw_digit(gc, number, x, true);

	// ask a refresh
	LLUI_DISPLAY_requestFlush(true);
}

// See the section 'Internal function definitions' for the function documentation
static uint32_t _draw_digit(MICROUI_GraphicsContext* gc, uint32_t digit, uint32_t x, bool draw_zero) {
	uint32_t digit_y = digit * __digit_height;
	if (digit > 0 || draw_zero)
	{
		// Software alforithms are not much slower on small images
		UI_DRAWING_SOFT_drawImage(gc, __digits, 0, digit_y, __digit_width, __digit_height, x, 0, 0xff);
	}
	return x += __digit_width;
}

// See the section 'Internal function definitions' for the function documentation
static void __display_overlay_toggle_callback(void *params) {
	display_overlay_toggle();
}

// -----------------------------------------------------------------------------
// EOF
// -----------------------------------------------------------------------------

