/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.annotation.processor.model;

import java.util.Objects;

/**
 * An attribute is a model for GIML attributes in XML files. There are 2 kind of attributes:
 * <ul>
 * <li>Standard attributes, which can have default values,</li>
 * <li>Required attributes, which cannot have default values.</li>
 * </ul>
 */
public class Attribute {

	private final String _name; // NOSONAR
	private final String _defaultValue; // NOSONAR
	private final String _normalizedDefaultValue; // NOSONAR
	private final String _class; // NOSONAR
	private final boolean _isConstructorAttribute; // NOSONAR

	/**
	 * Instantiates a required attribute.
	 *
	 * @param name
	 *            the name of this attribute
	 * @param clazz
	 *            the fully qualified name of this attribute type.
	 * @param isConstructorAttribute
	 *            {@code true} if this attribute has been declared in a constructor, {@code false} otherwise.
	 */
	public Attribute(String name, String clazz, boolean isConstructorAttribute) {
		this(name, null, null, clazz, isConstructorAttribute);
	}

	/**
	 * Instantiates a standard attribute.
	 *
	 * @param name
	 *            the name of this attribute
	 * @param defaultValue
	 *            default value of this attribute.
	 * @param clazz
	 *            the fully qualified name of this attribute type.
	 * @param isConstructorAttribute
	 *            {@code true} if this attribute has been declared in a constructor, {@code false} otherwise.
	 */
	public Attribute(String name, String defaultValue, String normalizedDefaultValue, String clazz,
			boolean isConstructorAttribute) {
		this._name = name;
		this._defaultValue = defaultValue;
		this._normalizedDefaultValue = normalizedDefaultValue;
		this._class = clazz;
		this._isConstructorAttribute = isConstructorAttribute;
	}

	/**
	 * Gets the name of this attribute.
	 *
	 * @return the name of this attribute.
	 */
	public String getName() {
		return this._name;
	}

	/**
	 * Gets fully qualified name of this attribute type.
	 *
	 * @return the fully qualified name of this attribute type.
	 */
	public String getAttributeClassName() {
		return this._class;
	}

	/**
	 * Gets default value of this attribute.
	 *
	 * @return the default value of this attribute, or {@code null} if this attribute is required.
	 */
	public String getDefaultValue() {
		return this._defaultValue;
	}

	/**
	 * Gets normalized default value of this attribute.
	 *
	 * @return the normalized default value of this attribute, or {@code null} if this attribute is required.
	 */
	public String getNormalizedDefaultValue() {
		return this._normalizedDefaultValue;
	}

	/**
	 * Gets of this attribute is required or standard.
	 *
	 * @return {@code true} if this attribute is required, {@code false} otherwise.
	 */
	public boolean isRequired() {
		return this._defaultValue == null;
	}

	/**
	 * Gets if this attribute has been declared in a constructor.
	 *
	 * @return {@code true} if this attribute has been declared in a constructor, {@code false} otherwise.
	 */
	public boolean isConstructorAttribute() {
		return this._isConstructorAttribute;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null || !(obj instanceof Attribute)) {
			return false;
		}

		Attribute other = (Attribute) obj;
		boolean defaultEquals = this._defaultValue != null ? this._defaultValue.equals(other._defaultValue)
				: other._defaultValue == null;
		return this._name.equalsIgnoreCase(other._name) && this._class.equals(other._class) && defaultEquals
				&& this._isConstructorAttribute == other._isConstructorAttribute;
	}

	@Override
	public int hashCode() {
		return Objects.hash(this._name, this._defaultValue, this._class, this._isConstructorAttribute);
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append(this._name).append('(').append(this._class);
		if (!this.isRequired()) {
			builder.append(",defaultValue=").append(this._defaultValue).append(']');
		}
		builder.append(')');
		return builder.toString();
	}
}
