/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.binary.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.io.ByteArrayOutputStream;
import java.io.IOException;

import org.junit.Ignore;
import org.junit.Test;

import com.is2t.giml.binary.GIMLObjectMap;
import com.is2t.giml.binary.GIMLObjectMap.Bucket;
import com.is2t.giml.binary.GIMLObjectMap.GIMLObjectReference;

/**
 * 
 */
public class GIMLObjectMapTest {

	/**
	 * Test empty map
	 */
	@Test
	public void testEmpty() {
		GIMLObjectMap map = new GIMLObjectMap(1.0f);
		//		map.put(new Object(), 12, 13);

		byte[] mapContent = completeMap(map);
		assertBucketTableLength(mapContent, 0);
	}

	/**
	 * Test simple map with one element
	 */
	@Test
	public void testSimple() {
		GIMLObjectMap map = new GIMLObjectMap(1.0f);
		map.put(new Object(), 12, 13);

		byte[] mapContent = completeMap(map);
		assertBucketTableLength(mapContent, 1);
	}

	/**
	 * Test dispatch of the entries in the bucket
	 */
	@Test
	public void testHash(){
		GIMLObjectMap map = new GIMLObjectMap(2.0f);

		putIntegerElements(map, 4);

		byte[] mapContent = completeMap(map);
		assertBucketTableLength(mapContent, 2);

		Bucket bucket = getBucket(mapContent, 0);
		assertEquals(2, bucket.size());

		GIMLObjectReference firstEntry = bucket.getEntry(0);
		GIMLObjectReference secondEntry = bucket.getEntry(1);
		if(firstEntry.identifierAddress==0 && firstEntry.objectAddress==1){
			assertEquals(20, secondEntry.identifierAddress);
			assertEquals(21, secondEntry.objectAddress);
		}
		else if(firstEntry.identifierAddress==20 && firstEntry.objectAddress==21){
			assertEquals(0, secondEntry.identifierAddress);
			assertEquals(1, secondEntry.objectAddress);
		}
		else {
			fail();
		}

		bucket = getBucket(mapContent, 1);
		assertEquals(2, bucket.size());
		firstEntry = bucket.getEntry(0);
		secondEntry = bucket.getEntry(1);
		if(firstEntry.identifierAddress==10 && firstEntry.objectAddress==11){
			assertEquals(30, secondEntry.identifierAddress);
			assertEquals(31, secondEntry.objectAddress);
		}
		else if(firstEntry.identifierAddress==30 && firstEntry.objectAddress==31){
			assertEquals(10, secondEntry.identifierAddress);
			assertEquals(11, secondEntry.objectAddress);
		}
		else {
			fail();
		}
	}

	/**
	 * Test load factor.
	 */
	@Test
	public void testLoadFactor(){
		testLoadFactor(1.0f, 4, 4);
		testLoadFactor(2.0f, 8, 4);
		testLoadFactor(0.5f, 8, 16);
		testLoadFactor(3f, 9, 3);
	}

	/**
	 * Test too many entries in the same bucket.
	 */
	@Test
	public void testTooManyEntriesInSameBucket(){
		GIMLObjectMap map = new GIMLObjectMap(255);
		putIntegerElements(map, 255);
		byte[] mapContent = completeMap(map);
		assertBucketTableLength(mapContent, 1);

		map = new GIMLObjectMap(255);
		putIntegerElements(map, 256);
		mapContent = completeMap(map);
		assertBucketTableLength(mapContent, 2);

		map = new GIMLObjectMap(255);
		putIntegerElements(map, 510);
		mapContent = completeMap(map);
		assertBucketTableLength(mapContent, 2);
	}

	/**
	 * Test too many entries with the same hashcode
	 */
	@Test
	public void testTooManyEntriesWithSameHashcode(){
		GIMLObjectMap map;
		byte[] mapContent;

		map = new GIMLObjectMap(1);
		for(int i=0 ; i<255 ; i++){
			map.put(new HashcodeTester(42), i*10, (i*10)+1);
		}
		mapContent = completeMap(map);
		assertBucketTableLength(mapContent, 255);

		map = new GIMLObjectMap(1);
		for(int i=0 ; i<256 ; i++){
			map.put(new HashcodeTester(42), i*10, (i*10)+1);
		}
		try{
			mapContent = completeMap(map);
			fail();
		}
		catch(InternalError e){

		}
	}

	/**
	 * Test too many entries.
	 */
	@Test
	public void testTooManyEntries(){
		GIMLObjectMap map = new GIMLObjectMap(255.0f);
		putIntegerElements(map, 65025);
		byte[] mapContent = completeMap(map);
		assertBucketTableLength(mapContent, 255);

		map = new GIMLObjectMap(0.5f);
		putIntegerElements(map, 30000);
		mapContent = completeMap(map);
		assertBucketTableLength(mapContent, 60000);

		map = new GIMLObjectMap(0.5f);
		putIntegerElements(map, 40000);
		try{
			mapContent = completeMap(map);
			fail();
		}
		catch(InternalError e){
		}
	}

	/**
	 * Test put after complete
	 */
	@Test
	public void testPutAfterComplete(){
		GIMLObjectMap map = new GIMLObjectMap(1f);
		map.complete();
		try{
			map.put(new Object(), 0, 0);
			fail();
		}
		catch(IllegalStateException e){

		}
	}

	@Ignore
	static class HashcodeTester {
		int hashcode;

		public HashcodeTester(int hashcode) {
			this.hashcode = hashcode;
		}

		@Override
		public int hashCode() {
			return hashcode;
		}
	}

	private void testLoadFactor(float loadFactor, int nbElements, int expectedBucketNumber){
		GIMLObjectMap map = new GIMLObjectMap(loadFactor);
		putIntegerElements(map, nbElements);
		byte[] mapContent = completeMap(map);
		assertBucketTableLength(mapContent, expectedBucketNumber);
	}

	private void putIntegerElements(GIMLObjectMap map, int number){
		for(int i=0 ; i<number ; i++){
			map.put(Integer.valueOf(i), i*10, (i*10)+1);
		}
	}

	/**
	 * Complete the given map and returns a byte[] with its content.
	 * Test the length of the byte[].
	 */
	private byte[] completeMap(GIMLObjectMap map){
		int size = map.complete();

		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			map.writeTo(out);
		} catch (IOException e) {
			fail(e.toString());
			return new byte[0];
		}

		assertEquals(size, out.size());

		return out.toByteArray();
	}

	public Bucket getBucket(byte[] mapContent, int offset){
		int bucketAddr = read16Bits(mapContent, 2 + (offset * 2));

		int nbEntries = mapContent[bucketAddr]&0xFF;

		Bucket bucket = new Bucket(nbEntries);
		int entryAddr = bucketAddr + 1;
		while(--nbEntries>=0){
			GIMLObjectReference objectRef = new GIMLObjectReference(read24Bits(mapContent, entryAddr), read24Bits(mapContent, entryAddr+3));
			try {
				bucket.addEntry(objectRef);
			} catch (IOException e) {
				fail(e.toString());
			}
			entryAddr += 6;
		}

		return bucket;
	}

	private int read16Bits(byte[] mapContent, int offset) {
		return ((mapContent[offset]&0xFF)<<8) | (mapContent[offset+1]&0xFF);
	}

	private int read24Bits(byte[] mapContent, int offset) {
		return ((mapContent[offset]&0xFF)<<16) | ((mapContent[offset+1]&0xFF)<<8) | (mapContent[offset+2]&0xFF);
	}

	private void assertBucketTableLength(byte[] mapContent, int expectedLength){
		int tableLength = ((mapContent[0]&0xFF)<<8) | (mapContent[1]&0xFF);
		assertEquals(expectedLength, tableLength);
	}



}
