/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.model.test;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import org.junit.Test;

import com.is2t.giml.model.GIMLModelFactory;
import com.is2t.giml.model.exception.GIMLMisplacedElementException;
import com.is2t.giml.model.exception.GIMLModelException;
import com.is2t.giml.model.exception.GIMLTemplateInvocationException;
import com.is2t.giml.model.exception.GIMLUnknownElementException;
import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.model.Template;

import junit.framework.Assert;

/**
 * The class {@code GIMLModelFactoryTest} contains unit tests for class {@link GIMLModelFactory}.
 */
public class GIMLModelFactoryTest extends AbstractTest {

	// Invalid content

	@Test(expected = GIMLUnknownElementException.class)
	public void testUnknownELement() throws GIMLModelException {
		List<Node> children = new ArrayList<Node>();
		children.add(createNode("unknown-node", NO_ATTRIBUTE));

		Document document = new Document("docId", "root", new HashSet<Attribute>(), children,
				new ArrayList<Document>());
		GIMLModelFactory.createGIMLDocument(document);
	}

	@Test(expected = GIMLTemplateInvocationException.class)
	public void testBadTemplateName() throws GIMLModelException {
		List<Template> templates = new ArrayList<Template>();
		Node templateNode = createNode(TEMPLATE_TAG_NAME, createAttributeSet(TEMPLATE_ATTR_NAME, "node1"));
		templates.add(createTemplate(templateNode));

		Document document = new Document("docId", "root", new HashSet<Attribute>(), new ArrayList<Node>(),
				new ArrayList<Document>(), templates);
		GIMLModelFactory.createGIMLDocument(document);
	}

	@Test(expected = GIMLMisplacedElementException.class)
	public void testMisplacedRootElement() throws GIMLModelException {
		List<Node> children = new ArrayList<Node>();
		children.add(createNode("root1", NO_ATTRIBUTE, createNode("root2", NO_ATTRIBUTE)));

		Document document = new Document("docId", "root", new HashSet<Attribute>(), children,
				new ArrayList<Document>());
		GIMLModelFactory.createGIMLDocument(document);
	}

	@Test(expected = GIMLMisplacedElementException.class)
	public void testMisplacedChildElement() throws GIMLModelException {
		List<Node> children = new ArrayList<Node>();
		children.add(createNode("node1", NO_ATTRIBUTE));

		Document document = new Document("docId", "root", new HashSet<Attribute>(), children,
				new ArrayList<Document>());
		GIMLModelFactory.createGIMLDocument(document);
	}

	// Valid content

	@Test
	public void testNoNormalizationOnAlreadyNormalizedElements() throws GIMLModelException {
		List<Node> children = new ArrayList<Node>();
		Node root = createNode("root1", NO_ATTRIBUTE,
				createNode("element", NO_ATTRIBUTE, createNode("node1", NO_ATTRIBUTE)));
		children.add(root);

		Document document = new Document("docId", "root", new HashSet<Attribute>(), children,
				new ArrayList<Document>());
		Document gimlDocument = GIMLModelFactory.createGIMLDocument(document);

		Assert.assertEquals(gimlDocument.getChildCount(), 1);
		Assert.assertEquals(gimlDocument.getChildren().get(0), root); // so there is no "element" node added
	}

	@Test
	public void testNoNormalizationOfEmptyElement() throws GIMLModelException {
		List<Node> children = new ArrayList<Node>();
		Node root = createNode("root1", NO_ATTRIBUTE);
		children.add(root);

		Document document = new Document("docId", "root", new HashSet<Attribute>(), children,
				new ArrayList<Document>());
		Document gimlDocument = GIMLModelFactory.createGIMLDocument(document);

		Assert.assertEquals(gimlDocument.getChildCount(), 1);
		Assert.assertEquals(gimlDocument.getChildren().get(0), root); // so there is no "element" node added
	}

	@Test
	public void testNormalizationOfRootElement() throws GIMLModelException {
		List<Node> children = new ArrayList<Node>();
		Node root = createNode("root1", NO_ATTRIBUTE, createNode("node1", NO_ATTRIBUTE));
		children.add(root);

		Document document = new Document("docId", "root", new HashSet<Attribute>(), children,
				new ArrayList<Document>());
		Document gimlDocument = GIMLModelFactory.createGIMLDocument(document);

		Node expectedNode = createNode("root1", NO_ATTRIBUTE,
				createNode("element", NO_ATTRIBUTE, createNode("node1", NO_ATTRIBUTE)));
		Assert.assertEquals(gimlDocument.getChildCount(), 1);
		Assert.assertEquals(gimlDocument.getChildren().get(0), expectedNode);
	}

	@Test
	public void testNormalizationOfChildElements() throws GIMLModelException {
		List<Node> children = new ArrayList<Node>();
		Node root = createNode("root2", NO_ATTRIBUTE,
				createNode("node1", NO_ATTRIBUTE, createNode("node1", NO_ATTRIBUTE)));
		children.add(root);

		Document document = new Document("docId", "root", new HashSet<Attribute>(), children,
				new ArrayList<Document>());
		Document gimlDocument = GIMLModelFactory.createGIMLDocument(document);

		Node expectedNode = createNode("root2", NO_ATTRIBUTE, createNode("node1", NO_ATTRIBUTE,
				createNode("element-node", NO_ATTRIBUTE, createNode("node1", NO_ATTRIBUTE))));
		Assert.assertEquals(gimlDocument.getChildCount(), 1);
		Assert.assertEquals(gimlDocument.getChildren().get(0), expectedNode);
	}

}
