/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Objects;
import java.util.Set;

/**
 * The class {@code Template} represents a template as a tree-like structure. A template has a name and an unique root
 * element and can be parameterized.
 */
public class Template {

	private final Node _templateNode;
	private final String _templateTagName; // NOSONAR
	private final String _templateAttributeName; // NOSONAR
	private final String _parameterPrefix; // NOSONAR
	private final String _parameterSuffix; // NOSONAR
	private final String _contentTagName; // NOSONAR
	private final String _contentAttributeName; // NOSONAR
	private final String _contentInstantiationAttributeName; // NOSONAR

	private final String _name; // NOSONAR
	private final Set<String> _parameters; // NOSONAR
	private final Set<String> _contentIds; // NOSONAR
	private final Node _root; // NOSONAR

	/**
	 * Create a template from the given node.
	 *
	 * @param templateNode
	 *            the node describing the template.
	 * @param templateTagName
	 *            the name of the template tag.
	 * @param templateNameAttributeName
	 *            the name of the attribute in the given template node that describe the template name.
	 * @param parameterPrefix
	 *            the prefix that mark the beginning of a parameter in attribute values.
	 * @param parameterSuffix
	 *            the suffix that mark the end of a parameter in attribute values.
	 * @param contentTagName
	 *            the name of the template content tag.
	 * @param contentIdAttributeName
	 *            the name of the attribute that describe the ID of template content.
	 * @param contentInstantiationIdAttributeName
	 *            the name of the attribute that describe the reference ID of the content to replace during template
	 *            instantiation.
	 *
	 * @throws NullPointerException
	 *             if given node is <code>null</code>.
	 * @throws IllegalArgumentException
	 *             if something wrong happened when creating the template node.
	 */
	public Template(Node templateNode, String templateTagName, String templateNameAttributeName, String parameterPrefix, // NOSONAR
			String parameterSuffix, String contentTagName, String contentIdAttributeName, // NOSONAR
			String contentInstantiationIdAttributeName) throws NullPointerException, IllegalArgumentException { // NOSONAR
		this._templateNode = templateNode;
		this._name = templateNode.getAttributeValue(templateNameAttributeName);
		this._templateTagName = templateTagName;
		this._templateAttributeName = templateNameAttributeName;
		this._parameterPrefix = parameterPrefix;
		this._parameterSuffix = parameterSuffix;
		this._contentTagName = contentTagName;
		this._contentAttributeName = contentIdAttributeName;
		this._contentInstantiationAttributeName = contentInstantiationIdAttributeName;
		this._contentIds = new HashSet<String>();
		this.checkTemplateNodeValidity(templateNode);

		// At this point, the given node is a valid representation of a template
		this._parameters = new HashSet<String>();
		for (Node child : templateNode.getChildren()) {
			this.extractTemplateParameters(child);
		}
		if (templateNode.getChildCount() == 1) {
			this._root = templateNode.getChildren().get(0);
		} else {
			this._root = null;
		}
	}

	/**
	 * Gets the node describing this template.
	 * 
	 * @return the node describing this template.
	 */
	public Node getTemplateNode() {
		return this._templateNode;
	}

	/**
	 * Gets the name of this template.
	 *
	 * @return the name of this template.
	 */
	public String getName() {
		return this._name;
	}

	/**
	 * Gets the parameter prefix that determine parameter start for this template.
	 *
	 * @return the parameter prefix for this template.
	 */
	public String getParameterPrefix() {
		return this._parameterPrefix;
	}

	/**
	 * Gets the parameter suffix that determine parameter end for this template.
	 *
	 * @return the parameter suffix for this template.
	 */
	public String getParameterSuffix() {
		return this._parameterSuffix;
	}

	/**
	 * Gets the template tag name for this template.
	 *
	 * @return the template tag name for this template.
	 */
	public String getTemplateTagName() {
		return this._templateTagName;
	}

	/**
	 * Gets the template attribute name for this template to determine the template name.
	 *
	 * @return the template attribute name for this template.
	 */
	public String getTemplateTagAttributeName() {
		return this._templateAttributeName;
	}

	/**
	 * Gets the content tag name of this template.
	 *
	 * @return the content tag name of this template.
	 */
	public String getContentTagName() {
		return this._contentTagName;
	}

	/**
	 * Gets the content attribute name for this template to determine the content ID.
	 *
	 * @return the content attribute name for this template.
	 */
	public String getContentTagAttributeName() {
		return this._contentAttributeName;
	}

	/**
	 * Gets the content attribute name in instantiation for this template to determine which content to replace.
	 *
	 * @return the content attribute name in instantiation for this template.
	 */
	public String getContentTagInstantiationAttributeName() {
		return this._contentInstantiationAttributeName;
	}

	/**
	 * Gets the number of content elements of this template.
	 *
	 * @return the number of content elements of this template.
	 */
	public int getContentCount() {
		return this._contentIds.size();
	}

	/**
	 * Gets the immutable set of content IDs of this template.
	 *
	 * @return the set of content IDs of this template
	 */
	public Set<String> getContentIds() {
		return Collections.unmodifiableSet(this._contentIds);
	}

	/**
	 * Gets if this template has a content with the given ID.
	 *
	 * @param value
	 *            the content ID to look for.
	 * @return <code>true</code> if this template has a content with the given ID, <code>false</code> otherwise.
	 */
	public boolean hasContentId(String value) {
		return this._contentIds.contains(value);
	}

	/**
	 * Gets the number of parameters in this template.
	 *
	 * @return the number of parameters in this template.
	 */
	public int getParameterCount() {
		return this._parameters.size();
	}

	/**
	 * Gets the immutable list of parameter names of this template.
	 *
	 * @return the list of parameter names of this template.
	 */
	public Set<String> getParameters() {
		return Collections.unmodifiableSet(this._parameters);
	}

	/**
	 * Gets if this template has the given parameter.
	 *
	 * @param value
	 *            the parameter to look for.
	 * @return <code>true</code> if this template has the given parameter, <code>false</code> otherwise.
	 */
	public boolean hasParameter(String value) {
		return this._parameters.contains(value);
	}

	/**
	 * Gets the root node of this template.
	 *
	 * @return the root node of this template.
	 */
	public Node getRoot() {
		return this._root;
	}

	/**
	 * Instantiates this template using information contained in the given node.
	 *
	 * @param instantationNode
	 *            the node containing information for template invocation (i.e. attributes, children,...).
	 * @return the content of this template with information contained in the given instantiation node, or
	 *         <code>null</code> if template is empty.
	 * @throws IllegalArgumentException
	 *             if something wrong happened when instantiating this template.
	 */
	public List<Node> instantiate(Node instantationNode) throws IllegalArgumentException {
		// First, check the node name is the same as this template name
		if (!this._name.equalsIgnoreCase(instantationNode.getName())) {
			throw new IllegalArgumentException(
					String.format("Wrong template instantiation name (expected '%s', found '%s')", this._name,
							instantationNode.getName()));
		}

		// Then, check invocation node attributes match template attribute (need perfect match)
		// We use Set#equals since we don't care about element order
		Set<String> instantationNodeParameters = new HashSet<String>();
		for (Attribute attr : instantationNode.getAttributes()) {
			instantationNodeParameters.add(attr.getName());
		}
		if (!this._parameters.equals(instantationNodeParameters)) {
			throw new IllegalArgumentException(String.format(
					"Invalid parameters in template '%s' instantiation (expected '%s', found '%s')", this._name,
					this.setToString(this._parameters), this.setToString(instantationNodeParameters)));
		}

		// Then, check all child of the instantiation node are "content" elements and if their reference ID perfectly
		// match a content of this template
		Set<String> instantiationNodeContents = new HashSet<String>();
		for (Node child : instantationNode.getChildren()) {
			String childName = child.getName();
			if (!this._contentTagName.equalsIgnoreCase(childName)) {
				throw new IllegalArgumentException(String.format(
						"Template instantion child elements must be named '%s' (found '%s' in '%s' template invocation)",
						this._contentTagName, childName, this._name));
			}

			if (child.getAttribute(this._contentInstantiationAttributeName) == null) {
				throw new IllegalArgumentException(String.format(
						"Template instantiation content elements must have attribute '%s' (in '%s' template invocation)",
						this._contentInstantiationAttributeName, this._name));
			}

			String childContentId = child.getAttributeValue(this._contentInstantiationAttributeName);
			instantiationNodeContents.add(childContentId);
		}
		if (!this._contentIds.equals(instantiationNodeContents)) {
			throw new IllegalArgumentException(String.format(
					"Invalid content in '%s' template instantiation (expected '%s', found '%s')", this._name,
					this.setToString(this._contentIds), this.setToString(instantiationNodeContents)));
		}

		// Template is empty, so there is nothing to instantiate.
		if (this._root == null) {
			return null;
		}

		// Here, instantiation node is valid. Re-create a node that is the template content with instantiation node
		// attribute values.
		List<Node> instantiationContent = instantationNode.getChildren();
		return this.instantiateNode(instantationNode, this._root, instantiationContent);
	}

	private void checkTemplateNodeValidity(Node templateNode) throws NullPointerException, IllegalArgumentException {
		// Check template tag name
		if (!this._templateTagName.equals(templateNode.getName())) {
			throw new IllegalArgumentException(
					String.format("Wrong tag name for template declaration (expected '%s', found '%s')",
							this._templateTagName, templateNode.getName()));
		}

		// Check template tag attribute name
		if (templateNode.getAttribute(this._templateAttributeName) == null) {
			throw new IllegalArgumentException(String.format("Expecting template declaration to have attribute '%s'",
					this._templateAttributeName));
		}

		// Check template contains only one child
		if (templateNode.getChildCount() > 1) {
			throw new IllegalArgumentException(
					String.format("Template '%s' must have at most one child element (found %d)",
							templateNode.getAttributeValue(this._templateAttributeName), templateNode.getChildCount()));
		}

		// Check if template contains "content" elements, and if they are well formed
		for (Node child : templateNode.getChildren()) {
			checkAndExtractValidContentElements(child);
		}
	}

	private void checkAndExtractValidContentElements(Node node) throws IllegalArgumentException {
		// Check current node
		if (this._contentTagName.equalsIgnoreCase(node.getName())) {
			// Check content node has ID attribute
			if (node.getAttribute(this._contentAttributeName) == null) {
				throw new IllegalArgumentException(
						String.format("Template content node must have property '%s' in template '%s'",
								this._contentAttributeName, this._name));
			}

			// Check content node has no child
			if (node.getChildCount() > 0) {
				throw new IllegalArgumentException(
						String.format("Template content cannot have child element in template '%s'", this._name));
			}

			// Register this content ID
			this._contentIds.add(node.getAttributeValue(this._contentAttributeName));
		}

		// Check in children too
		for (Node child : node.getChildren()) {
			checkAndExtractValidContentElements(child);
		}
	}

	private void extractTemplateParameters(Node templateNode) throws IllegalArgumentException {
		for (Attribute attribute : templateNode.getAttributes()) {
			// template are necessarily in attribute values
			String value = attribute.getValue();
			this.extractParameters(value);
		}
		// Extract parameters in children too
		for (Node child : templateNode.getChildren()) {
			this.extractTemplateParameters(child);
		}
	}

	private void extractParameters(String value) throws IllegalArgumentException {
		int indexOfParameterPrefix = value.indexOf(this._parameterPrefix);
		int indexOfParameterSuffix = value.indexOf(this._parameterSuffix);

		// We necessarily have one parameter in the given value (at least one)
		while (indexOfParameterPrefix != -1 || indexOfParameterSuffix != -1) {
			// while there exist a parameter, extract it
			if (indexOfParameterPrefix == -1 || indexOfParameterSuffix < indexOfParameterPrefix) {
				// there exist a suffix but the first suffix is before first prefix, malformed expression
				throw new IllegalArgumentException(
						String.format("Malformed template parameter invocation ('%s')", value));
			}

			String parameter = value.substring(indexOfParameterPrefix + this._parameterPrefix.length(),
					indexOfParameterSuffix);
			if (parameter.indexOf(this._parameterPrefix) != -1 || parameter.indexOf(this._parameterSuffix) != -1) {
				// in the declared parameter, there is the parameter prefix > malformed
				throw new IllegalArgumentException(String
						.format("Cannot invoke template parameter in another template parameter ('%s')", parameter));
			}

			// parameter does not contain nested parameter invocation, check if it can be a value attribute
			if (parameter.matches(".*\\s.*")) {
				throw new IllegalArgumentException(
						String.format("Template parameter cannot contain space characters ('%s')", parameter));
			}

			// parameter is well formed for sure
			this._parameters.add(parameter);

			// try to find next parameters
			value = value.substring(indexOfParameterSuffix + this._parameterSuffix.length());
			indexOfParameterPrefix = value.indexOf(this._parameterPrefix);
			indexOfParameterSuffix = value.indexOf(this._parameterSuffix);
		}
	}

	private List<Node> instantiateNode(Node templateInstantiation, Node node, List<Node> instantiationContent) {
		if (this._contentTagName.equalsIgnoreCase(node.getName())) {
			return this.instantiateContentNode(templateInstantiation, node, instantiationContent);
		} else {
			return this.instantiateNormalNode(templateInstantiation, node, instantiationContent);
		}
	}

	private List<Node> instantiateContentNode(Node templateInstantiation, Node contentNode,
			List<Node> instantiationContent) {
		List<Node> newChildren = new ArrayList<Node>();

		// First get the instantiation content
		List<Node> instantationContentChildren = null;
		for (Node content : instantiationContent) {
			if (contentNode.getAttributeValue(this._contentAttributeName)
					.equals(content.getAttributeValue(this._contentInstantiationAttributeName))) {
				instantationContentChildren = content.getChildren();
			}
		}
		// instantationContentChildren cannot be null (due to previous check)

		// Then, add to new child list the instantiated content
		for (Node toInstantiate : instantationContentChildren) {
			newChildren.addAll(this.instantiateNode(templateInstantiation, toInstantiate, instantiationContent));
		}

		return newChildren;
	}

	private List<Node> instantiateNormalNode(Node templateInstantiation, Node node, List<Node> instantiationContent) {
		// Instantiate all values in attributes
		Set<Attribute> parameterValues = templateInstantiation.getAttributes();
		Set<Attribute> parametrizedAttributes = new HashSet<Attribute>();
		for (Attribute templateAttribute : node.getAttributes()) {
			parametrizedAttributes.add(this.getParametrizedValue(parameterValues, templateAttribute));
		}

		// Do same thing for children
		List<Node> parametrizedChildren = new ArrayList<Node>();
		for (Node child : node.getChildren()) {
			parametrizedChildren.addAll(this.instantiateNode(templateInstantiation, child, instantiationContent));
		}

		List<Node> nodes = new ArrayList<Node>();
		nodes.add(new Node(node.getName(), parametrizedAttributes, parametrizedChildren));
		return nodes;
	}

	private Attribute getParametrizedValue(Set<Attribute> parameterValues, Attribute templateAttribute) {
		String newValue = templateAttribute.getValue();
		for (Attribute parameter : parameterValues) {
			String toReplace = this._parameterPrefix + parameter.getName() + this._parameterSuffix;
			newValue = newValue.replace(toReplace, parameter.getValue());
		}
		return new Attribute(templateAttribute.getName(), newValue);
	}

	private String setToString(Set<String> parameters) {
		StringBuilder builder = new StringBuilder();
		builder.append('[');
		Iterator<String> iterator = parameters.iterator();
		while (iterator.hasNext()) {
			String parameter = iterator.next();
			builder.append(parameter);
			if (iterator.hasNext()) {
				builder.append(',');
			}
		}
		builder.append(']');
		return builder.toString();
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}

		if (!(obj instanceof Template)) {
			return false;
		}

		Template other = (Template) obj;

		return this._name.equalsIgnoreCase(other._name) && this._parameters.equals(other._parameters)
				&& (this._root != null ? this._root.equals(other._root) : (other._root == null))
				&& equalsTemplateConfiguration(other) && equalsParametersConfiguration(other)
				&& equalsContentConfiguration(other);
	}

	@Override
	public int hashCode() {
		return Objects.hash(this._name, this._parameters, this._root, this._parameterPrefix, this._parameterSuffix,
				this._templateTagName, this._templateAttributeName);
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();

		builder.append("T[");
		builder.append("name=");
		builder.append(this._name);
		builder.append(',');
		builder.append("parameters=");
		builder.append('{');
		Iterator<String> iterator = this._parameters.iterator();
		while (iterator.hasNext()) {
			builder.append(iterator.next());
			if (iterator.hasNext()) {
				builder.append(',');
			}
		}
		builder.append('}');
		builder.append(',');
		if (this._root != null) {
			builder.append("root=");
			builder.append(this._root.getName());
		}
		builder.append(']');

		return builder.toString();
	}

	private boolean equalsTemplateConfiguration(Template other) {
		return this._templateTagName.equalsIgnoreCase(other._templateTagName)
				&& this._templateAttributeName.equalsIgnoreCase(other._templateAttributeName);
	}

	private boolean equalsParametersConfiguration(Template other) {
		return this._parameterPrefix.equals(other._parameterPrefix)
				&& this._parameterSuffix.equals(other._parameterSuffix);
	}

	private boolean equalsContentConfiguration(Template other) {
		return this._contentTagName.equalsIgnoreCase(other._contentTagName)
				&& this._contentAttributeName.equalsIgnoreCase(other._contentAttributeName)
				&& this._contentInstantiationAttributeName.equalsIgnoreCase(other._contentInstantiationAttributeName);
	}
}
