/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.model.test.document;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.junit.Test;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.model.Template;
import com.is2t.giml.xml.model.test.AbstractTest;

import junit.framework.Assert;

/**
 * Unit tests for {@link Document} class.
 */
public class DocumentTest extends AbstractTest {

	// We do not test behavior related to Node. @see NodeTest class

	private static final String VALID_ID = "avalidid";
	private static final String VALID_NAME = "avalidname";

	private List<Document> createReferencedDocumentList() {
		List<Document> referencedDocuments = new ArrayList<Document>();
		HashSet<Attribute> emptyAttr = new HashSet<Attribute>();

		List<Node> nodes1 = new ArrayList<Node>();
		nodes1.add(new Node("node1-1"));
		nodes1.add(new Node("node1-2"));

		List<Node> nodes2 = new ArrayList<Node>();
		nodes2.add(new Node("node2-1"));
		nodes2.add(new Node("node2-2"));
		nodes2.add(new Node("node3-2"));

		List<Node> nodes3 = new ArrayList<Node>();
		nodes3.add(new Node("node3-1"));

		Document document3 = new Document("id-doc-3", VALID_NAME, emptyAttr, nodes3);
		List<Document> doc2Referenced = new ArrayList<Document>();
		doc2Referenced.add(document3);
		Document document2 = new Document("id-doc-2", VALID_NAME, emptyAttr, nodes2, doc2Referenced);
		Document document1 = new Document("id-doc-1", VALID_NAME, emptyAttr, nodes1);

		referencedDocuments.add(document1);
		referencedDocuments.add(document2);

		return referencedDocuments;
	}

	private List<Template> createTemplateList() {
		List<Template> templates = new ArrayList<Template>();
		Set<Attribute> templateAttributes1 = createAttributeSet("name", "my-template1");
		Set<Attribute> templateAttributes2 = createAttributeSet("name", "my-template2");

		Set<Attribute> childAttributes = createAttributeSet("text", createTemplateParam("param1"));
		Node child = createNode("node1", childAttributes);
		Node templateNode = createNode("template", templateAttributes1, child);
		templates.add(createTemplate(templateNode));

		Node templateNode2 = createNode("template", templateAttributes2);
		templates.add(createTemplate(templateNode2));

		return templates;
	}

	private void checkReferencedDocumentCount(Document document, int expectedCount) {
		Assert.assertEquals(document.getReferencedDocumentCount(), expectedCount);
		Assert.assertEquals(document.getReferencedDocuments().size(), expectedCount);
	}

	private void checkTemplateCount(Document document, int expectedCount) {
		Assert.assertEquals(document.getTemplateCount(), expectedCount);
		Assert.assertEquals(document.getTemplates().size(), expectedCount);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testNullDocumentId() {
		new Document(null, VALID_NAME);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testEmptyDocumentId() {
		new Document("", VALID_NAME);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testNullReferencedDocuments() {
		new Document(VALID_ID, VALID_NAME, new HashSet<Attribute>(), new ArrayList<Node>(), null);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testNullTemplatesDocuments() {
		new Document(VALID_ID, VALID_NAME, new HashSet<Attribute>(), new ArrayList<Node>(), new ArrayList<Document>(),
				null);
	}

	@Test
	public void testGetId() {
		Document document = new Document(VALID_ID, VALID_NAME);
		Assert.assertEquals(document.getId(), VALID_ID);
	}

	@Test
	public void testReferencedDocuments() {
		List<Document> referencedDocumentList = createReferencedDocumentList();
		List<Node> docNode = new ArrayList<Node>();
		docNode.add(new Node("node0-1"));
		Document document = new Document("id", "doc", new HashSet<Attribute>(), docNode, referencedDocumentList);
		checkReferencedDocumentCount(document, referencedDocumentList.size());

		List<Document> documentReferencedDocuments = document.getReferencedDocuments();
		for (Document doc : referencedDocumentList) {
			Assert.assertTrue(documentReferencedDocuments.contains(doc));
		}
	}

	@Test
	public void testTemplates() {
		List<Template> templateList = createTemplateList();
		List<Node> docNode = new ArrayList<Node>();
		docNode.add(new Node("node0-1"));
		Document document = new Document("id", "doc", new HashSet<Attribute>(), docNode, new ArrayList<Document>(),
				templateList);
		checkTemplateCount(document, templateList.size());

		List<Template> templateDocuments = document.getTemplates();
		for (Template template : templateList) {
			Assert.assertTrue(templateDocuments.contains(template));
			Assert.assertNotNull(document.getTemplate(template.getName()));
		}
	}

	private void checkNodeExists(Document root, Document referencedDocument) {
		List<Node> rootChildren = root.getChildren();
		for (Node node : referencedDocument.getChildren()) {
			Assert.assertTrue(rootChildren.contains(node));
		}
		for (Document innerDoc : referencedDocument.getReferencedDocuments()) {
			checkNodeExists(root, innerDoc);
		}
	}

	@Test
	public void testCompactDocument() {
		List<Document> referencedDocumentList = createReferencedDocumentList();
		List<Node> docNode = new ArrayList<Node>();
		docNode.add(new Node("node0-1"));
		Document compactedDocument = new Document("id", "doc", new HashSet<Attribute>(), docNode,
				referencedDocumentList).compact();
		checkReferencedDocumentCount(compactedDocument, 0);

		// now check all referenced document children are in the root document
		for (Document referencedDocument : referencedDocumentList) {
			checkNodeExists(compactedDocument, referencedDocument);
		}
	}

	@Test
	public void resolveEmptyTemplate() {
		Node templateNode = createNode("template", createAttributeSet("name", "template-empty"));
		Template template = createTemplate(templateNode);
		List<Template> templates = new ArrayList<Template>();
		templates.add(template);

		Document document = new Document("id", "doc", new HashSet<Attribute>(), new ArrayList<Node>(),
				new ArrayList<Document>(), templates);

		Assert.assertEquals(document.getTemplateCount(), 1);
		Assert.assertEquals(document.getChildCount(), 0);

		document = document.resolveTemplates();

		Assert.assertEquals(document.getTemplateCount(), 0);
		Assert.assertEquals(document.getChildCount(), 0);
	}

	@Test
	public void testResolveTemplatesInDocument() {
		List<Template> templateList = createTemplateList();
		List<Node> docNode = new ArrayList<Node>();
		docNode.add(createNode("my-template1", createAttributeSet("param1", "value1")));
		Document document = new Document("id", "doc", new HashSet<Attribute>(), docNode, new ArrayList<Document>(),
				templateList).resolveTemplates();
		checkTemplateCount(document, 0);

		Assert.assertEquals(document.getChildCount(), 1);
		Node child = document.getChildren().get(0);
		Assert.assertEquals(child.getName(), "node1");
		Assert.assertEquals(child.getChildCount(), 0);
		Assert.assertEquals(child.getAttributeCount(), 1);
		Assert.assertEquals(child.getAttributeValue("text"), "value1");
	}

	@Test
	public void testResolveTemplateDefinedInIncludedDocumentWithCompact() {
		testResolveTemplateDefinedInIncludedDocument(true);
	}

	@Test
	public void testResolveTemplateDefinedInIncludedDocumentNoCompact() {
		testResolveTemplateDefinedInIncludedDocument(false);
	}

	public void testResolveTemplateDefinedInIncludedDocument(boolean compact) {
		// Sub document that contains template declaration
		List<Template> templateList = createTemplateList();
		Document includedDocument = new Document("id", "doc", new HashSet<Attribute>(), new ArrayList<Node>(),
				new ArrayList<Document>(), templateList);

		// Root document, that invoke templates defined in imported document
		List<Node> docNode = new ArrayList<Node>();
		docNode.add(createNode("my-template1", createAttributeSet("param1", "my param value")));
		List<Document> referencedDocuments = new ArrayList<Document>();
		referencedDocuments.add(includedDocument);
		Document document = new Document("id", "doc", new HashSet<Attribute>(), docNode, referencedDocuments);

		// Now resolve template in main document without compacting it, and check that template invocation is ok
		if (compact) {
			document = document.compact();
		}
		document = document.resolveTemplates();
		Node expectedNode = createNode("node1", createAttributeSet("text", "my param value"));
		Assert.assertEquals(document.getReferencedDocumentCount(), (compact ? 0 : 1));
		Assert.assertEquals(document.getTemplateCount(), 0);
		Assert.assertEquals(document.getChildCount(), 1);
		Assert.assertTrue(document.getChildren().contains(expectedNode));
	}

	@Test
	public void testDocumentEquals() {
		Document document1 = new Document(VALID_ID, VALID_NAME);
		Document document2 = new Document(VALID_ID, VALID_NAME);
		Document document3 = new Document(VALID_ID + "--", VALID_NAME);
		Document document4 = new Document(VALID_ID, VALID_NAME + "--");
		Document document5 = new Document(VALID_ID, VALID_NAME.toUpperCase());

		Assert.assertFalse(document1.equals(null));
		Assert.assertFalse(document1.equals(new String()));
		Assert.assertFalse(document1.equals(document3));
		Assert.assertFalse(document1.equals(document4));
		Assert.assertFalse(document3.equals(document4));

		Assert.assertTrue(document1.equals(document2));
		Assert.assertTrue(document1.equals(document5));
	}

}
