/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.model.test.node;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.UUID;

import org.junit.Test;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.model.test.AbstractTest;

import junit.framework.Assert;

/**
 * Unit tests for {@link Node} class.
 */
public class NodeTest extends AbstractTest {

	private static final String VALID_NAME = "avalidname";
	private static final String INTEGER_PREFIX = "integer";
	private static final String FLOAT_PREFIX = "float";
	private static final String DOUBLE_PREFIX = "double";

	private void checkAttributeCount(Node node, int expectedCount) {
		Assert.assertEquals(node.getAttributeCount(), expectedCount);
		Assert.assertEquals(node.getAttributes().size(), expectedCount);
		Assert.assertEquals(node.getAttributeNames().size(), expectedCount);
	}

	private void checkChildCount(Node node, int expectedCount) {
		Assert.assertEquals(node.getChildCount(), expectedCount);
		Assert.assertEquals(node.getChildren().size(), expectedCount);
	}

	private Set<Attribute> createAttributeList() {
		Set<Attribute> attributes = new HashSet<Attribute>();
		Random generator = new Random(System.currentTimeMillis());
		for (int i = 0; i < 5; i++) {
			attributes.add(new Attribute("attr-" + i, UUID.randomUUID().toString()));
			attributes.add(new Attribute(INTEGER_PREFIX + "-attr-" + i, String.valueOf(generator.nextInt())));
			attributes.add(new Attribute(FLOAT_PREFIX + "-attr-" + i, String.valueOf(generator.nextFloat())));
			attributes.add(new Attribute(DOUBLE_PREFIX + "-attr-" + i, String.valueOf(generator.nextDouble())));
		}
		return attributes;
	}

	private List<Node> createNodes() {
		List<Node> nodes = new ArrayList<Node>();
		for (int i = 0; i < 10; i++) {
			nodes.add(new Node(UUID.randomUUID().toString()));
		}
		return nodes;
	}

	@Test(expected = IllegalArgumentException.class)
	public void testNullNodeName() {
		new Node(null);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testEmptyNodeName() {
		new Node("");
	}

	@Test(expected = IllegalArgumentException.class)
	public void testNullNodeNameWithValidAttributes() {
		new Node(null, new HashSet<Attribute>());
	}

	@Test(expected = IllegalArgumentException.class)
	public void testEmptyNodeNameWithValidAttributes() {
		new Node("", new HashSet<Attribute>());
	}

	@Test(expected = IllegalArgumentException.class)
	public void testNullChildrenWithValidNameAndAttributes() {
		new Node(VALID_NAME, new HashSet<Attribute>(), null);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testValidNodeNameWithNullAttributes() {
		new Node(VALID_NAME, null);
	}

	@Test
	public void testValidNodeName() {
		Node node = new Node(VALID_NAME, new HashSet<Attribute>());
		Assert.assertEquals(node.getName(), VALID_NAME);
	}

	@Test
	public void testEmptyNode() {
		Node node = new Node(VALID_NAME);
		checkAttributeCount(node, 0);
		checkChildCount(node, 0);

		node = new Node(VALID_NAME, new HashSet<Attribute>());
		checkAttributeCount(node, 0);
		checkChildCount(node, 0);

		node = new Node(VALID_NAME, new HashSet<Attribute>(), new ArrayList<Node>());
		checkAttributeCount(node, 0);
		checkChildCount(node, 0);
	}

	@Test
	public void testNodeHasChild() {
		List<Node> nodeList = createNodes();
		Node node = new Node(VALID_NAME, new HashSet<Attribute>(), nodeList);

		// Must contains all this node
		for (Node n : nodeList) {
			Assert.assertTrue(node.hasChild(n.getName()));
		}

		// Must not contains those nodes
		Assert.assertFalse(node.hasChild(null));
		Assert.assertFalse(node.hasChild(""));
		Assert.assertFalse(node.hasChild("not-in-this-node"));
	}

	@Test
	public void testNodeAttributes() {
		Set<Attribute> attributeList = createAttributeList();
		Node node = new Node(VALID_NAME, attributeList);
		checkAttributeCount(node, attributeList.size());

		// Check all attributes can be read as string
		for (Attribute attribute : attributeList) {
			Assert.assertEquals(node.getAttributeValue(attribute.getName()), attribute.getValue());
		}

		// Check integer attributes
		for (Attribute attribute : attributeList) {
			if (attribute.getName().startsWith(INTEGER_PREFIX)) {
				Assert.assertEquals(node.getIntegerAttributeValue(attribute.getName()),
						new Integer(Integer.parseInt(attribute.getValue())));
			}
		}

		// Check float attributes
		for (Attribute attribute : attributeList) {
			if (attribute.getName().startsWith(FLOAT_PREFIX)) {
				Assert.assertEquals(node.getFloatAttributeValue(attribute.getName()),
						new Float(Float.parseFloat(attribute.getValue())));
			}
		}

		// Check double attributes
		for (Attribute attribute : attributeList) {
			if (attribute.getName().startsWith(DOUBLE_PREFIX)) {
				Assert.assertEquals(node.getDoubleAttributeValue(attribute.getName()),
						new Double(Double.parseDouble(attribute.getValue())));
			}
		}
	}

	@Test
	public void testNonExistingNodeAttributeValue() {
		Set<Attribute> attributeList = createAttributeList();
		Node node = new Node(VALID_NAME, attributeList);

		Assert.assertNull(node.getAttributeValue("non-existing-attr"));
		Assert.assertNull(node.getIntegerAttributeValue("non-existing-attr"));
		Assert.assertNull(node.getFloatAttributeValue("non-existing-attr"));
		Assert.assertNull(node.getDoubleAttributeValue("non-existing-attr"));
	}

	@Test
	public void testNodeChildren() {
		List<Node> nodeList = createNodes();
		Node node = new Node(VALID_NAME, new HashSet<Attribute>(), nodeList);
		checkChildCount(node, nodeList.size());

		List<Node> children = node.getChildren();
		for (Node n : nodeList) {
			Assert.assertTrue(children.contains(n));
		}
	}

	@Test
	public void testgetChildrenByName() {
		List<Node> nodeList = createNodes();
		Node node = new Node(VALID_NAME, new HashSet<Attribute>(), nodeList);

		String node1Name = nodeList.get(0).getName();
		String node2Name = nodeList.get(5).getName().toUpperCase();

		Assert.assertEquals(node.getChildrenByName(node1Name).size(), 1);
		Assert.assertEquals(node.getChildrenByName(node2Name).size(), 1);

		// Must contains all this node
		for (Node n : nodeList) {
			Assert.assertTrue(node.hasChild(n.getName()));
		}
	}

	@Test
	public void testNodeEquals() {
		Node node1 = new Node(VALID_NAME);
		Node node2 = new Node(VALID_NAME);
		Node node3 = new Node(VALID_NAME + "--");
		Node node4 = new Node(VALID_NAME.toUpperCase());

		Assert.assertFalse(node1.equals(null));
		Assert.assertFalse(node1.equals(new String()));
		Assert.assertFalse(node1.equals(node3));
		Assert.assertFalse(node2.equals(node3));

		Assert.assertTrue(node1.equals(node2));
		Assert.assertTrue(node1.equals(node4));
	}
}
