/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.nls.generator;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.Map.Entry;

public class NLSLocale {

	private static final String LOCALES_IMMUTABLES_SUFFIX = "locales";
	private String localeName;

	private String displayName;

	// the translations pool
	private final Map<String, String> translations;

	public NLSLocale(String localeName, String displayName) {
		this.translations = new Hashtable<String, String>();
		setLocaleName(localeName);
		setDisplayName(displayName);
	}

	/**
	 * Gets the translation message of the given message key.
	 *
	 * @param msgKey
	 *            the message key.
	 * @return the translation message.
	 */
	public String getTranslation(String msgKey) {
		return this.translations.get(msgKey);
	}

	/**
	 * Associates the the given message key to the given translation message.
	 *
	 * @param msgKey
	 *            the message key.
	 * @param translation
	 *            the translation message.
	 */
	public void addTranslation(String msgKey, String translation) {
		this.translations.put(msgKey, translation);

	}

	private Map<String, String> normalizeTranslations() {
		Map<String, String> normalizedTranslations = new HashMap<String, String>();
		for (Entry<String, String> entry : this.translations.entrySet()) {
			normalizedTranslations.put(entry.getKey(), normalizeTranslation(entry.getValue()));
		}
		return normalizedTranslations;
	}

	private String normalizeTranslation(String translation) {
		StringBuilder sb = new StringBuilder();
		int len = translation.length();
		for (int i = -1; ++i < len;) {
			char currentChar = translation.charAt(i);
			switch (currentChar) {

			case '\n':
				sb.append('\\');
				sb.append('n');
				break;
			case '\r':
				sb.append('\\');
				sb.append('r');
				break;
			case '\t':
				sb.append('\\');
				sb.append('t');
				break;
			case '\\':
				sb.append(currentChar);
				sb.append(currentChar);
				break;
			default:
				sb.append(currentChar);
			}
		}
		return sb.toString();
	}

	/**
	 * Sets the local display name.
	 *
	 * @param displayName
	 *            the display name.
	 * @throws NullPointerException
	 *             if the given display name is <code>null</code>
	 */
	public void setDisplayName(String displayName) {
		if (displayName == null) {
			throw new NullPointerException("null display name");
		}
		this.displayName = displayName;

	}

	/**
	 * Sets the locale name.
	 *
	 * @param localeName
	 *            the locale name.
	 * @throws NullPointerException
	 *             if the given locale name is <code>null</code>
	 * @throws IllegalArgumentException
	 *             if the given localeName is equal to {@link #LOCALES_IMMUTABLES_SUFFIX}.
	 *
	 */
	public void setLocaleName(String localeName) {
		if (localeName.equals(LOCALES_IMMUTABLES_SUFFIX)) {
			throw new IllegalArgumentException("invalid locale name " + localeName);
		}
		this.localeName = localeName;

	}

	/**
	 * Gets the display name.
	 *
	 * @return the display name.
	 */
	public String getDisplayName() {
		return this.displayName;
	}

	/**
	 * Gets the locale name.
	 *
	 * @return the locale name.
	 */
	public String getLocaleName() {
		return this.localeName;
	}

	@Override
	public boolean equals(Object o) {

		if (o == null || !(o instanceof NLSLocale)) {
			return false;
		}

		NLSLocale that = (NLSLocale) o;
		return this.localeName.equals(that.getLocaleName());
	}

	@Override
	public int hashCode() {
		return this.localeName.hashCode();
	}

}
