/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.nls.generator;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.Normalizer;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.fedorahosted.tennera.jgettext.Catalog;
import org.fedorahosted.tennera.jgettext.Message;
import org.fedorahosted.tennera.jgettext.catalog.parse.ParseException;

public class PoParser {

	private static final String PO = ".po";

	private final File poFilesDir;
	private final String header;
	private final Set<NLSLocale> locales;
	private final Set<String> messagesKeys;

	private File fileLastError;

	public PoParser(File poFilesDir, String header) {
		this.poFilesDir = poFilesDir;
		this.header = header;
		this.locales = new HashSet<NLSLocale>();
		this.messagesKeys = new HashSet<>();
	}

	/**
	 * Parses all existing .po files in the po files directory; creates a locale representation for each po file and
	 * computes the whole messages keys for all translations.
	 *
	 * @throws FileNotFoundException
	 *             if a po file can not be loaded.
	 * @throws NullPointerException
	 *             if the directory of the po files is <code>null</code>.
	 * @throws ParseException
	 *             if an error occurs during the parsing.
	 * @throws IllegalArgumentException
	 *             if the directory of the po files does not exist or does not denote a folder pathname.
	 * @throws IOException
	 *             if an I/O error occurs.
	 */
	public void parse() throws FileNotFoundException, IOException {
		this.fileLastError = null;
		if (this.poFilesDir == null) {
			return;
		}
		List<File> listFiles = listFiles(this.poFilesDir, this.header, PO);
		for (File file : listFiles) {
			org.fedorahosted.tennera.jgettext.PoParser parser = new org.fedorahosted.tennera.jgettext.PoParser();
			Catalog catalog = null;
			try {
				catalog = parser.parseCatalog(file);
			} catch (IOException | ParseException e) {
				this.fileLastError = file;
				throw new IOException(String.format("error while parsing the po file '%s'", file.getAbsoluteFile()), e);
			}
			Iterator<Message> messages = catalog.iterator();
			String fileName = file.getName();
			// use the file name as local name and display name by default
			String localeName = fileName.substring(0, fileName.length() - PO.length());
			String displayName = localeName;
			// new locale with the file name
			NLSLocale locale = new NLSLocale(localeName, displayName);

			while (messages.hasNext()) {
				Message msg = messages.next();
				if (msg.isObsolete() || msg.isPlural()) {
					// do not manage obsolete msg and plural msg
					continue;
				}
				if (msg.isHeader()) {
					String headerMsg = msg.getMsgstr();
					Pattern pat = Pattern.compile("Language-Team: *(.*)\n.*");
					Matcher matcher = pat.matcher(headerMsg);
					if (matcher.find()) {
						// get the display name
						displayName = matcher.group(1);
						// update display name
						locale.setDisplayName(displayName);
					}
					pat = Pattern.compile("Language: *(.*)\n.*");
					matcher = pat.matcher(headerMsg);
					if (matcher.find()) {
						// get the locale name
						localeName = matcher.group(1);
						// update the locale name
						locale.setLocaleName(localeName);
					}
					continue;
				}
				String poOriginalMsg = msg.getMsgid();
				String poTranslationMsg = msg.getMsgstr();
				// compute the msg key
				String nlsMsgKey = stringToKey(poOriginalMsg);
				// add the translation with the nls msg key
				locale.addTranslation(nlsMsgKey, poTranslationMsg);
				this.messagesKeys.add(nlsMsgKey);

			}
			this.locales.add(locale);
		}
	}

	/**
	 * Returns the list of created locales after the parsing.
	 *
	 * @return the list of created locales after the parsing.
	 */
	public Set<NLSLocale> getLocales() {
		return this.locales;
	}

	/**
	 * Returns the set of computed messages keys after the parsing.
	 *
	 * @return the set of computed messages keys after the parsing.
	 */
	public Set<String> getMessagesKeys() {
		return this.messagesKeys;
	}

	/**
	 * Returns the file on which last parsing error has occurred.
	 *
	 * @return the file on which last parsing error has occurred.
	 */
	public File getLastErrorOnFile() {
		return this.fileLastError;
	}

	/*
	 * Computes a key from the given string value. All characters that is not a letter or a digit are replaced by an
	 * underscore.
	 */
	static String stringToKey(String string) {
		string = Normalizer.normalize(string, Normalizer.Form.NFD);
		string = string.replaceAll("\\p{IncombiningDiacriticalMarks}+", "");
		int len = string.length();
		StringBuffer key = new StringBuffer();
		char firstChar = string.charAt(0);
		if (Character.isLetter(firstChar) || firstChar == '$' || firstChar == '_') {
			key.append(firstChar);
		} else {
			key.append('_');
			if (Character.isDigit(firstChar)) {
				key.append(firstChar);
			}
		}
		for (int i = 0; ++i < len;) {
			char c = string.charAt(i);
			if (Character.isLetterOrDigit(c) || firstChar == '$' || firstChar == '_') {
				key.append(c);
			} else {
				key.append('_');
			}
		}
		return key.toString();
	}

	/**
	 * Gets the list of files in the given folder and its sub-folders, which match the given extension.
	 *
	 * @param directory
	 *            the parent directory to look for files.
	 * @param extension
	 *            the extension of file to look for.
	 * @return a possibly empty list of files that match the given extension.
	 * @throws NullPointerException
	 *             if given arguments are <code>null</code>.
	 * @throws IllegalArgumentException
	 *             if the given folder does not exist or does not denote a folder pathname.
	 */
	public static List<File> listFiles(File directory, String prefix, String extension)
			throws NullPointerException, IllegalArgumentException {
		if (!directory.exists() || !directory.isDirectory()) {
			throw new IllegalArgumentException("Given file does not denote an existing directory pathname");
		}
		List<File> files = new ArrayList<File>();

		File[] children = directory.listFiles();
		for (File child : children) {
			if (child.isFile()) {
				if (child.getName().startsWith(prefix) && child.getName().endsWith(extension)) {
					files.add(child);
				}
			} else {
				// is directory
				files.addAll(listFiles(child, prefix, extension));
			}
		}

		return files;
	}

}
