/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2010-2021, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.net;

import java.net.SocketException;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.NetworkInterfaceNatives;

import ej.error.Message;
import ej.sni.SNI;

public final class NetworkInterface implements INetworkInterface {

	// ipv4 address info size (tag (1) + IP (4) + prefix (1) + hasBroadcast (1) + broadcast IP (4))
	private static final int IPV4_ADDR_INFO_SIZE = 11;
	// ipv6 address info size (tag (1) + IP (16) + prefix (1)
	private static final int IPV6_ADDR_INFO_SIZE = 18;

	// ipv4 address tag
	private static final int IPV4_ADDR_TAG = 4;
	// ipv6 address tag
	private static final int IPV6_ADDR_TAG = 6;

	// Hardware address size
	private static final int MAC_ADDR_SIZE = 6;

	private static final int INTERFACE_MAX_NAME_LENGTH = 128;

	/**
	 * Return a list of VM network interface objects.
	 *
	 * @return The list of network interfaces.
	 * @throws SocketException
	 */
	@Override
	public INetworkInterfaceData[] getVMInterfaces() throws SocketException {
		int nIfaces = NetworkInterfaceNatives.getVMInterfacesCount(false);

		//retry native returned with result
		while(nIfaces == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			nIfaces = NetworkInterfaceNatives.getVMInterfacesCount(true);
		}

		if (nIfaces < 0) {
			switch (nIfaces) {
			case NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED:
				throw new RuntimeException(NetErrors.NOT_IMPLEMENTED);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), nIfaces));
			}
		}
		INetworkInterfaceData[] res = new INetworkInterfaceData[nIfaces];
		for(int i = nIfaces; --i >= 0; ) {
			byte[] ifnameBuffer = new byte[INTERFACE_MAX_NAME_LENGTH];

			int ifnameLen = NetworkInterfaceNatives.getVMInterface(i, ifnameBuffer, ifnameBuffer.length, false);

			//retry native returned with result
			while(ifnameLen == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
				ifnameLen = NetworkInterfaceNatives.getVMInterface(i, ifnameBuffer, ifnameBuffer.length, true);
			}

			if(ifnameLen < 0){
				throw new SocketException(Message.at(NetErrors.getInstance(), ifnameLen));
			}

			if(ifnameLen == 0){
				//no interface found
				continue;
			}

			String name = SNI.toJavaString(ifnameBuffer);

			int count = NetworkInterfaceNatives.getVMInterfaceAddressesCount(i, false);

			//retry native returned with result
			while(count == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
				count = NetworkInterfaceNatives.getVMInterfaceAddressesCount(i, true);
			}

			if(count < 0){
				throw new SocketException(Message.at(NetErrors.getInstance(), count));
			}
			byte[][] addresses = new byte[count][];
			short[] prefixes = new short[count];
			byte[][] broadcasts = new byte[count][];
			while (--count >= 0) {

				int size = NetworkInterfaceNatives.getVMInterfaceAddress(i, count, ifnameBuffer,
						ifnameBuffer.length, false);

				//retry native returned with result
				while(size == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
					size = NetworkInterfaceNatives.getVMInterfaceAddress(i, count, ifnameBuffer,
							ifnameBuffer.length, true);
				}

				int addrTag;
				int addrSize;
				boolean addrMayHaveBroadcast;
				switch(size) {
				case IPV4_ADDR_INFO_SIZE:
					addrTag = IPV4_ADDR_TAG;
					addrSize = 4;
					addrMayHaveBroadcast = true;
					break;
				case IPV6_ADDR_INFO_SIZE:
					addrTag = IPV6_ADDR_TAG;
					addrSize = 16;
					addrMayHaveBroadcast = false;
					break;
				default:
					throw new SocketException("network interfaces contains an invalid address size");
				}
				if (ifnameBuffer[0] != addrTag) {
					throw new SocketException("Wrong format, expected " + addrTag);
				}
				byte[] address = new byte[addrSize];
				// +1 because of the tag
				System.arraycopy(ifnameBuffer, 1, address, 0, addrSize);
				addresses[count] = address;
				// prefix can be from 0 to 128
				// so we encode it on 1 byte, from 0 to 0x80
				prefixes[count] = (short) (ifnameBuffer[1 + addrSize] & 0xFF);

				byte[] broadcast;
				if (addrMayHaveBroadcast) {
					if (ifnameBuffer[addrSize + 2] != 0) {
						broadcast = new byte[addrSize];
						// +1 because of the tag
						// +1 because of the prefix
						// +1 because of hasBroadcast
						System.arraycopy(ifnameBuffer, addrSize + 3, broadcast, 0, addrSize);
					} else {
						broadcast = null;
					}
				} else {
					broadcast = null;
				}
				broadcasts[count] = broadcast;
			}
			res[i] = new VMNetworkInterfaceData(name, addresses, prefixes, broadcasts);
		}
		return res;
	}

	@Override
	public boolean isUp(String name) throws SocketException {

		byte[] nameBuffer = new byte[name.length() + 1];

		SNI.toCString(name, nameBuffer);
		int ret = NetworkInterfaceNatives.isUp(nameBuffer, name.length() + 1, false);

		//retry native returned with result
		while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			ret = NetworkInterfaceNatives.isUp(nameBuffer, name.length() + 1, true);
		}

		if (ret < 0) {
			switch (ret) {
			case NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED:
				throw new RuntimeException(NetErrors.NOT_IMPLEMENTED);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), ret));
			}
		}

		return (ret == 0); // 0 means true
	}

	@Override
	public boolean isLoopback(String name) throws SocketException {
		byte[] nameBuffer = new byte[name.length() + 1];
		SNI.toCString(name, nameBuffer);
		int ret = NetworkInterfaceNatives.isLoopback(nameBuffer,
				name.length() + 1, false);

		//retry native returned with result
		while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			ret = NetworkInterfaceNatives.isLoopback(nameBuffer,
					name.length() + 1, true);
		}

		if (ret < 0) {
			switch (ret) {
			case NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED:
				throw new RuntimeException(NetErrors.NOT_IMPLEMENTED);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), ret));
			}
		}

		return (ret == 0); // 0 means true
	}

	@Override
	public boolean isPointToPoint(String name) throws SocketException {
		byte[] nameBuffer = new byte[name.length() + 1];
		SNI.toCString(name, nameBuffer);
		int ret = NetworkInterfaceNatives.isPointToPoint(nameBuffer,
				name.length() + 1, false);

		//retry native returned with result
		while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			ret = NetworkInterfaceNatives.isPointToPoint(nameBuffer,
					name.length() + 1, true);
		}

		if (ret < 0) {
			switch (ret) {
			case NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED:
				throw new RuntimeException(NetErrors.NOT_IMPLEMENTED);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), ret));
			}
		}

		return (ret == 0); // 0 means true
	}

	@Override
	public boolean supportsMulticast(String name) throws SocketException {
		byte[] nameBuffer = new byte[name.length() + 1];
		SNI.toCString(name, nameBuffer);
		int ret = NetworkInterfaceNatives.supportsMulticast(nameBuffer,
				name.length() + 1, false);

		//retry native returned with result
		while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			ret = NetworkInterfaceNatives.supportsMulticast(nameBuffer,
					name.length() + 1, true);
		}

		if (ret < 0) {
			switch (ret) {
			case NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED:
				throw new RuntimeException(NetErrors.NOT_IMPLEMENTED);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), ret));
			}
		}

		return (ret == 0); // 0 means true
	}

	@Override
	public byte[] getHardwareAddress(String name)
			throws java.net.SocketException {
		byte[] nameBuffer = new byte[name.length() + 1];
		byte[] hwAddr = new byte[MAC_ADDR_SIZE];
		SNI.toCString(name, nameBuffer);
		int ret = NetworkInterfaceNatives.getHardwareAddress(nameBuffer,
				name.length() + 1, hwAddr, hwAddr.length, false);

		//retry native returned with result
		while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			ret = NetworkInterfaceNatives.getHardwareAddress(nameBuffer,
					name.length() + 1, hwAddr, hwAddr.length, true);
		}

		if (ret < 0) {
			switch (ret) {
			case NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED:
				throw new RuntimeException(NetErrors.NOT_IMPLEMENTED);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), ret));
			}
		}
		if (ret == 0) {
			return null;
		}
		byte[] resHwAddr = new byte[ret];
		System.arraycopy(hwAddr, 0, resHwAddr, 0, ret);
		return resHwAddr;
	}

	@Override
	public int getMTU(String name) throws SocketException {
		byte[] nameBuffer = new byte[name.length() + 1];
		SNI.toCString(name, nameBuffer);
		int ret = NetworkInterfaceNatives.getMTU(nameBuffer, name.length() + 1, false);

		// retry native returned with result
		while (ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			ret = NetworkInterfaceNatives.getMTU(nameBuffer, name.length() + 1, true);
		}

		// MTU == -1 means that MTU value is unknown
		if (ret < -1) {
			switch (ret) {
			case NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED:
				throw new RuntimeException(NetErrors.NOT_IMPLEMENTED);
			default:
				throw new SocketException(Message.at(NetErrors.getInstance(), ret));
			}
		}

		return ret;
	}
}
