/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2010-2021, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.net;

import java.io.IOException;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.StreamSocketChannelNatives;

public class StreamSocketChannel extends SocketChannel implements IStreamSocketChannel{

	public StreamSocketChannel() throws IOException {
		super();
		this.kind = Kind.SOCK_STREAM;
	}

	public StreamSocketChannel(int fd) throws IOException {
		super(fd);
	}

	@Override
	public int read(byte[] data, int offset, int length) throws IOException {
		// data is kept into heap
		int ret = readNative(data, offset, length);
		if (ret < -1) { // don't throw error if ret is -1 (-1 means end of file)
			checkError(ret);
		}
		if (ret == -1) {
			checkSockConnection(true);
		}
		return ret;
	}

	public int readNative(byte[] data, int offset, int length)
			throws IOException {
		int native_fd = this.nfd.getNativeFD();
		int ret = StreamSocketChannelNatives.readByteBufferNative(native_fd, kind, data, offset, length, false);

		//retry read operation until there are available data or an error
		while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			ret = StreamSocketChannelNatives.readByteBufferNative(native_fd, kind, data, offset, length, true);
		}

		return ret;
	}

	@Override
	public int write(byte[] data, int offset, int length) throws IOException {
		int ret = writeNative(data, offset, length);
		if (ret < 0 && ret != NetErrors.J_EPIPE /* BrokenPipe */) {
			checkSockConnection(false);
		}
		checkError(ret);
		return ret;
	}

	public int writeNative(byte[] data, int offset, int length)
			throws IOException {
		// retrive file descriptor associated to the socket
		int native_fd = this.nfd.getNativeFD();

		// Default SNI compliant implementation
		int writtenBytes = 0;

		int remaining;
		//while there is data to write, transmit by "IMMORTAL capacity" (at most) chunks
		while((remaining = (length - writtenBytes)) > 0) { //we did not reach the end of buffer
			//put into IMMORTAL
			int len = Math.min(remaining, data.length);

			int ret;

			ret = StreamSocketChannelNatives.writeByteBufferNative(native_fd, kind, data, offset, len, false);

			//retry write operation until data has been sent or an error occurred
			while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
				ret = StreamSocketChannelNatives.writeByteBufferNative(native_fd, kind, data, offset, len, true);
			}

			if (ret < 0) {
				return ret;// error occurred
			}
			writtenBytes += ret;
			offset += ret;
		}
		return writtenBytes;

	}

	@Override
	public int available() throws IOException {
		int ret = StreamSocketChannelNatives.available(this.nfd.getNativeFD(), false);

		//retry native returned with result
		while(ret == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT){
			ret = StreamSocketChannelNatives.available(this.nfd.getNativeFD(), true);
		}

		checkError(ret);
		return ret;
	}

	// FIXME server socket methods

	/**
	 * Accept an incoming connection, returning a new VMChannel, or null if the
	 * channel is nonblocking and no connection is pending.
	 *
	 * @return The accepted connection, or null.
	 * @throws IOException
	 *             If an IO error occurs.
	 */
	@Override
	public IStreamSocketChannel accept() throws IOException {
		int fd = this.nfd.getNativeFD();
		int new_fd = 0;

		// The accept operation blocks until a new connection is attempted
		// or timeout reached

		new_fd = StreamSocketChannelNatives.accept(fd, false);

		while (new_fd == NetErrors.J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			new_fd = StreamSocketChannelNatives.accept(fd, true);
		}

		checkError(new_fd);
		StreamSocketChannel ssch = new StreamSocketChannel(new_fd);
		ssch.kind = Kind.SOCK_STREAM;
		return ssch;
	}

}
