/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2014-2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.is2t.support.net.natives;

import com.is2t.errors.NetErrors;


public class ChannelNatives {

	/**
	 * Initializes the TCP/IP stack components.
	 * 
	 * @return 0 on success or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int initialize();

	/**
	 * Returns option value for the socket associated with the file descriptor {@code fd}.
	 * 
	 * @param fd the socket file descriptor
	 * @param option the socket option identifier
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the option value for the socket associated with the file descriptor {@code fd} or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getOption(int fd, int option, boolean retry);

	/**
	 * Sets option value for the socket associated with the file descriptor {@code fd}.
	 * 
	 * @param fd the socket file descriptor
	 * @param option the socket option identifier
	 * @param value the option value to set
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 on success or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int setOption(int fd, int option, int value, boolean retry);

	/**
	 * Returns option value for the socket associated with the file descriptor {@code fd}.
	 * 
	 * @param fd the socket file descriptor
	 * @param option the socket option identifier
	 * @param dst the destination byte array for the option value
	 * @param dstLength the maximum length of the value the array can hold
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return the length of the option value for the socket associated with the file descriptor {@code fd} or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getOption(int fd, int option, byte[] dst, int dstLength, boolean retry);

	/**
	 * Sets option value for the socket associated with the file descriptor {@code fd}.
	 * 
	 * @param fd the socket file descriptor
	 * @param option the socket option identifier
	 * @param value the option value to set as a byte array
	 * @param valueLength the length of the value in the array
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 on success or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int setOption(int fd, int option, byte[] value, int valueLength, boolean retry);

	/**
	 * Binds the socket associated with the file descriptor {@code fd} to the specified local address {@code addr} and {@code port}.
	 * 
	 * @param fd the native file descriptor
	 * @param addr the IP address array, in network byte order
	 * @param length the address size (4 bytes long for IPv4 addresses and 16 bytes long for IPv6 addresses)
	 * @param port the port to bind to
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 on success or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int bind(int fd, byte[] addr, int length, int port, boolean retry);

	/**
	 * Configures the socket associated with the file descriptor {@code fd} in blocking or non-blocking mode.
	 * 
	 * @param fd the socket file descriptor
	 * @param blocking true for blocking mode, false for non-blocking mode
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 on success or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int setBlocking(int fd, boolean blocking, boolean retry);

	/**
	 * Closes the socket associated with the file descriptor {@code fd}.
	 * 
	 * @param fd the socket file descriptor
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 on success or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int closeNative(int fd, boolean retry);

	/**
	 * Shuts down the socket associated with the file descriptor {@code fd}'s input/output streams.
	 * 
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @param fd the socket file descriptor
	 * @return 0 on success or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int shutdown(int fd, boolean retry);

	/**
	 * Listens on the socket associated with the given file descriptor {@code fd} for incoming connections.
	 * 
	 * @param fd the socket file descriptor
	 * @param backlog the listen backlog size
	 * @param retry true when the previous call returned {@link NetErrors#J_NET_NATIVE_CODE_BLOCKED_WITHOUT_RESULT}
	 * and the calling process repeats the call to this operation for its completion
	 * @return 0 on success or a negative error code
	 * @see {@link NetErrors} for error codes
	 */
	public static native int listen(int fd, int backlog, boolean retry);
}
