/*
 * Copyright (C) 2017-2019, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security;

import java.security.KeyFactorySpi;
import java.security.MessageDigestSpi;
import java.security.SecureRandomSpi;
import java.security.SignatureSpi;

import javax.crypto.CipherSpi;
import javax.crypto.MacSpi;

import com.is2t.support.security.cipher.NativeCipherSpi;
import com.is2t.support.security.digest.NativeMessageDigestSpi;
import com.is2t.support.security.keyfactory.NativeKeyFactorySpi;
import com.is2t.support.security.mac.NativeMacSpi;
import com.is2t.support.security.securerandom.NativeSecureRandomSpi;
import com.is2t.support.security.signature.NativeSignatureSpi;

import ej.bon.ByteArray;
import ej.sni.SNI;

/**
 *
 */
public class NativeSecurityProvider implements SecurityProvider {

	/**
	 * Length of the string to provide for the message digest algorithm name.
	 */
	public static final int MESSAGE_DIGEST_ALGORITHM_NAME_SIZE = 16;


	@Override
	public CipherSpi getCipherSpi(String transformation) {

		byte[] cString = new byte[transformation.length() + 1];// add 1 for the '\0'
		SNI.toCString(transformation, cString);

		byte[] cipherDescription = new byte[NativeCipherSpi.CICPHER_DESCRIPTION_SIZE];

		int nativeTransformationId = NativeCipherSpi.nativeGetTransformationDescription(cString, cipherDescription);
		if (nativeTransformationId == -1) {
			return null;
		}

		int blockSize = ByteArray.readInt(cipherDescription, 0);
		int unitBytes = ByteArray.readInt(cipherDescription, ByteArray.INT_SIZE);
		int cipherMode = ByteArray.readInt(cipherDescription, ByteArray.INT_SIZE * 2);

		return new NativeCipherSpi(nativeTransformationId, blockSize, unitBytes, cipherMode);
	}

	@Override
	public SignatureSpi getSignatureSpi(String algorithm) {

		byte[] cAlgorithm = new byte[algorithm.length() + 1];// add 1 for the '\0'
		SNI.toCString(algorithm, cAlgorithm);

		byte[] cMessageDigestAlgorithmName = new byte[MESSAGE_DIGEST_ALGORITHM_NAME_SIZE];

		int nativeAlgorithmId = NativeSignatureSpi.nativeGetAlgorithmDescription(cAlgorithm,
				cMessageDigestAlgorithmName, cMessageDigestAlgorithmName.length);
		if (nativeAlgorithmId == -1) {
			return null;
		}

		String messageDigestAlgorithmName = SNI.toJavaString(cMessageDigestAlgorithmName);

		return new NativeSignatureSpi(nativeAlgorithmId, messageDigestAlgorithmName);
	}

	@Override
	public MessageDigestSpi getMessageDigestSpi(String algorithm) {

		byte[] cString = new byte[algorithm.length() + 1];// add 1 for the '\0'
		SNI.toCString(algorithm, cString);

		byte[] algorithmDescription = new byte[NativeMessageDigestSpi.ALGORITHM_DESCRIPTION_SIZE];

		int nativeAlgorithmId = NativeMessageDigestSpi.nativeGetAlgorithmDescription(cString, algorithmDescription);
		if (nativeAlgorithmId == -1) {
			return null;
		}

		int digestLength = ByteArray.readInt(algorithmDescription, 0);

		return new NativeMessageDigestSpi(nativeAlgorithmId, digestLength);
	}

	@Override
	public MacSpi getMacSpi(String algorithm) {

		byte[] cString = new byte[algorithm.length() + 1];// add 1 for the '\0'
		SNI.toCString(algorithm, cString);

		byte[] algorithmDescription = new byte[NativeMacSpi.ALGORITHM_DESCRIPTION_SIZE];

		int nativeAlgorithmId = NativeMacSpi.nativeGetAlgorithmDescription(cString, algorithmDescription);
		if (nativeAlgorithmId == -1) {
			return null;
		}

		int macLength = ByteArray.readInt(algorithmDescription, 0);

		return new NativeMacSpi(nativeAlgorithmId, macLength);
	}

	@Override
	public SecureRandomSpi getSecureRandomSpi() {
		return new NativeSecureRandomSpi();
	}

	@Override
	public KeyFactorySpi getKeyFactorySpi(String algorithm) {
		byte[] cString = new byte[algorithm.length() + 1]; // add 1 for the '\0'
		SNI.toCString(algorithm, cString);

		int nativeAlgorithmId = NativeKeyFactorySpi.nativeGetAlgorithmDescription(cString);
		if (nativeAlgorithmId == -1) {
			return null;
		}

		return new NativeKeyFactorySpi(nativeAlgorithmId, algorithm);
	}

}
