/*
 * Copyright (c) 2002, 2012, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security.mac;

import java.security.DigestException;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.ProviderException;
import java.security.spec.AlgorithmParameterSpec;

import javax.crypto.MacSpi;
import javax.crypto.SecretKey;

/**
 * This class constitutes the core of HMAC-<MD> algorithms, where
 * <MD> can be SHA1 or MD5, etc. See RFC 2104 for spec.
 *
 * It also contains the implementation classes for SHA-224, SHA-256,
 * SHA-384, and SHA-512 HMACs.
 *
 * Implementation copied from <code>com.sun.crypto.provider.HmacCore</code>.
 *
 * @author Jan Luehe
 */
public class HmacCore extends MacSpi {

	private MessageDigest md;
	private final byte[] k_ipad; // inner padding - key XORd with ipad
	private final byte[] k_opad; // outer padding - key XORd with opad
	private boolean first;       // Is this the first data to be processed?

	private final int blockLen;

	public HmacCore(String messageDigestAlgorithmName, int bl) {
		try {
			md = MessageDigest.getInstance(messageDigestAlgorithmName);
		} catch (NoSuchAlgorithmException e) {
			throw new ProviderException(e);
		}
		this.blockLen = bl;
		this.k_ipad = new byte[bl];
		this.k_opad = new byte[bl];
		this.first = true;
	}

	@Override
	protected int engineGetMacLength() {
		return this.md.getDigestLength();
	}

	@Override
	protected void engineInit(Key key, AlgorithmParameterSpec params) throws InvalidKeyException, InvalidAlgorithmParameterException {
		if (params != null) {
			throw new InvalidAlgorithmParameterException
			("HMAC does not use parameters");
		}

		if (!(key instanceof SecretKey)) {
			throw new InvalidKeyException("Secret key expected");
		}

		byte[] secret = key.getEncoded();
		if (secret == null) {
			throw new InvalidKeyException("Missing key data");
		}

		// if key is longer than the block length, reset it using
		// the message digest object.
		if (secret.length > blockLen) {
			byte[] tmp = md.digest(secret);
			// now erase the secret
			// TODO: use array util fonction to reset the array
			for(int i=secret.length ; --i>=0 ;) {
				secret[i] = 0;
			}
			secret = tmp;
		}

		// XOR k with ipad and opad, respectively
		for (int i = 0; i < blockLen; i++) {
			int si = (i < secret.length) ? secret[i] : 0;
			k_ipad[i] = (byte)(si ^ 0x36);
			k_opad[i] = (byte)(si ^ 0x5c);
		}

		// now erase the secret
		// TODO: use array util fonction to reset the array
		for(int i=secret.length ; --i>=0 ;) {
			secret[i] = 0;
		}
		secret = null;

		engineReset();

	}

	@Override
	protected void engineUpdate(byte input) {
		if (first == true) {
			// compute digest for 1st pass; start with inner pad
			md.update(k_ipad);
			first = false;
		}

		// add the passed byte to the inner digest
		md.update(input);
	}

	@Override
	protected void engineUpdate(byte[] input, int offset, int len) {
		if (first == true) {
			// compute digest for 1st pass; start with inner pad
			md.update(k_ipad);
			first = false;
		}

		// add the selected part of an array of bytes to the inner digest
		md.update(input, offset, len);
	}

	@Override
	protected byte[] engineDoFinal() {
		if (first == true) {
			// compute digest for 1st pass; start with inner pad
			md.update(k_ipad);
		} else {
			first = true;
		}

		try {
			// finish the inner digest
			byte[] tmp = md.digest();

			// compute digest for 2nd pass; start with outer pad
			md.update(k_opad);
			// add result of 1st hash
			md.update(tmp);

			md.digest(tmp, 0, tmp.length);
			return tmp;
		} catch (DigestException e) {
			// should never occur
			throw new ProviderException(e);
		}
	}

	@Override
	protected void engineReset() {
		if (first == false) {
			md.reset();
			first = true;
		}
	}

}
