/*
 * Copyright (C) 2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security.mac;

import java.security.InvalidKeyException;

import com.is2t.support.security.NativeResource;

import ej.bon.ByteArray;
import ej.sni.NativeException;

/**
 *
 */
public class NativeMacSpi extends AbstractMacSpi {

	/**
	 * @see #nativeGetAlgorithmDescription(byte[], byte[])
	 */
	public static final int ALGORITHM_DESCRIPTION_SIZE = ByteArray.INT_SIZE * 1;

	private final int nativeAlgorithmId;

	/**
	 * Native identifier for the cipher.
	 * Initialized during engine init.
	 */
	private int nativeId;

	/**
	 * Native resource related to the Mac. Initialized during engine init.
	 */
	private NativeMacResource nativeResource;

	/**
	 * @param macLength
	 */
	public NativeMacSpi(int nativeAlgorithmId, int macLength) {
		super(macLength);
		this.nativeAlgorithmId = nativeAlgorithmId;
		this.nativeId = -1;
	}


	@Override
	void implInit(byte[] secretKey) throws InvalidKeyException {
		try {

			if (nativeResource != null) {
				// close and reset the existing resource
				nativeResource.close();
				this.nativeId = -1;
				this.nativeResource = null;
			}

			int nativeAlgorithmId = this.nativeAlgorithmId;

			try {
				// Maybe some resources was garbage collected.
				// Try to release their related underlying native resources without trigger
				// the GC before.
				NativeResource.recycle(false);
				this.nativeId = nativeInit(nativeAlgorithmId, secretKey, secretKey.length);
				// the initialization succeeds; create the associated resource.
				this.nativeResource = new NativeMacResource(this, nativeAlgorithmId, this.nativeId);

			} catch (NativeException e) {
				// An error occurs during initialization, may be there is no space left to
				// allocate new native resource.
				// Trigger a new fresh GC and release new potential useless resources and try
				// the initialization again!
				NativeResource.recycle(true);
				this.nativeId = nativeInit(nativeAlgorithmId, secretKey, secretKey.length);
				// the initialization succeeds; create the associated resource.
				this.nativeResource = new NativeMacResource(this, nativeAlgorithmId, this.nativeId);
			}
		}
		catch(NativeException e) {
			throw new InvalidKeyException(e);
		}
	}


	@Override
	void implUpdate(byte[] b, int ofs, int len) {
		nativeUpdate(nativeAlgorithmId, nativeId, b, ofs, len);
	}

	@Override
	void implDoFinal(byte[] out, int ofs, int len) {
		nativeDoFinal(nativeAlgorithmId, nativeId, out, ofs, len);
	}

	@Override
	void implReset() {
		nativeReset(nativeAlgorithmId, nativeId);
	}

	/**
	 * Gets for the given algorithm the message digest description.
	 * <p>
	 * <code>description</code> must be filled-in with:
	 * <ul>
	 * 	<li>[0-3]: macLength: length of the message digest in bytes</li>
	 * </ul>
	 *
	 * @param algorithm Null terminated string that describes the algorithm
	 * @return algorithm ID on success or -1 on error.
	 */
	native public static int nativeGetAlgorithmDescription(byte[] algorithm, byte[] description);

	/**
	 * Throw NativeException on error.
	 * Returns a nativeId.
	 */
	native private static int nativeInit(int nativeAlgorithmId, byte[] secretKey, int secretKeyLength) throws NativeException;

	/**
	 * @param nativeAlgorithmId
	 * @param nativeId
	 *
	 * @throws {@link NativeException} on error
	 */
	native private static void nativeUpdate(int nativeAlgorithmId, int nativeId, byte[] buffer, int offset, int length) throws NativeException;

	/**
	 * @param nativeAlgorithmId
	 * @param nativeId
	 *
	 * @throws {@link NativeException} on error
	 */
	native private static void nativeDoFinal(int nativeAlgorithmId, int nativeId, byte[] out, int offset, int length) throws NativeException;

	/**
	 * @param nativeAlgorithmId
	 * @param nativeId
	 *
	 * @throws {@link NativeException} on error
	 */
	native private static void nativeReset(int nativeAlgorithmId, int nativeId) throws NativeException;

	/**
	 * @param nativeAlgorithmId
	 * @param nativeId
	 *
	 * @throws {@link
	 *             NativeException} on error
	 */
	native public static void nativeClose(int nativeAlgorithmId, int nativeId) throws NativeException;
}
