/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2019, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security.x509;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactorySpi;

public class X509Factory extends CertificateFactorySpi {

	@Override
	public Certificate engineGenerateCertificate(InputStream stream)
			throws CertificateException {
		if (stream == null) {
			throw new CertificateException("Missing input stream");
		}
		try {
			int c = stream.read();
			if (c == -1) {
				return null;
			}
			if (c == 0x30) { // 0x30 == DerValue.tag_Sequence
				ByteArrayOutputStream bout = new ByteArrayOutputStream(1024);
				bout.write(c);
				readBERInternal(stream, bout, c);
				return new X509CertImpl(bout.toByteArray());
			} else {
				byte[] data = new byte[256];
				ByteArrayOutputStream out = new ByteArrayOutputStream(256);
				try {
					// write the byte previously read to check the DER tag sequence
					out.write(c);

					/* get certificate data */
					int len = stream.read(data);
					if (len == -1) {
						throw new IOException("Empty input");
					}
					do {
						out.write(data, 0, len);
					} while ((len = stream.read(data)) > 0);

					// create a new X.509 certificate
					return new X509CertImpl(out.toByteArray());

				} catch (IOException ioe) {
					throw new CertificateException("Could not parse certificate", ioe);
				}
			}
		} catch (IOException ioe) {
			throw new CertificateException("Could not parse certificate", ioe);
		}
	}

	/**
	 * Read one BER data block. This method is aware of indefinite-length BER encoding and will read all of the
	 * sub-sections in a recursive way
	 *
	 * @param is
	 *            Read from this InputStream
	 * @param bout
	 *            Write into this OutputStream
	 * @param tag
	 *            Tag already read (-1 mean not read)
	 * @returns The current tag, used to check EOC in indefinite-length BER
	 * @throws IOException
	 *             Any parsing error
	 */
	private static int readBERInternal(InputStream is, ByteArrayOutputStream bout, int tag) throws IOException {

		if (tag == -1) { // Not read before the call, read now
			tag = is.read();
			if (tag == -1) {
				throw new IOException("BER/DER tag info absent");
			}
			if ((tag & 0x1f) == 0x1f) {
				throw new IOException("Multi octets tag not supported");
			}
			bout.write(tag);
		}

		int n = is.read();
		if (n == -1) {
			throw new IOException("BER/DER length info absent");
		}
		bout.write(n);

		int length;

		if (n == 0x80) { // Indefinite-length encoding
			if ((tag & 0x20) != 0x20) {
				throw new IOException("Non constructed encoding must have definite length");
			}
			while (true) {
				int subTag = readBERInternal(is, bout, -1);
				if (subTag == 0) { // EOC, end of indefinite-length section
					break;
				}
			}
		} else {
			if (n < 0x80) {
				length = n;
			} else if (n == 0x81) {
				length = is.read();
				if (length == -1) {
					throw new IOException("Incomplete BER/DER length info");
				}
				bout.write(length);
			} else if (n == 0x82) {
				int highByte = is.read();
				int lowByte = is.read();
				if (lowByte == -1) {
					throw new IOException("Incomplete BER/DER length info");
				}
				bout.write(highByte);
				bout.write(lowByte);
				length = (highByte << 8) | lowByte;
			} else if (n == 0x83) {
				int highByte = is.read();
				int midByte = is.read();
				int lowByte = is.read();
				if (lowByte == -1) {
					throw new IOException("Incomplete BER/DER length info");
				}
				bout.write(highByte);
				bout.write(midByte);
				bout.write(lowByte);
				length = (highByte << 16) | (midByte << 8) | lowByte;
			} else if (n == 0x84) {
				int highByte = is.read();
				int nextByte = is.read();
				int midByte = is.read();
				int lowByte = is.read();
				if (lowByte == -1) {
					throw new IOException("Incomplete BER/DER length info");
				}
				if (highByte > 127) {
					throw new IOException("Invalid BER/DER data (a little huge?)");
				}
				bout.write(highByte);
				bout.write(nextByte);
				bout.write(midByte);
				bout.write(lowByte);
				length = (highByte << 24) | (nextByte << 16) | (midByte << 8) | lowByte;
			} else { // ignore longer length forms
				throw new IOException("Invalid BER/DER data (too huge?)");
			}
			if (readFully(is, bout, length) != length) {
				throw new IOException("Incomplete BER/DER data");
			}
		}
		return tag;
	}

	/**
	 * Read from the stream until length bytes have been read or EOF has been reached. Return the number of bytes
	 * actually read.
	 */
	private static int readFully(InputStream in, ByteArrayOutputStream bout, int length) throws IOException {
		int read = 0;
		byte[] buffer = new byte[2048];
		while (length > 0) {
			int n = in.read(buffer, 0, length < 2048 ? length : 2048);
			if (n <= 0) {
				break;
			}
			bout.write(buffer, 0, n);
			read += n;
			length -= n;
		}
		return read;
	}

}
