/*
 * Java
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.verisilicon.vivante;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.LinearGradientPaint;
import java.awt.MultipleGradientPaint;
import java.awt.Paint;
import java.awt.RenderingHints;
import java.awt.geom.GeneralPath;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;

import com.is2t.hil.HIL;
import com.is2t.hil.NativeInterface;
import com.verisilicon.vivante.path.PathParserBase;

import ej.microui.display.LLUIDisplay;
import ej.microui.display.LLUIPainter.MicroUIGraphicsContext;

/**
 * This class provides a simulation of the Verisilicon VGLite APIs on the front-panel
 */
public class Nvg_lite {

	/**
	 * Non-zero fill rule. A pixel is drawn if it crosses at least one path pixel.
	 * <p>
	 * For drawing any path, the hardware supports both non-zero and odd-even fill rules.
	 * <p>
	 * To determine whether any point is contained inside an object, imagine drawing a line from that point out to
	 * infinity in any direction such that the line does not cross any vertex of the path. For each edge that is crossed
	 * by the line, add 1 to the counter if the edge crosses from left to right, as seen by an observer walking across
	 * the line towards infinity, and subtract 1 if the edge crosses from right to left. In this way, each region of the
	 * plane will receive an integer value.
	 * <p>
	 * The non-zero fill rule says that a point is inside the shape if the resulting sum is not equal to zero. The
	 * even/odd rule says that a point is inside the shape if the resulting sum is odd, regardless of sign. See
	 */
	public static final int VG_LITE_FILL_NON_ZERO = 0;

	/**
	 * Even-odd fill rule. A pixel is drawn it it crosses an odd number of path pixels.
	 *
	 * @see #VG_LITE_FILL_NON_ZERO
	 */
	public static final int VG_LITE_FILL_EVEN_ODD = 1;

	/**
	 * Blending mode: "S, i.e. no blending."
	 * <p>
	 * Some of the VGLite API functions calls support blending. S and D represent source and destination color channels
	 * and Sa and Da represent the source and destination alpha channels.
	 */
	public static final int VG_LITE_BLEND_NONE = 0;

	/**
	 * Blending mode: "S + (1 - Sa) * D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_SRC_OVER = 1;

	/**
	 * Blending mode: "(1 - Da) * S + D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_DST_OVER = 2;

	/**
	 * Blending mode: "Da * S"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_SRC_IN = 3;

	/**
	 * Blending mode: "Sa * D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_DST_IN = 4;

	/**
	 * Blending mode: "S + D - S * D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_SCREEN = 5;

	/**
	 * Blending mode: "S * (1 - Da) + D * (1 - Sa) + S * D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_MULTIPLY = 6;

	/**
	 * Blending mode: "S + D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_ADDITIVE = 7;

	/**
	 * Blending mode: "D * (1 - S)"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_SUBTRACT = 8;

	private static final Point2D ORIGIN = new Point2D.Float(0f, 0f);
	private static final Point2D GRADIENT_END = new Point2D.Float(255f, 0f);

	/**
	 * Draw a path to a target buffer.
	 * <p>
	 * The specified path will be transformed by the given matrix and drawn into the specified target buffer using the
	 * supplied color. Blending can be specified.
	 * <p>
	 * See <code>vg_lite_draw_path</code> C function. Target buffer is managed in C.
	 *
	 * @param sd
	 *            The byte array where to draw.
	 * @param pathLeft
	 *            The path to draw bounding_box left coordinate.
	 * @param pathTop
	 *            The path to draw bounding_box top coordinate.
	 * @param pathRight
	 *            The path to draw bounding_box right coordinate.
	 * @param pathBottom
	 *            The path to draw bounding_box bottom coordinate.
	 * @param path
	 *            Array containing the physical description of the path.
	 * @param pathOffset
	 *            The start index in the path array.
	 * @param pathLength
	 *            Number of bytes in the path data.
	 * @param pathQuality
	 *            Quality hint for the path.
	 * @param pathFormat
	 *            Coordinate format.
	 * @param fillRule
	 *            Fill rule for the path: {@link #VG_LITE_FILL_NON_ZERO} or {@link #VG_LITE_FILL_EVEN_ODD}.
	 * @param matrixHandle
	 *            The matrix that defines the transformation matrix of the path.
	 * @param blend
	 *            The blending mode to be applied to each drawn pixel. If no blending is required, set this value to
	 *            {@link #VG_LITE_BLEND_NONE}.
	 * @param color
	 *            The color is formed using 8-bit ARGB channels. The blue channel is in the lower 8-bit of the color
	 *            value, followed by the green and red channels. The alpha channel is in the upper 8-bit of the color
	 *            value.
	 * @return vg_lite_error An explicit exception in case an error occurred.
	 */
	public static int vg_lite_draw_path(byte[] sd, float pathLeft, float pathTop, float pathRight, float pathBottom,
			byte[] path, int pathOffset, int pathLength, int pathQuality, int pathFormat, int fillRule,
			int matrixHandle, int blend, int color) {
		try {
			return drawPath(sd, pathLeft, pathTop, pathRight, pathBottom, path, pathOffset, pathLength, pathQuality,
					pathFormat, fillRule, matrixHandle, blend, color);
		} catch (Throwable e) {
			// TODO manage error codes
			return -1;
		}
	}

	private static int drawPath(byte[] sd, float pathLeft, float pathTop, float pathRight, float pathBottom,
			byte[] path, int pathOffset, int pathLength, int pathQuality, int pathFormat, int fillRule,
			int matrixHandle, int blend, int color) {

		LLUIDisplay engine = LLUIDisplay.Instance;

		synchronized (engine) {

			MicroUIGraphicsContext gc = engine.mapMicroUIGraphicsContext(sd);
			if (gc.requestDrawing()) {

				// retrieve destination
				BufferedImage image = (BufferedImage) gc.getImage().getRAWImage();
				Graphics2D gc2d = image.createGraphics();

				// apply rendering hints
				RenderingHints rh = new RenderingHints(RenderingHints.KEY_ANTIALIASING,
						RenderingHints.VALUE_ANTIALIAS_ON);
				gc2d.setRenderingHints(rh);

				// apply color and composite
				setComposite(gc2d, blend);
				gc2d.setColor(new Color(getDisplayColor(color), true));

				// create VG path on user VG path
				refreshContent(path, pathOffset, pathLength);
				GeneralPath p = PathParserBase.get(path, pathOffset, pathLength, pathFormat);
				p.setWindingRule(getFillRule(fillRule));

				// apply VG path
				Matrix matrix = Nvg_lite_matrix.get(matrixHandle);
				matrix.transform(p);

				// Set the clip
				setClip(gc2d, gc, pathLeft, pathTop, pathRight, pathBottom, matrix);

				gc2d.fill(p);

				// convert colors in drawn area
				final Rectangle2D bounds = gc2d.getClipBounds();
				convertColorToColorToDraw(image, (int) bounds.getMinX(), (int) bounds.getMinY(),
						(int) bounds.getWidth(), (int) bounds.getHeight());
			}
		}

		return 0;
	}

	private static int getFillRule(int fillRule) {
		switch (fillRule) {
		case VG_LITE_FILL_EVEN_ODD:
			return Path2D.WIND_EVEN_ODD;
		case VG_LITE_FILL_NON_ZERO:
		default:
			return Path2D.WIND_NON_ZERO;
		}
	}

	/**
	 * Clips the Graphics2D from the current MicroEJ GraphicsContext current clip and update drawing flush limits.
	 * <p>
	 * This code performs the same work than vg_lite library for embedded. It uses an orthogonal clip.
	 *
	 * @param gc2d
	 *            Graphics2D drawing object
	 * @param gc
	 *            MicroUI GraphicsContext where performing the drawing
	 * @param pathLeft
	 *            left bound of the path
	 * @param pathTop
	 *            top bound of the path
	 * @param pathRight
	 *            right bound of the path
	 * @param pathBottom
	 *            bottom bound of the path
	 * @param matrix
	 *            3x3 matrix that defines the transformation matrix of path
	 */
	private static void setClip(Graphics2D gc2d, MicroUIGraphicsContext gc, float pathLeft, float pathTop,
			float pathRight, float pathBottom, Matrix matrix) {

		Point2D min = new Point2D.Double();
		matrix.transform(new Point2D.Double(pathLeft, pathTop), min);
		Point2D max = (Point2D) min.clone();

		updateBounds(pathRight, pathTop, matrix, min, max);
		updateBounds(pathRight, pathBottom, matrix, min, max);
		updateBounds(pathLeft, pathBottom, matrix, min, max);
		updateBounds(pathRight, pathTop, matrix, min, max);

		int clipX1 = Math.max((int) min.getX(), gc.getClipX1());
		int clipY1 = Math.max((int) min.getY(), gc.getClipY1());
		int clipX2 = Math.min((int) max.getX(), gc.getClipX2());
		int clipY2 = Math.min((int) max.getY(), gc.getClipY2());

		gc2d.clipRect(clipX1, clipY1, clipX2 - clipX1 + 1, clipY2 - clipY1 + 1);

		// update rendering bounds
		gc.setDrawingLimits(clipX1, clipY1, clipX2, clipY2);
	}

	private static void updateBounds(float x, float y, Matrix matrix, Point2D min, Point2D max) {
		Point2D temp = new Point2D.Double();
		matrix.transform(new Point2D.Double(x, y), temp);

		double minX = min.getX();
		double minY = min.getY();
		double maxX = max.getX();
		double maxY = max.getY();

		if (temp.getX() < minX) {
			minX = temp.getX();
		}
		if (temp.getY() < minY) {
			minY = temp.getY();
		}
		if (temp.getX() > maxX) {
			maxX = temp.getX();
		}
		if (temp.getY() > maxY) {
			maxY = temp.getY();
		}

		min.setLocation(minX, minY);
		max.setLocation(maxX, maxY);
	}

	/**
	 * Refresh a range of an array; the array may be immortal or not.
	 *
	 * @param array
	 *            the array to refresh
	 * @param offset
	 *            the start index of the range.
	 * @param length
	 *            the length of the range.
	 * @return true if the array is immortal
	 */
	private static boolean refreshContent(byte[] array, int offset, int length) {
		NativeInterface hil = HIL.getInstance();
		boolean isImmortal = hil.isImmortal(array);
		if (isImmortal) {
			hil.refreshImmortalContent(array, offset, length);
		} else {
			hil.refreshContent(array, offset, length);
		}
		return isImmortal;
	}

	/**
	 * Refresh a rangee of an array; the array may be immortal or not.
	 *
	 * @param array
	 *            the array to refresh
	 * @param offset
	 *            the start index of the range.
	 * @param length
	 *            the length of the range.
	 * @return true if the array is immortal
	 */
	private static boolean refreshContent(int[] array, int offset, int length) {
		NativeInterface hil = HIL.getInstance();
		boolean isImmortal = hil.isImmortal(array);
		if (isImmortal) {
			hil.refreshImmortalContent(array, offset, length);
		} else {
			hil.refreshContent(array, offset, length);
		}
		return isImmortal;
	}

	/**
	 * Prints a warning message and sets the default graphics composite SRC_OVER.
	 *
	 * @param gc2d
	 *            the AWT graphics
	 * @param blendLabel
	 *            the name of the unsupported blend algorithm
	 */
	private static void setUnsupportedComposite(Graphics2D gc2d, String blendLabel) {
		System.err.println("Warning: unsupported blend mode: " + blendLabel + ", using BLEND_SRC_OVER instead");
		gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
	}

	/**
	 * Set graphics composite according user blending value.
	 *
	 * @param gc2d
	 *            the AWT graphics
	 * @param blend
	 *            the user blending value
	 * @throws IllegalArgumentException
	 *             if user blending value is not an expected value
	 */
	private static void setComposite(Graphics2D gc2d, int blend) {
		switch (blend) {
		case VG_LITE_BLEND_NONE:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC));
			break;
		case VG_LITE_BLEND_SRC_OVER:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
			break;
		case VG_LITE_BLEND_DST_OVER:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.DST_OVER));
			break;
		case VG_LITE_BLEND_SRC_IN:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_IN));
			break;
		case VG_LITE_BLEND_DST_IN:
			gc2d.setComposite(AlphaComposite.getInstance(AlphaComposite.DST_IN));
			break;
		case VG_LITE_BLEND_SCREEN:
			setUnsupportedComposite(gc2d, "BLEND_SCREEN");
			break;
		case VG_LITE_BLEND_MULTIPLY:
			setUnsupportedComposite(gc2d, "BLEND_MULTIPLY");
			break;
		case VG_LITE_BLEND_ADDITIVE:
			setUnsupportedComposite(gc2d, "BLEND_ADDITIVE");
			break;
		default:
			throw new IllegalArgumentException("Unsupported blend mode: " + blend);
		}
	}

	/**
	 * Fill a path with a gradient pattern.
	 * <p>
	 * The specified path will be transformed by the given matrix and filled by the tranformed gradient pattern.
	 * <p>
	 * See <code>vg_lite_draw_gradient</code> C function. Target buffer is managed in C.
	 *
	 * @param gcArray
	 *            The byte array where to draw.
	 * @param pathLeft
	 *            The path to draw bounding_box left coordinate.
	 * @param pathTop
	 *            The path to draw bounding_box top coordinate.
	 * @param pathRight
	 *            The path to draw bounding_box right coordinate.
	 * @param pathBottom
	 *            The path to draw bounding_box bottom coordinate.
	 * @param path
	 *            Array containing the physical description of the path.
	 * @param pathOffset
	 *            The start index in the path array.
	 * @param pathLength
	 *            Number of bytes in the path data.
	 * @param pathQuality
	 *            Quality hint for the path.
	 * @param pathFormat
	 *            Coordinate format.
	 * @param fillRule
	 *            Fill rule for the path: {@link #VG_LITE_FILL_NON_ZERO} or {@link #VG_LITE_FILL_EVEN_ODD}.
	 * @param matrixHandle
	 *            The matrix that defines the transformation matrix of the path.
	 * @param gradientHandle
	 *            The gradient that will be filled the path with.
	 * @param blend
	 *            The blending mode to be applied to each drawn pixel. If no blending is required, set this value to
	 *            {@link #VG_LITE_BLEND_NONE}.
	 * @return vg_lite_error An explicit exception in case an error occurred
	 */
	public static int vg_lite_draw_gradient(byte[] gcArray, float pathLeft, float pathTop, float pathRight,
			float pathBottom, byte[] path, int pathOffset, int pathLength, int pathQuality, int pathFormat,
			int fillRule, int matrixHandle, int gradientHandle, int blend) {
		try {
			return drawGradient(gcArray, pathLeft, pathTop, pathRight, pathBottom, path, pathOffset, pathLength,
					pathQuality, pathFormat, fillRule, matrixHandle, gradientHandle, blend);
		} catch (Throwable e) {
			// TODO manage error codes
			return -1;
		}
	}

	private static int drawGradient(byte[] gcArray, float pathLeft, float pathTop, float pathRight, float pathBottom,
			byte[] path, int pathOffset, int pathLength, int pathQuality, int pathFormat, int fillRule,
			int matrixHandle, int gradientHandle, int blend) {

		LLUIDisplay engine = LLUIDisplay.Instance;
		synchronized (engine) {
			MicroUIGraphicsContext graphicContext = engine.mapMicroUIGraphicsContext(gcArray);
			if (graphicContext.requestDrawing()) {

				final Gradient gradient = Nvg_lite_gradient.get(gradientHandle);

				final Point2D gradientOrigin = new Point2D.Float();
				final Point2D gradientEnd = new Point2D.Float();

				final Matrix gradientMatrix = gradient.getMatrix();
				gradientMatrix.transform(ORIGIN, gradientOrigin);
				gradientMatrix.transform(GRADIENT_END, gradientEnd);

				final MultipleGradientPaint gradientPaint = new LinearGradientPaint(gradientOrigin, gradientEnd,
						gradient.getFractions(), gradient.getColors());

				BufferedImage bi = (BufferedImage) graphicContext.getImage().getRAWImage();
				Graphics2D g2d = (Graphics2D) bi.getGraphics();

				final Paint originalPaint = g2d.getPaint();
				g2d.setPaint(gradientPaint);

				// apply rendering hints
				RenderingHints rh = new RenderingHints(RenderingHints.KEY_ANTIALIASING,
						RenderingHints.VALUE_ANTIALIAS_ON);
				g2d.setRenderingHints(rh);
				setComposite(g2d, blend);

				// create VG path on user VG path
				refreshContent(path, pathOffset, pathLength);
				GeneralPath p = PathParserBase.get(path, pathOffset, pathLength, pathFormat);

				// apply VG path
				Matrix matrix = Nvg_lite_matrix.get(matrixHandle);
				matrix.transform(p);

				// Set the clip
				setClip(g2d, graphicContext, pathLeft, pathTop, pathRight, pathBottom, matrix);

				// paint
				g2d.fill(p);

				// convert colors in drawn area
				final Rectangle2D bounds = g2d.getClipBounds();
				convertColorToColorToDraw(bi, (int) bounds.getMinX(), (int) bounds.getMinY(), (int) bounds.getWidth(),
						(int) bounds.getHeight());

				// reset paint
				g2d.setPaint(originalPaint);
			}
		}
		return 0;
	}

	/**
	 * Initializes the gradient to its default settings.
	 * <p>
	 * See <code>vg_lite_init_grad</code> C function.
	 *
	 * @param handle
	 *            gradient to initialize
	 * @return 0
	 */
	public static int vg_lite_init_grad(int handle) {
		// Nothing to do on simulator
		return 0;
	}

	/**
	 * Sets the values for the members of the gradient definition.
	 * <p>
	 * See <code>vg_lite_set_grad</code> C function.
	 *
	 * @param handle
	 *            gradient to set
	 * @param count
	 *            The count of the colors in gradient.
	 * @param colors
	 *            A color is formed using 8-bit ARGB channels. The blue channel is in the lower 8-bit of the color
	 *            value, followed by the green and red channels. The alpha channel is in the upper 8-bit of the color
	 *            value.
	 * @param stops
	 *            Numbers ranging from 0 to 255 specifying the distribution of colors along the gradient.
	 * @return 0
	 */
	public static int vg_lite_set_grad(int handle, int count, int[] colors, int[] stops) {
		refreshContent(colors, 0, count);
		refreshContent(stops, 0, count);

		final Gradient gradient = Nvg_lite_gradient.get(handle);
		gradient.setFractions(stops);
		gradient.setColors(colors);

		return 0;
	}

	/**
	 * Updates or generates the corresponding gradient to render with.
	 * <p>
	 * The gradient has an internal buffer which is used to render the gradient pattern. The buffer will be
	 * created/updated by the corresponding gradient parameters.
	 * <p>
	 * See <code>vg_lite_update_grad</code> C function.
	 *
	 * @param handle
	 *            gradient to update
	 * @return 0
	 */
	public static int vg_lite_update_grad(int handle) {
		// Nothing to do on simulator
		return 0;
	}

	/**
	 * Convert RGB color to RGB color corresponding to the display configuration.
	 *
	 * @param color
	 *            RGB color code to convert.
	 * @return Converted RGB color
	 */
	public static int getDisplayColor(int color) {
		int alpha = (color & 0xFF000000);
		int rgb = (LLUIDisplay.Instance.convertARGBColorToColorToDraw(color) & 0x00FFFFFF);
		return (alpha | rgb);
	}

	/**
	 * Converts all pixels colors to Display Color in given region.
	 */
	private static void convertColorToColorToDraw(BufferedImage bi, int x, int y, int width, int height) {

		// (width, height) represents first outside pixel
		width += x;
		height += y;

		// sanity check
		x = Math.max(x, 0);
		y = Math.max(y, 0);
		width = Math.min(width, bi.getWidth());
		height = Math.min(height, bi.getHeight());

		for (int i = x; i < width; i++) {
			for (int j = y; j < height; j++) {
				bi.setRGB(i, j, LLUIDisplay.Instance.convertARGBColorToColorToDraw(bi.getRGB(i, j)));
			}
		}
	}
}
