/*
 * Java
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.verisilicon.vivante.path;

import java.awt.geom.GeneralPath;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * Expected to be an immutable so that it can be passed as-is to native side
 */
public abstract class PathParserBase {

	private static final int VG_LITE_S8 = 0;
	private static final int VG_LITE_S16 = 1;
	private static final int VG_LITE_S32 = 2;
	private static final int VG_LITE_FP32 = 3;

	private static final int CMD_MOVE = 2;
	private static final int CMD_MOVE_REL = 3;
	private static final int CMD_LINE = 4;
	private static final int CMD_LINE_REL = 5;
	private static final int CMD_QUAD = 6;
	private static final int CMD_QUAD_REL = 7;
	private static final int CMD_CUBIC = 8;
	private static final int CMD_CUBIC_REL = 9;

	/**
	 * get vglite command from path
	 *
	 * @param b
	 *            ByteBuffer containing the path
	 * @return next command
	 */
	protected abstract double getCommand(ByteBuffer b);

	/**
	 * get a point coordinate from path
	 *
	 * @param b
	 *            ByteBuffer containing the path
	 * @return next coordinate
	 */
	protected abstract double getCoordinate(ByteBuffer b);

	/**
	 * Compute VGlite path to a GeneralPath
	 *
	 * @param pathData
	 *            VGlite path
	 *
	 * @param pathOffset
	 *            offset on the path
	 *
	 * @param pathLength
	 *            path length
	 *
	 * @return GeneralPath
	 */
	protected GeneralPath toPath(byte[] pathData, int pathOffset, int pathLength) {
		double xPos = 0;
		double yPos = 0;

		GeneralPath p = new GeneralPath();

		ByteBuffer b = ByteBuffer.wrap(pathData, pathOffset, pathLength);
		b.order(ByteOrder.LITTLE_ENDIAN);

		ByteBuffer buffer = ByteBuffer.wrap(pathData).order(ByteOrder.LITTLE_ENDIAN);

		try {
			double cmd;

			while (true) {
				cmd = getCommand(buffer);

				if (cmd == 0) {
					break;
				}

				switch ((int) cmd) {
				case CMD_MOVE: {
					xPos = getCoordinate(buffer);
					yPos = getCoordinate(buffer);

					p.moveTo(xPos, yPos);
					break;
				}
				case CMD_MOVE_REL: {
					xPos += getCoordinate(buffer);
					yPos += getCoordinate(buffer);

					p.moveTo(xPos, yPos);
					break;
				}
				case CMD_LINE: {
					xPos = getCoordinate(buffer);
					yPos = getCoordinate(buffer);

					p.lineTo(xPos, yPos);
					break;
				}
				case CMD_LINE_REL: {
					xPos += getCoordinate(buffer);
					yPos += getCoordinate(buffer);

					p.lineTo(xPos, yPos);
					break;
				}
				case CMD_QUAD: {
					double cx = getCoordinate(buffer);
					double cy = getCoordinate(buffer);
					xPos = getCoordinate(buffer);
					yPos = getCoordinate(buffer);

					p.quadTo(cx, cy, xPos, yPos);
					break;
				}
				case CMD_QUAD_REL: {
					double cx = xPos + getCoordinate(buffer);
					double cy = yPos + getCoordinate(buffer);
					xPos += getCoordinate(buffer);
					yPos += getCoordinate(buffer);

					p.quadTo(cx, cy, xPos, yPos);
					break;
				}
				case CMD_CUBIC: {
					double c1x = getCoordinate(buffer);
					double c1y = getCoordinate(buffer);
					double c2x = getCoordinate(buffer);
					double c2y = getCoordinate(buffer);
					double x = getCoordinate(buffer);
					double y = getCoordinate(buffer);

					// Path debug
					// Prints the command formulae that can be used with a plotter
					// System.out.println(String.format("(%f, %f)", xPos, yPos));
					// System.out.println(String.format("(%f, %f)", c1x, c1y));
					// System.out.println(String.format("(%f, %f)", c2x, c2y));
					// System.out.println(String.format("(%f, %f)", x, y));
					// System.out.println(String.format("((1-t)*%f+t*%f,(1-t)*%f+t*%f)", xPos, c1x,
					// yPos, c1y));
					// System.out.println(String.format("((1-t)*%f+t*%f,(1-t)*%f+t*%f)", c1x, c2x,
					// c1y, c2y));
					// System.out.println(String.format("((1-t)*%f+t*%f,(1-t)*%f+t*%f)", c2x, x,
					// c2y, y));
					// System.out.println(String.format("("
					// +
					// "(1-t)((1-t)((1-t)*%f+%f*t)+t((1-t)*%f+%f*t))+t((1-t)((1-t)*%f+%ft)+t((1-t)*%f+%f*t)),"
					// +
					// "(1-t)((1-t)((1-t)*%f+%f*t)+t((1-t)*%f+%f*t))+t((1-t)((1-t)*%f+%ft)+t((1-t)*%f+%f*t))"
					// + ")", xPos, c1x, c1x, c2x, c1x, c2x, c2x, x, yPos, c1y, c1y, c2y, c1y, c2y,
					// c2y, y));

					p.curveTo(c1x, c1y, c2x, c2y, x, y);

					xPos = x;
					yPos = y;

					break;
				}
				case CMD_CUBIC_REL: {
					double c1x = xPos + getCoordinate(buffer);
					double c1y = yPos + getCoordinate(buffer);
					double c2x = xPos + getCoordinate(buffer);
					double c2y = yPos + getCoordinate(buffer);
					xPos += getCoordinate(buffer);
					yPos += getCoordinate(buffer);

					p.curveTo(c1x, c1y, c2x, c2y, xPos, yPos);
					break;
				}
				default: {
					System.out.println("Unsupported command: " + cmd);
				}
				}
			}
		} catch (java.nio.BufferUnderflowException e) {

		}

		p.closePath();

		return p;
	}

	/**
	 * Compute VGlite path to a GeneralPath with a specific parser
	 *
	 * @param pathData
	 *            VGlite path
	 *
	 * @param pathOffset
	 *            offset on the path
	 *
	 * @param pathLength
	 *            path length
	 * @param pathFormat
	 *            path format
	 * @return GeneralPath
	 */
	public static GeneralPath get(byte[] pathData, int pathOffset, int pathLength, int pathFormat) {
		PathParserBase parser;

		switch (pathFormat) {
		case VG_LITE_S8:
			parser = PathParserS8.getInstance();
			break;
		case VG_LITE_S16:
			parser = PathParserS16.getInstance();
			break;
		case VG_LITE_S32:
			parser = PathParserS32.getInstance();
			break;
		case VG_LITE_FP32:
			parser = PathParserFP32.getInstance();
			break;
		default:
			throw new IllegalArgumentException("Unkown format: " + pathFormat);
		}

		return parser.toPath(pathData, pathOffset, pathLength);
	}
}
