/*
 * Java
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.verisilicon.vivante;

import ej.microui.display.GraphicsContext;

/**
 * Provides access to vg_lite native APIs.
 * <p>
 * see vg_lite.h for more information.
 */
public class vg_lite {

	/**
	 * Non-zero fill rule. A pixel is drawn if it crosses at least one path pixel.
	 * <p>
	 * For drawing any path, the hardware supports both non-zero and odd-even fill rules.
	 * <p>
	 * To determine whether any point is contained inside an object, imagine drawing a line from that point out to
	 * infinity in any direction such that the line does not cross any vertex of the path. For each edge that is crossed
	 * by the line, add 1 to the counter if the edge crosses from left to right, as seen by an observer walking across
	 * the line towards infinity, and subtract 1 if the edge crosses from right to left. In this way, each region of the
	 * plane will receive an integer value.
	 * <p>
	 * The non-zero fill rule says that a point is inside the shape if the resulting sum is not equal to zero. The
	 * even/odd rule says that a point is inside the shape if the resulting sum is odd, regardless of sign. See
	 */
	public static final int VG_LITE_FILL_NON_ZERO = 0;

	/**
	 * Even-odd fill rule. A pixel is drawn it it crosses an odd number of path pixels.
	 *
	 * @see #VG_LITE_FILL_NON_ZERO
	 */
	public static final int VG_LITE_FILL_EVEN_ODD = 1;

	/**
	 * Blending mode: "S, i.e. no blending."
	 * <p>
	 * Some of the VGLite API functions calls support blending. S and D represent source and destination color channels
	 * and Sa and Da represent the source and destination alpha channels.
	 */
	public static final int VG_LITE_BLEND_NONE = 0;

	/**
	 * Blending mode: "S + (1 - Sa) * D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_SRC_OVER = 1;

	/**
	 * Blending mode: "(1 - Da) * S + D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_DST_OVER = 2;

	/**
	 * Blending mode: "Da * S"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_SRC_IN = 3;

	/**
	 * Blending mode: "Sa * D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_DST_IN = 4;

	/**
	 * Blending mode: "S + D - S * D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_SCREEN = 5;

	/**
	 * Blending mode: "S * (1 - Da) + D * (1 - Sa) + S * D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_MULTIPLY = 6;

	/**
	 * Blending mode: "S + D"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_ADDITIVE = 7;

	/**
	 * Blending mode: "D * (1 - S)"
	 *
	 * @see #VG_LITE_BLEND_NONE
	 */
	public static final int VG_LITE_BLEND_SUBTRACT = 8;

	/**
	 * Fill a (partial) buffer with a specified color.
	 * <p>
	 * Either an entire buffer or a partial rectangle of a buffer will be filled with a specific color.
	 * <p>
	 * See <code>vg_lite_clear</code> C function. Target buffer is managed in C.
	 *
	 * @param x
	 *            The x coordinate of the rectangle.
	 * @param y
	 *            The y coordinate of the rectangle.
	 * @param width
	 *            The width of the rectangle.
	 * @param height
	 *            The height of the rectangle.
	 * @param color
	 *            The color is formed using 8-bit ARGB channels. The blue channel is in the lower 8-bit of the color
	 *            value, followed by the green and red channels. The alpha channel is in the upper 8-bit of the color
	 *            value.
	 * @throws vg_lite_error
	 *             An explicit exception in case an error occurred
	 */
	public static void vg_lite_clear(int x, int y, int width, int height, int color) throws vg_lite_error {
		throw new RuntimeException();
	}

	/**
	 * Draw a path to a target buffer.
	 * <p>
	 * The specified path will be transformed by the given matrix and drawn into the specified target buffer using the
	 * supplied color. Blending can be specified.
	 * <p>
	 * See <code>vg_lite_draw_path</code> C function. Target buffer is managed in C.
	 *
	 * @param gc
	 *            The instance of GraphicsContext describing the destination where to draw.
	 * @param path
	 *            The path to draw.
	 * @param fill_rule
	 *            Fill rule for the path: {@link #VG_LITE_FILL_NON_ZERO} or {@link #VG_LITE_FILL_EVEN_ODD}.
	 * @param matrix
	 *            The matrix that defines the transformation matrix of the path.
	 * @param blend
	 *            The blending mode to be applied to each drawn pixel. If no blending is required, set this value to
	 *            {@link #VG_LITE_BLEND_NONE}.
	 * @param color
	 *            The color is formed using 8-bit ARGB channels. The blue channel is in the lower 8-bit of the color
	 *            value, followed by the green and red channels. The alpha channel is in the upper 8-bit of the color
	 *            value.
	 * @throws vg_lite_error
	 *             An explicit exception in case an error occurred.
	 */
	public static void vg_lite_draw_path(GraphicsContext gc, vg_lite_path path, int fill_rule, vg_lite_matrix matrix,
			int blend, int color) throws vg_lite_error {
		throw new RuntimeException();
	}

	/**
	 * Fill a path with a gradient pattern.
	 * <p>
	 * The specified path will be transformed by the given matrix and filled by the tranformed gradient pattern.
	 * <p>
	 * See <code>vg_lite_draw_gradient</code> C function. Target buffer is managed in C.
	 *
	 * @param gc
	 *            the instance of GraphicsContext describing the destination where to draw
	 * @param path
	 *            The path to draw.
	 * @param fill_rule
	 *            Fill rule for the path: {@link #VG_LITE_FILL_NON_ZERO} or {@link #VG_LITE_FILL_EVEN_ODD}.
	 * @param matrix
	 *            The matrix that defines the transformation matrix of the path.
	 * @param gradient
	 *            The gradient that will be filled the path with.
	 * @param blend
	 *            The blending mode to be applied to each drawn pixel. If no blending is required, set this value to
	 *            {@link #VG_LITE_BLEND_NONE}.
	 * @throws vg_lite_error
	 *             An explicit exception in case an error occurred
	 */
	public static void vg_lite_draw_gradient(GraphicsContext gc, vg_lite_path path, int fill_rule,
			vg_lite_matrix matrix, vg_lite_gradient gradient, int blend) throws vg_lite_error {
		throw new RuntimeException();
	}
}
