/*
 * Java
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.vector.font;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.LinearGradientPaint;
import java.awt.geom.Point2D;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * A linear multiple gradient.
 *
 * This is a Java representation of the gradient loaded in C native code. A new {@link LinearGradient} cannot be
 * directly instantiated calling its constructor. New instances are created when native method
 * {@link #initNative(int, int, int, int, int[], int[])} is called from FreeType API. Gradients can then be applied
 * calling static method {@link #setPaint(int, Graphics2D)}.
 */
public class LinearGradient {

	private static final Map<Integer, LinearGradient> gradients = new ConcurrentHashMap<>();

	private static int nextHandle;

	private final int handle;

	private LinearGradientPaint inner;

	/**
	 * Initializes a {@link LinearGradient}.
	 *
	 * This is the front panel implementation of native method defined in FreeType API. It instantiate a new
	 * {@link LinearGradient} at each call and stores it in a map. The gradient can then be retrieved from its handle.
	 *
	 * @param startX
	 *            the x coordinate of the start of the gradient
	 * @param startY
	 *            the y coordinate of the start of the gradient
	 * @param endX
	 *            the x coordinate of the end of the gradient
	 * @param endY
	 *            the y coordinate of the end of the gradient
	 * @param fractions
	 *            the colors distribution along the gradient from 0 (start) to 255 (end)
	 * @param colors
	 *            the colors in ARGB format
	 * @return the handle of the gradient
	 */
	public static int initNative(int startX, int startY, int endX, int endY, int[] fractions, int[] colors) {
		final LinearGradient gradient = new LinearGradient(startX, startY, endX, endY, fractions, colors);
		final int handle = gradient.getHandle();

		gradients.put(handle, gradient);

		return handle;
	}

	/**
	 * Translates this gradient with the (x, y) vector.
	 *
	 * @param handle
	 *            the handle of the gradient on which the translation is applied.
	 * @param x
	 *            the x coordinate of the translation vector.
	 * @param y
	 *            the y coordinate of the translation vector.
	 */
	public static void translateNative(int handle, int x, int y) {
		final LinearGradient gradient = gradients.get(handle);
		gradient.translate(x, y);
	}

	/**
	 * Closes the given gradient.
	 *
	 * @param handle
	 *            the handle of the gradient to close.
	 */
	public static void closeNative(int handle) {
		gradients.remove(handle);
	}

	/**
	 * Tries to set the gradient to the graphic context.
	 *
	 * The gradient to apply is retrieved from its handle. This handle may be 0 if there is no gradient to apply. This
	 * can happen when the user sets a "simple" color.
	 *
	 * @param handle
	 *            the gradient handle in native code
	 * @param g2d
	 *            the graphic context in which the gradient will be paint
	 * @return true if the gradient has been set, false otherwise
	 */
	public static boolean setPaint(int handle, Graphics2D g2d) {
		if (handle == 0) {
			// No gradient to apply
			return false;
		}

		final LinearGradient gradient = gradients.get(handle);

		if (gradient == null) {
			// The handle does not correspond to a known gradient
			return false;
		}

		g2d.setPaint(gradient.inner);

		return true;
	}

	private LinearGradient(int startX, int startY, int endX, int endY, int[] intFractions, int[] intColors) {
		this.handle = nextHandle++;

		final float[] fractions = new float[intFractions.length];
		final Color[] colors = new Color[intColors.length];

		NativeUtils.refreshContent(intFractions);
		NativeUtils.refreshContent(intColors);

		for (int i = 0; i < fractions.length; i++) {
			// in FreeType API, fractions are integers from 0 to 255 (as in VG Lite APi) whereas in Java 2D they are
			// floats from 0 to 1
			fractions[i] = intFractions[i] / 255f;
		}

		for (int i = 0; i < colors.length; i++) {
			colors[i] = new Color(intColors[i], true);
		}

		this.inner = new LinearGradientPaint(startX, startY, endX, endY, fractions, colors);
	}

	private int getHandle() {
		// The handle cannot be 0 which means null.
		return this.handle + 1;
	}

	private void translate(int x, int y) {
		final Point2D start = this.inner.getStartPoint();
		final Point2D end = this.inner.getEndPoint();

		final float newStartX = (float) start.getX() + x;
		final float newStartY = (float) start.getY() + y;
		final float newEndX = (float) end.getX() + x;
		final float newEndY = (float) end.getY() + y;

		this.inner = new LinearGradientPaint(newStartX, newStartY, newEndX, newEndY, this.inner.getFractions(),
				this.inner.getColors());
	}
}
