/*
 * Copyright (c) 1995, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2016, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package java.net;

import sun.net.util.IPAddressUtil;

import com.is2t.support.DependencyInjectionHelper;
import com.is2t.support.net.IDNS;
import com.is2t.support.net.INetworkAddress;
import com.is2t.support.net.security.IURLSecurityManager;

public
class InetAddress implements java.io.Serializable {

	private static INetworkAddress VMInetAddress = (INetworkAddress) DependencyInjectionHelper.getImplementation(INetworkAddress.class);
	public static IDNS VMIDNS = getDNSImpl();
	private static IURLSecurityManager URLSecurityManager = (IURLSecurityManager) DependencyInjectionHelper.getImplementation(IURLSecurityManager.class);
	private static final String LOCALHOST =  "localhost";
	/**
	 * Specify the address family: Internet Protocol, Version 4
	 * @since 1.4
	 */
	static final int IPv4 = 1;

	/**
	 * Specify the address family: Internet Protocol, Version 6
	 * @since 1.4
	 */
	static final int IPv6 = 2;

	/* Specify address family preference */
	static transient boolean preferIPv6Address = false;

	static class InetAddressHolder {

		InetAddressHolder() {}

		InetAddressHolder(String hostName, int address, int family) {
			this.hostName = hostName;
			this.address = address;
			this.family = family;
		}

		void init(String hostName, int family) {
			this.hostName = hostName;
			if (family != -1) {
				this.family = family;
			}
		}

		String hostName;

		String getHostName() {
			return hostName;
		}

		/**
		 * Holds a 32-bit IPv4 address.
		 */
		int address;

		int getAddress() {
			return address;
		}

		/**
		 * Specifies the address family type, for instance, '1' for IPv4
		 * addresses, and '2' for IPv6 addresses.
		 */
		int family;

		int getFamily() {
			return family;
		}
	}

	/* Used to store the serializable fields of InetAddress */
	final transient InetAddressHolder holder;

	InetAddressHolder holder() {
		return holder;
	}


	/* Used to store the best available hostname */
	private transient String canonicalHostName = null;

	/** use serialVersionUID from JDK 1.0.2 for interoperability */
	private static final long serialVersionUID = 3286316764910316507L;


	InetAddress() {
		holder = new InetAddressHolder();
	}


	public boolean isMulticastAddress() {
		return false;
	}

	public boolean isAnyLocalAddress() {
		return false;
	}

	public boolean isLoopbackAddress() {
		return false;
	}

	public boolean isLinkLocalAddress() {
		return false;
	}

	public boolean isSiteLocalAddress() {
		return false;
	}

	public boolean isMCGlobal() {
		return false;
	}

	public boolean isMCNodeLocal() {
		return false;
	}

	public boolean isMCLinkLocal() {
		return false;
	}

	public boolean isMCSiteLocal() {
		return false;
	}

	public boolean isMCOrgLocal() {
		return false;
	}

	public String getHostName() {
		return getHostName(true);
	}

	String getHostName(boolean check) {
		if (holder().getHostName() == null) {
			holder().hostName = InetAddress.getHostFromNameService(this, check);
		}
		return holder().getHostName();
	}

	public String getCanonicalHostName() {
		if (canonicalHostName == null) {
			canonicalHostName =
					InetAddress.getHostFromNameService(this, true);
		}
		return canonicalHostName;
	}

	/**
	 * Returns the hostname for this address.
	 *
	 * <p>If there is a security manager, this method first
	 * calls its {@code checkConnect} method
	 * with the hostname and {@code -1}
	 * as its arguments to see if the calling code is allowed to know
	 * the hostname for this IP address, i.e., to connect to the host.
	 * If the operation is not allowed, it will return
	 * the textual representation of the IP address.
	 *
	 * @return  the host name for this IP address, or if the operation
	 *    is not allowed by the security check, the textual
	 *    representation of the IP address.
	 *
	 * @param check make security check if true
	 *
	 * @see SecurityManager#checkConnect
	 */
	private static String getHostFromNameService(InetAddress addr, boolean check) {
		String host = null;
		try {
			// first lookup the hostname
			host = VMIDNS.getHostByAddr(addr.getAddress());

			/* check to see if calling code is allowed to know
			 * the hostname for this IP address, ie, connect to the host
			 */
			if (check) {
				SecurityManager sec = System.getSecurityManager();
				if (sec != null) {
					URLSecurityManager.checkConnect(sec, host, -1);
				}
			}

			/* now get all the IP addresses for this hostname,
			 * and make sure one of them matches the original IP
			 * address. We do this to try and prevent spoofing.
			 */

			InetAddress[] arr = InetAddress.getAllByName0(host, check);
			boolean ok = false;

			if(arr != null) {
				for(int i = 0; !ok && i < arr.length; i++) {
					ok = addr.equals(arr[i]);
				}
			}

			//XXX: if it looks a spoof just return the address?
			if (!ok) {
				host = addr.getHostAddress();
				return host;
			}


		} catch (SecurityException e) {
			host = addr.getHostAddress();
		} catch (UnknownHostException e) {
			host = addr.getHostAddress();
			// let next provider resolve the hostname
		}

		return host;
	}

	public byte[] getAddress() {
		return null;
	}

	public String getHostAddress() {
		return null;
	}

	@Override
	public int hashCode() {
		return -1;
	}

	@Override
	public boolean equals(Object obj) {
		return false;
	}

	@Override
	public String toString() {
		String hostName = holder().getHostName();
		return ((hostName != null) ? hostName : "")
				+ "/" + getHostAddress();
	}

	static InetAddress[]    unknown_array; // put THIS in cache

	public static InetAddress getByAddress(String host, byte[] addr)
			throws UnknownHostException {
		if (host != null && host.length() > 0 && host.charAt(0) == '[') {
			if (host.charAt(host.length()-1) == ']') {
				host = host.substring(1, host.length() -1);
			}
		}
		if (addr != null) {
			if (addr.length == Inet4Address.INADDRSZ) {
				return new Inet4Address(host, addr);
			} else if (addr.length == Inet6Address.INADDRSZ) {
				byte[] newAddr
				= IPAddressUtil.convertFromIPv4MappedAddress(addr);
				if (newAddr != null) {
					return new Inet4Address(host, newAddr);
				} else {
					return new Inet6Address(host, addr);
				}
			}
		}
		throw new UnknownHostException("addr is of illegal length");
	}

    public static InetAddress getLoopbackAddress() {
    	try {
			return getByAddress(VMInetAddress.loopbackAddress());
		} catch (UnknownHostException e) {
			// given the specified address (loopback), this exception shouldn't happen
			throw new InternalError(e.getMessage());
		}
    }

    public static InetAddress getByName(String host)
			throws UnknownHostException {
		return InetAddress.getAllByName(host)[0];
	}


	public static InetAddress[] getAllByName(String host)
			throws UnknownHostException {
		return getAllByName(host, null);
	}

	private static InetAddress[] getAllByName(String host, InetAddress reqAddr)
			throws UnknownHostException {

		if (host == null || host.length() == 0) {
			InetAddress[] ret = new InetAddress[1];
			ret[0] = getByAddress(VMInetAddress.loopbackAddress());
			return ret;
		}

		boolean ipv6Expected = false;
		if (host.charAt(0) == '[') {
			// This is supposed to be an IPv6 literal
			if (host.length() > 2 && host.charAt(host.length()-1) == ']') {
				host = host.substring(1, host.length() -1);
				ipv6Expected = true;
			} else {
				// This was supposed to be a IPv6 address, but it's not!
				throw new UnknownHostException(host + ": invalid IPv6 address");
			}
		}

		// if host is an IP address, we won't do further lookup
		if (Character.digit(host.charAt(0), 16) != -1
				|| (host.charAt(0) == ':')) {
			byte[] addr = null;
			int numericZone = -1;
			String ifname = null;
			// see if it is IPv4 address
			addr = IPAddressUtil.textToNumericFormatV4(host);
			if (addr == null) {
				// This is supposed to be an IPv6 literal
				// Check if a numeric or string zone id is present
				int pos;
				if ((pos=host.indexOf ("%")) != -1) {
					numericZone = checkNumericZone (host);
					if (numericZone == -1) { /* remainder of string must be an ifname */
						ifname = host.substring (pos+1);
					}
				}
				if ((addr = IPAddressUtil.textToNumericFormatV6(host)) == null && host.contains(":")) {
					throw new UnknownHostException(host + ": invalid IPv6 address");
				}
			} else if (ipv6Expected) {
				// Means an IPv4 litteral between brackets!
				throw new UnknownHostException("["+host+"]");
			}
			InetAddress[] ret = new InetAddress[1];
			if(addr != null) {
				if (addr.length == Inet4Address.INADDRSZ) {
					ret[0] = new Inet4Address(null, addr);
				} else {
					if (ifname != null) {
						ret[0] = new Inet6Address(null, addr, ifname);
					} else {
						ret[0] = new Inet6Address(null, addr, numericZone);
					}
				}
				return ret;
			}
		} else if (ipv6Expected) {
			// We were expecting an IPv6 Litteral, but got something else
			throw new UnknownHostException("["+host+"]");
		}
		return getAllByName0(host, reqAddr, true);
	}

	/**
	 * check if the literal address string has %nn appended
	 * returns -1 if not, or the numeric value otherwise.
	 *
	 * %nn may also be a string that represents the displayName of
	 * a currently available NetworkInterface.
	 */
	private static int checkNumericZone (String s) throws UnknownHostException {
		int percent = s.indexOf ('%');
		int slen = s.length();
		int digit, zone=0;
		if (percent == -1) {
			return -1;
		}
		for (int i=percent+1; i<slen; i++) {
			char c = s.charAt(i);
			if (c == ']') {
				if (i == percent+1) {
					/* empty per-cent field */
					return -1;
				}
				break;
			}
			if ((digit = Character.digit (c, 10)) < 0) {
				return -1;
			}
			zone = (zone * 10) + digit;
		}
		return zone;
	}


	/**
	 * package private so SocketPermission can call it
	 */
	static InetAddress[] getAllByName0 (String host, boolean check)
			throws UnknownHostException  {
		return getAllByName0 (host, null, check);
	}

	private static InetAddress[] getAllByName0 (String host, InetAddress reqAddr, boolean check)
			throws UnknownHostException  {

		/* If it gets here it is presumed to be a hostname */
		/* Cache.get can return: null, unknownAddress, or InetAddress[] */

		/* make sure the connection to the host is allowed, before we
		 * give out a hostname
		 */
		if (check) {
			SecurityManager security = System.getSecurityManager();
			if (security != null) {
				URLSecurityManager.checkConnect(security, host, -1);
			}
		}

		InetAddress[] addresses = getAddressesFromNameService(host, reqAddr);

		if (addresses == unknown_array) {
			throw new UnknownHostException(host);
		}

		return addresses.clone();
	}

	private static InetAddress[] getAddressesFromNameService(String host, InetAddress reqAddr)
			throws UnknownHostException
			{
		InetAddress[] addresses = null;
		byte[][] addrs = null;
		boolean success = false;
		UnknownHostException ex = null;

		try {
			/*
			 * Do not put the call to lookup() inside the
			 * constructor.  if you do you will still be
			 * allocating space when the lookup fails.
			 */

			addrs = VMIDNS.getHostByName(host);
			addresses = new InetAddress[addrs.length];
			for(int i = -1; ++i < addrs.length;){
				addresses[i]  = InetAddress.getByAddress(host, addrs[i]);
			}
			success = true;
		} catch (UnknownHostException uhe) {
			if (host.equalsIgnoreCase(LOCALHOST)) {
				InetAddress[] local = new InetAddress[] { getByAddress(LOCALHOST, VMInetAddress.loopbackAddress()) };
				addresses = local;
				success = true;
			}
			else {
				addresses = unknown_array;
				success = false;
				ex = uhe;
			}
		}

		// More to do?
		if (reqAddr != null && addresses.length > 1 && !addresses[0].equals(reqAddr)) {
			// Find it?
			int i = 1;
			for (; i < addresses.length; i++) {
				if (addresses[i].equals(reqAddr)) {
					break;
				}
			}
			// Rotate
			if (i < addresses.length) {
				InetAddress tmp, tmp2 = reqAddr;
				for (int j = 0; j < i; j++) {
					tmp = addresses[j];
					addresses[j] = tmp2;
					tmp2 = tmp;
				}
				addresses[i] = tmp2;
			}
		}

		if (!success && ex != null){
			throw ex;
		}

		return addresses;
			}



	public static InetAddress getByAddress(byte[] addr)
			throws UnknownHostException {
		return getByAddress(null, addr);
	}


	public static InetAddress getLocalHost() throws UnknownHostException {

		SecurityManager security = System.getSecurityManager();
		try {
			String local = VMInetAddress.getLocalHostname();

			if (security != null) {
				URLSecurityManager.checkConnect(security, local, -1);
			}

			if (local.equals(LOCALHOST)) {
				return getByAddress(LOCALHOST, VMInetAddress.loopbackAddress());
			}

			InetAddress[] localAddrs;
			try {
				localAddrs =
						InetAddress.getAddressesFromNameService(local, null);
			} catch (UnknownHostException uhe) {
				// Rethrow with a more informative error message.
				UnknownHostException uhe2 =
						new UnknownHostException(local + ": " +
								uhe.getMessage());
				uhe2.initCause(uhe);
				throw uhe2;
			}
			return localAddrs[0];
		} catch (java.lang.SecurityException e) {
			return getByAddress(LOCALHOST, VMInetAddress.loopbackAddress());
		}
	}


	/*
	 * Returns the InetAddress representing anyLocalAddress
	 * (typically 0.0.0.0 or ::0)
	 */
	static InetAddress anyLocalAddress() {

		try {
			return getByAddress(VMInetAddress.lookupInaddrAny());
		} catch (UnknownHostException e) {
			return new Inet4Address("0.0.0.0", new byte[]{0,0,0,0}); //0.0.0.0
		}
	}

	public static IDNS getDNSImpl(){
		return (IDNS) DependencyInjectionHelper.getImplementation(IDNS.class, false);
	}

}
